@@ -58,8 +58,8 @@ menu('&File',
      ]).
 menu('&Settings',
      [ --,
-       '&User init file ...' = prolog_edit_preferences(prolog),
-       '&Stack sizes ...'    = prolog_edit_preferences(stack_sizes)
+       '&User init file ...'  = prolog_edit_preferences(prolog),
+       '&GUI preferences ...' = prolog_edit_preferences(xpce)
      ],
      []).
 menu('&Debug',
@@ -77,23 +77,19 @@ menu('&Debug',
      [ before_menu(-)
      ]).
 menu('&Help',
-     [ '&About ...' 				= about,
-       '&Online manual ...' 			= help,
-       '&Package documentation ...' 		= html_open(swi('doc/packages/index.html')),
+     [ '&About ...'				= about,
+       '&Help ...'				= help,
+       'SWI-Prolog &Manual (on www) ...'	= www_open(swipl_man),
        --,
-       'SWI-Prolog &WWW home (on www) ...'	= www_open(pl),
-       'SWI-Prolog &FAQ (on www) ...'		= www_open(pl_faq),
-       'SWI-Prolog &Quick Start (on www) ...'	= www_open(pl_quick),
-       'SWI-Prolog &Manual (on www) ...'	= www_open(pl_man),
-       'SWI-Prolog Mailing &List (on www) ...'	= www_open(pl_mail),
-       'SWI-Prolog &Download (on www) ...'	= www_open(pl_download),
+       'SWI-Prolog &WWW home (on www) ...'	= www_open(swipl),
+       'SWI-Prolog &FAQ (on www) ...'		= www_open(swipl_faq),
+       'SWI-Prolog &Quick Start (on www) ...'	= www_open(swipl_quick),
+       'SWI-Prolog Mailing &List (on www) ...'	= www_open(swipl_mail),
+       'SWI-Prolog &Download (on www) ...'	= www_open(swipl_download),
        --,
-       '&XPCE (GUI) Manual ...' 		= manpce,
-       'XPCE &User Guide (on www) ...'   	= www_open(xpce_man),
+       '&XPCE (GUI) Manual ...'			= manpce,
        --,
-       'Submit &Bug report (on www) ...'  	= www_open(pl_bugs),
-       --,
-       '&Donate to the SWI-Prolog project (on www) ...' = www_open(pl_donate)
+       'Submit &Bug report (on www) ...'	= www_open(swipl_bugs)
      ],
      [ before_menu(-)
      ]).
@@ -185,7 +181,7 @@ gather_arg(file(Mode, Title), File) :-
 	Filter =.. [chain|AllTuples],
 	current_prolog_flag(hwnd, HWND),
 	working_directory(CWD, CWD),
-	call(get(@display, win_file_name, 	% avoid autoloading
+	call(get(@display, win_file_name,	% avoid autoloading
 		 Mode, Filter, Title,
 		 directory := CWD,
 		 owner := HWND,
