@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2005, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -57,18 +58,20 @@ static atom_t ATOM_unicode_le;
 static atom_t ATOM_wchar_t;
 static atom_t ATOM_read;
 static atom_t ATOM_write;
+static atom_t ATOM_free_on_close;
 
 #define MEMFILE_MAGIC	0x5624a6b3L
-#define NOSIZE ((unsigned int)-1)
+#define NOSIZE ((size_t)-1)
 
 typedef struct
 { long		magic;			/* MEMFILE_MAGIC */
   IOENC		encoding;		/* encoding of the data */
+  int		free_on_close;		/* free if it is closed */
   char	       *data;			/* data of the file */
   size_t	data_size;		/* byte-size of data */
   size_t	size;			/* size in characters */
   IOSTREAM     *stream;			/* Stream hanging onto it */
-  atom_t 	atom;			/* Created from atom */
+  atom_t	atom;			/* Created from atom */
 } memfile;
 
 
@@ -86,7 +89,7 @@ get_memfile(term_t handle, memfile **f)
   { term_t a = PL_new_term_ref();
     void *ptr;
 
-    PL_get_arg(1, handle, a);
+    _PL_get_arg(1, handle, a);
     if ( PL_get_pointer(a, &ptr) )
     { memfile *m = ptr;
 
@@ -109,7 +112,8 @@ new_memory_file(term_t handle)
 { memfile *m = calloc(1, sizeof(*m));
 
   if ( !m )
-    return pl_error(NULL, 0, NULL, ERR_ERRNO, errno, "create", "memory_file", handle);
+    return pl_error(NULL, 0, NULL, ERR_ERRNO, errno,
+		    "create", "memory_file", handle);
 
   m->magic = MEMFILE_MAGIC;
   m->encoding = ENC_UTF8;
@@ -125,22 +129,27 @@ new_memory_file(term_t handle)
 }
 
 
+static int
+destroy_memory_file(memfile *m)
+{ if ( m->stream )
+    Sclose(m->stream);
+  if ( m->atom )
+    PL_unregister_atom(m->atom);
+  else if ( m->data )
+    Sfree(m->data);			/* MS-Windows: malloc by other DLL! */
+  m->magic = 0;
+  free(m);
+
+  return TRUE;
+}
+
+
 static foreign_t
 free_memory_file(term_t handle)
 { memfile *m;
 
   if ( get_memfile(handle, &m) )
-  { if ( m->stream )
-      Sclose(m->stream);
-    if ( m->atom )
-      PL_unregister_atom(m->atom);
-    else if ( m->data )
-      Sfree(m->data);			/* MS-Windows: malloc by other DLL! */
-    m->magic = 0;
-    free(m);
-
-    return TRUE;
-  }
+    return destroy_memory_file(m);
 
   return FALSE;
 }
@@ -151,6 +160,8 @@ closehook(void *closure)
 { memfile *m = closure;
 
   m->stream = NULL;
+  if ( m->free_on_close )
+    destroy_memory_file(m);
 }
 
 
@@ -164,7 +175,7 @@ alreadyOpen(term_t handle, const char *op)
 static struct encname
 { IOENC  code;
   atom_t *name;
-} encoding_names[] = 
+} encoding_names[] =
 { { ENC_UNKNOWN,     &ATOM_unknown },
   { ENC_OCTET,       &ATOM_octet },
   { ENC_ASCII,       &ATOM_ascii },
@@ -178,12 +189,12 @@ static struct encname
 };
 
 
-IOENC
+static IOENC
 atom_to_encoding(atom_t a)
 { struct encname *en;
 
   for(en=encoding_names; en->name; en++)
-  { if ( *en->name == a ) 
+  { if ( *en->name == a )
       return en->code;
   }
 
@@ -216,6 +227,7 @@ open_memory_file4(term_t handle, term_t mode, term_t stream, term_t options)
   atom_t iom;
   IOSTREAM *fd;
   IOENC encoding;
+  int free_on_close = FALSE;
 
   if ( !get_memfile(handle, &m) )
     return FALSE;
@@ -237,9 +249,15 @@ open_memory_file4(term_t handle, term_t mode, term_t stream, term_t options)
       if ( PL_get_name_arity(head, &name, &arity) && arity == 1 )
       { term_t arg = PL_new_term_ref();
 
-	PL_get_arg(1, head, arg);
-	if ( !get_encoding(arg, &encoding) )
-	  return FALSE;
+	_PL_get_arg(1, head, arg);
+	if ( name == ATOM_encoding )
+	{ if ( !get_encoding(arg, &encoding) )
+	    return FALSE;
+	} else if ( name == ATOM_free_on_close )
+	{ if ( !PL_get_bool(arg, &free_on_close) )
+	    return pl_error("open_memory_file", 4, NULL, ERR_TYPE,
+			    arg, "boolean");
+	}
       } else
 	return pl_error("open_memory_file", 4, NULL, ERR_TYPE, head, "option");
     }
@@ -261,6 +279,7 @@ open_memory_file4(term_t handle, term_t mode, term_t stream, term_t options)
     m->encoding = encoding;
   } else if ( iom == ATOM_read )
   { x = "r";
+    m->free_on_close = free_on_close;
   } else
   { return pl_error("open_memory_file", 3, NULL, ERR_DOMAIN,
 		    mode, "io_mode");
@@ -287,39 +306,68 @@ open_memory_file(term_t handle, term_t mode, term_t stream)
 
 
 static foreign_t
-size_memory_file(term_t handle, term_t size)
+size_memory_file(term_t handle, term_t sizeh, term_t encoding)
 { memfile *m;
 
   if ( get_memfile(handle, &m) )
   { if ( m->stream && !m->atom )
       return alreadyOpen(handle, "size");
+
     if ( m->data )
-    { if ( m->size == NOSIZE )
-      { switch( m->encoding )
+    { IOENC size_enc = m->encoding;
+      size_t size;
+
+      if ( encoding )
+      { if ( !get_encoding(encoding, &size_enc) )
+	  return FALSE;
+      } else
+	size_enc = m->encoding;
+
+      if ( m->size != NOSIZE && size_enc == m->encoding )
+      { size = m->size;
+      } else
+      { switch( size_enc )
 	{ case ENC_ISO_LATIN_1:
 	  case ENC_OCTET:
-	    m->size = m->data_size;
+	    size = m->data_size;
 	    break;
 	  case ENC_WCHAR:
-	    m->size = m->data_size / sizeof(wchar_t);
+	    size = m->data_size / sizeof(wchar_t);
 	    break;
 	  case ENC_UTF8:
-	    m->size = PL_utf8_strlen(m->data, m->data_size);
+	    size = PL_utf8_strlen(m->data, m->data_size);
 	    break;
 	  default:
 	    assert(0);
 	    return FALSE;
 	}
+
+	if ( size_enc == m->encoding )
+	  m->size = size;
       }
-      return PL_unify_integer(size, m->size);
+
+      return PL_unify_int64(sizeh, size);
     } else
-      return PL_unify_integer(size, 0);
+    { return PL_unify_integer(sizeh, 0);
+    }
   }
 
   return FALSE;
 }
 
 
+static foreign_t
+size_memory_file2(term_t handle, term_t size)
+{ return size_memory_file(handle, size, 0);
+}
+
+
+static foreign_t
+size_memory_file3(term_t handle, term_t size, term_t encoding)
+{ return size_memory_file(handle, size, encoding);
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 utf8_position_memory_file(+MF, -Here, -Size)
 
@@ -403,7 +451,7 @@ memory_file_to_text(term_t handle, term_t atom, term_t encoding, int flags)
 
   if ( get_memfile(handle, &m) )
   { IOENC enc;
-  
+
     if ( encoding )
     { if ( !get_encoding(encoding, &enc) )
 	return FALSE;
@@ -478,10 +526,12 @@ install_memfile()
   MKATOM(wchar_t);
   MKATOM(read);
   MKATOM(write);
+  MKATOM(free_on_close);
 
   PL_register_foreign("new_memory_file",      1, new_memory_file,      0);
   PL_register_foreign("free_memory_file",     1, free_memory_file,     0);
-  PL_register_foreign("size_memory_file",     2, size_memory_file,     0);
+  PL_register_foreign("size_memory_file",     2, size_memory_file2,    0);
+  PL_register_foreign("size_memory_file",     3, size_memory_file3,    0);
   PL_register_foreign("open_memory_file",     3, open_memory_file,     0);
   PL_register_foreign("open_memory_file",     4, open_memory_file4,    0);
   PL_register_foreign("atom_to_memory_file",  2, atom_to_memory_file,  0);
