@@ -127,6 +127,13 @@ leave the details to this function.
 #include <malloc.h>
 #endif
 
+#ifdef __WINDOWS__
+#define GET_ERRNO WSAGetLastError()
+#define GET_H_ERRNO WSAGetLastError()
+#else
+#define GET_ERRNO errno
+#define GET_H_ERRNO h_errno
+#endif
 
 #ifndef __WINDOWS__
 #define closesocket(n) close((n))	/* same on Unix */
@@ -147,30 +154,41 @@ leave the details to this function.
 static CRITICAL_SECTION mutex;
 static CRITICAL_SECTION mutex_free;
 
-#define LOCK() EnterCriticalSection(&mutex)
-#define UNLOCK() LeaveCriticalSection(&mutex)
-#define LOCK_FREE() EnterCriticalSection(&mutex_free)
-#define UNLOCK_FREE() LeaveCriticalSection(&mutex_free)
-#define INITLOCK() (InitializeCriticalSection(&mutex), \
-		    InitializeCriticalSection(&mutex_free))
-#else
+#define LOCK()			EnterCriticalSection(&mutex)
+#define UNLOCK()		LeaveCriticalSection(&mutex)
+#define LOCK_FREE()		EnterCriticalSection(&mutex_free)
+#define UNLOCK_FREE()		LeaveCriticalSection(&mutex_free)
+#define INITLOCK()		( InitializeCriticalSection(&mutex), \
+				  InitializeCriticalSection(&mutex_free))
+#define LOCK_SOCKET(s)		EnterCriticalSection(&s->socket_mutex)
+#define UNLOCK_SOCKET(s)	LeaveCriticalSection(&s->socket_mutex)
+#define INIT_SOCKET_LOCK(s)	InitializeCriticalSection(&s->socket_mutex)
+#define FREE_SOCKET_LOCK(s)	DeleteCriticalSection(&s->socket_mutex)
+#else /*__WINDOWS__*/
 #include <pthread.h>
 
 static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
 
-#define LOCK() pthread_mutex_lock(&mutex)
-#define UNLOCK() pthread_mutex_unlock(&mutex)
-#define LOCK_FREE()
-#define UNLOCK_FREE()
-#define INITLOCK()
-#endif
-#else
-#define LOCK()
-#define UNLOCK()
-#define LOCK_FREE()
-#define UNLOCK_FREE()
-#define INITLOCK()
+#define LOCK()			pthread_mutex_lock(&mutex)
+#define UNLOCK()		pthread_mutex_unlock(&mutex)
+#define LOCK_FREE()		(void)0
+#define UNLOCK_FREE()		(void)0
+#define INITLOCK()		(void)0
+#define FREE_SOCKET_LOCK(s)	(void)0
+#endif /*__WINDOWS__*/
+#else /* _REENTRANT */
+#define LOCK()			(void)0
+#define UNLOCK()		(void)0
+#define LOCK_FREE()		(void)0
+#define UNLOCK_FREE()		(void)0
+#define INITLOCK()		(void)0
+#define FREE_SOCKET_LOCK(s)	(void)0
+#if __WINDOWS__
+#define LOCK_SOCKET(s)		(void)0
+#define UNLOCK_SOCKET(s)	(void)0
+#define INIT_SOCKET_LOCK(s)	(void)0
 #endif
+#endif /* _REENTRANT */
 
 #define set(s, f)   ((s)->flags |= (f))
 #define clear(s, f) ((s)->flags &= ~(f))
@@ -196,8 +214,10 @@ typedef struct _plsocket
   nbio_request	    request;		/* our request */
   DWORD		    thread;		/* waiting thread */
   DWORD		    error;		/* error while executing request */
+  DWORD             close_timeout;      /* Time to reap socket if FD_CLOSE is not received */
   int		    done;		/* request completed */
   int		    w32_flags;		/* or of received FD_* */
+  CRITICAL_SECTION  socket_mutex;       /* synchronization mutex */
   union
   { struct
     { struct sockaddr_in addr;		/* accepted address */
@@ -230,10 +250,10 @@ typedef struct _plsocket
     struct
     { int bytes;			/* byte count */
       void *buffer;			/* the buffer */
-      size_t size;			/* buffer size */
+      int size;				/* buffer size */
       int flags;
       const struct sockaddr *to;
-      socklen_t tolen;
+      int tolen;
     } sendto;
   } rdata;
 #endif
@@ -242,6 +262,7 @@ typedef struct _plsocket
 static plsocket *allocSocket(SOCKET socket);
 #ifdef __WINDOWS__
 static plsocket *lookupOSSocket(SOCKET socket);
+static int socketIsPendingClose(plsocket *s);
 static const char *WinSockError(unsigned long eno);
 #endif
 
@@ -256,7 +277,7 @@ need_retry(int error)
 #ifdef O_DEBUG
 static int debugging;
 
-int
+NBIO_EXPORT(int)
 nbio_debug(int level)
 { int old = debugging;
 
@@ -377,9 +398,12 @@ static int
 doneRequest(plsocket *s)
 { SOCKET sock;
 
+  LOCK_SOCKET(s);
   s->done = TRUE;
   s->request = REQ_NONE;
+  UNLOCK_SOCKET(s);
 
+  /* If we have FD_CLOSE, then we know we can release the socket to the OS */
   if ( (s->w32_flags & FD_CLOSE) && (sock=s->socket) >= 0 )
   { s->socket = -1;
     closesocket(sock);
@@ -435,11 +459,13 @@ nbio_wait(nbio_sock_t socket, nbio_request request)
   if ( !(s=nbio_to_plsocket(socket)) )
     return -1;
 
+  LOCK_SOCKET(s);
   s->flags  |= PLSOCK_WAITING;
   s->done    = FALSE;
   s->error   = 0;
   s->thread  = GetCurrentThreadId();
   s->request = request;
+  UNLOCK_SOCKET(s);
 
   SendMessage(State()->hwnd, WM_REQUEST, 1, (LPARAM)&s);
 
@@ -501,11 +527,13 @@ nbio_select(int n,
       { plsocket *s = nbio_to_plsocket(i);
 
 	if ( s )
-	{ s->flags  |= PLSOCK_WAITING;
+	{ LOCK_SOCKET(s);
+	  s->flags  |= PLSOCK_WAITING;
 	  s->done    = FALSE;
 	  s->error   = 0;
 	  s->thread  = GetCurrentThreadId();
 	  s->request = (s->flags & PLSOCK_LISTEN) ? REQ_ACCEPT : REQ_READ;
+	  UNLOCK_SOCKET(s);
 	  sockets[i] = s;
 	} else
 	{ DEBUG(2, Sdprintf("nbio_select(): no socket for %d\n", i));
@@ -586,11 +614,13 @@ placeRequest(plsocket *s, nbio_request request)
 { if ( s->magic != PLSOCK_MAGIC )
     Sdprintf("placeRequest: %p has bad magic\n", s);
 
+  LOCK_SOCKET(s);
   s->error   = 0;
   s->done    = FALSE;
   s->thread  = GetCurrentThreadId();
   s->request = request;
   clear(s, PLSOCK_WAITING);
+  UNLOCK_SOCKET(s);
 
   SendMessage(State()->hwnd, WM_REQUEST, 1, (LPARAM)&s);
   DEBUG(2, Sdprintf("%d (%ld): Placed %s request for %d\n",
@@ -710,7 +740,9 @@ doRequest(plsocket *s)
       break;
     case REQ_RECVFROM:
       if ( s->w32_flags & (FD_READ|FD_CLOSE) )
-      { s->w32_flags &= ~FD_READ;
+      { int iflen = (int)*s->rdata.recvfrom.fromlen;
+
+	s->w32_flags &= ~FD_READ;
 
 	if ( true(s, PLSOCK_WAITING) )
 	{ doneRequest(s);
@@ -723,7 +755,7 @@ doRequest(plsocket *s)
 			 (int)s->rdata.recvfrom.size,
 			 s->rdata.recvfrom.flags,
 			 s->rdata.recvfrom.from,
-			 s->rdata.recvfrom.fromlen);
+			 &iflen);
 
 	if ( s->rdata.recvfrom.bytes < 0 )
 	{ s->error = WSAGetLastError();
@@ -734,7 +766,8 @@ doRequest(plsocket *s)
 	    doneRequest(s);
 	  }
 	} else
-	{ doneRequest(s);
+	{ *s->rdata.recvfrom.fromlen = iflen;
+	  doneRequest(s);
 	}
       }
       break;
@@ -772,7 +805,14 @@ doRequest(plsocket *s)
 	{ s->rdata.write.bytes = (int)s->rdata.write.written;
 	  doneRequest(s);
 	}
+      } else if ( s->w32_flags & FD_CLOSE )
+      { if ( s->rdata.write.written > 0 )
+	  s->rdata.write.bytes = (int)s->rdata.write.written;
+	else
+	  s->rdata.write.bytes = -1;
+	doneRequest(s);
       }
+      break;
     case REQ_SENDTO:
       if ( s->w32_flags & FD_WRITE )
       { int n;
@@ -806,6 +846,7 @@ doRequest(plsocket *s)
 	s->rdata.sendto.bytes = n;
 	doneRequest(s);
       }
+      break;
   }
 
   return TRUE;
@@ -881,8 +922,13 @@ socket_wnd_proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
 
 	s->error = err;
 	if ( sock )
-	{ s->socket = -1;
-	  closesocket(sock);
+	{ /* We cannot close the socket yet, since the late arrival
+             of FD_CLOSE might be delivered to this socket even after
+	     it has been reallocated. Instead, calculate a timeout to
+             allow for the case when FD_CLOSE never comes, and then
+	     continue as normal
+          */
+          socketIsPendingClose(s);
 	}
 
 	switch(s->request)
@@ -1077,7 +1123,7 @@ nbio_fcntl(nbio_sock_t socket, int op, int arg)
   { if ( op == F_SETFL && arg == O_NONBLOCK )
       s->flags |= PLSOCK_NONBLOCK;
   } else
-    nbio_error(errno, TCP_ERRNO);
+    nbio_error(GET_ERRNO, TCP_ERRNO);
 
   return rc;
 }
@@ -1201,7 +1247,7 @@ nbio_to_plsocket(nbio_sock_t socket)
 }
 
 
-SOCKET
+NBIO_EXPORT(SOCKET)
 nbio_fd(nbio_sock_t socket)
 { plsocket *p;
 
@@ -1225,6 +1271,8 @@ allocSocket(SOCKET socket)
   size_t i;
 
 #ifdef __WINDOWS__
+  DWORD now = GetTickCount();
+
   if ( (p=lookupOSSocket(socket)) )
   { DEBUG(1, Sdprintf("WinSock %d already registered on %d\n",
 		      (int)socket, p->id));
@@ -1249,10 +1297,25 @@ allocSocket(SOCKET socket)
   }
 
   for(i=0; i<socks_allocated; i++)
-  { if ( sockets[i] == NULL )
+  { if ( (p=sockets[i]) == NULL )
     { sockets[i] = p = PL_malloc(sizeof(*p));
       socks_count++;
       break;
+#ifdef __WINDOWS__
+    } else
+    { if ( p->close_timeout && p->close_timeout < now )
+      { SOCKET sock;
+
+	if ( (sock=p->socket) )		/* is this ever the case? */
+	{ int rval;
+
+	  do
+	  { rval = closesocket(sock);
+	  } while(rval == SOCKET_ERROR && errno == EINTR);
+	}
+	break;
+      }
+#endif
     }
   }
   UNLOCK();
@@ -1267,6 +1330,7 @@ allocSocket(SOCKET socket)
 #ifdef __WINDOWS__
   p->w32_flags = 0;
   p->request   = REQ_NONE;
+  INIT_SOCKET_LOCK(p);
 #endif
   p->input = p->output = (IOSTREAM*)NULL;
 
@@ -1298,6 +1362,7 @@ freeSocket(plsocket *s)
   sock = s->socket;
   socket = s->id;
   s->magic = 0;
+  FREE_SOCKET_LOCK(s);
   PL_free(s);
   UNLOCK_FREE();
 
@@ -1322,13 +1387,13 @@ freeSocket(plsocket *s)
 		 *******************************/
 
 #ifdef __WINDOWS__
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 The code in BILLY_GETS_BETTER is, according to various documents the
 right code, but it doesn't work, so we do it by hand.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 #ifdef BILLY_GETS_BETTER
-
 static const char *
 WinSockError(unsigned long eno)
 { char buf[1024];
@@ -1410,10 +1475,10 @@ WinSockError(unsigned long error)
     { WSANOTINITIALISED, "Socket layer not initialised" },
 					/* WinSock 2 errors */
     { WSAHOST_NOT_FOUND, "Host not found" },
+    { WSANO_DATA, "Valid name, no data record of requested type" },
     { 0, NULL }
   };
   char tmp[100];
-
   for(ep=edefs; ep->string; ep++)
   { if ( ep->index == (int)error )
       return ep->string;
@@ -1464,7 +1529,6 @@ nbio_error(int code, nbio_error_map mapid)
 
   if ( code == EPLEXCEPTION )
     return FALSE;
-
   switch( mapid )
   { case TCP_HERRNO:
       map = h_errno_codes;
@@ -1472,7 +1536,6 @@ nbio_error(int code, nbio_error_map mapid)
     default:
       map = NULL;
   }
-
   {
 #ifdef __WINDOWS__
   msg = WinSockError(code);
@@ -1495,11 +1558,12 @@ nbio_error(int code, nbio_error_map mapid)
     msg = strerror(code);
 #endif /*__WINDOWS__*/
 
-  PL_unify_term(except,
-		CompoundArg("error", 2),
-		  CompoundArg("socket_error", 1),
-		    AtomArg(msg),
-		  PL_VARIABLE);
+  if ( !PL_unify_term(except,
+		      CompoundArg("error", 2),
+		        CompoundArg("socket_error", 1),
+		          AtomArg(msg),
+		        PL_VARIABLE) )
+    return FALSE;
   }
 
   return PL_raise_exception(except);
@@ -1527,17 +1591,14 @@ nbio_last_error(nbio_sock_t socket)
 		 *	  INITIALISATION	*
 		 *******************************/
 
-int
+NBIO_EXPORT(int)
 nbio_init(const char *module)
-{ INITLOCK();				/* is this ok? */
-
-  LOCK();
-  if ( initialised )
-  { UNLOCK();
-    return TRUE;
-  }
+{ if ( initialised )			/* called from install handlers, which */
+    return TRUE;			/* are serialized by the compiler mutex */
   initialised = TRUE;
 
+  INITLOCK();
+
   FUNCTOR_module2  = PL_new_functor(PL_new_atom(":"), 2);
   FUNCTOR_ip4	   = PL_new_functor(PL_new_atom("ip"), 4);
   FUNCTOR_ip1	   = PL_new_functor(PL_new_atom("ip"), 1);
@@ -1558,19 +1619,16 @@ nbio_init(const char *module)
 #endif
 
   if ( WSAStartup(MAKEWORD(2,0), &WSAData) )
-  { UNLOCK();
     return PL_warning("nbio_init() - WSAStartup failed.");
-  }
   startSocketThread();
 }
 #endif /*__WINDOWS__*/
 
-  UNLOCK();
   return TRUE;
 }
 
 
-int
+NBIO_EXPORT(int)
 nbio_cleanup(void)
 { if ( initialised )
   {
@@ -1582,6 +1640,25 @@ nbio_cleanup(void)
   return 0;
 }
 
+#ifdef __WINDOWS__
+
+/* socketIsPendingClose() is called if a Windows socket is closed, but
+   we did not yet see an FD_CLOSE message for it.  We will reuse this
+   socket using lookupTimedOutSocket() if we have not seen the FD_CLOSE
+   after 30 seconds.  This is a random time.  Are we guaranteed to get
+   an FD_CLOSE?
+*/
+
+static int
+socketIsPendingClose(plsocket *s)
+{ s->close_timeout = GetTickCount() + 30 * 1000;
+  shutdown(s->socket, SD_BOTH);
+  DEBUG(3, Sdprintf("Setting timeout for FD_CLOSE on socket %d...\n", s->id));
+
+  return 0;
+}
+
+#endif /*__WINDOWS__*/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 socket(-Socket)
@@ -1589,7 +1666,7 @@ socket(-Socket)
     the format $socket(Id).
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-nbio_sock_t
+NBIO_EXPORT(nbio_sock_t)
 nbio_socket(int domain, int type, int protocol)
 { SOCKET sock;
   plsocket *s;
@@ -1597,10 +1674,9 @@ nbio_socket(int domain, int type, int protocol)
   assert(initialised);
 
   if ( (sock = socket(domain, type , protocol)) < 0)
-  { nbio_error(errno, TCP_ERRNO);
+  { nbio_error(GET_ERRNO, TCP_ERRNO);
     return -1;
   }
-
   if ( !(s=allocSocket(sock)) )		/* register it */
   { closesocket(sock);
     return -1;
@@ -1615,7 +1691,7 @@ nbio_socket(int domain, int type, int protocol)
 }
 
 
-int
+NBIO_EXPORT(int)
 nbio_closesocket(nbio_sock_t socket)
 { plsocket *s;
 
@@ -1637,20 +1713,18 @@ nbio_closesocket(nbio_sock_t socket)
     }
   } else
   {
+  // We cannot free the socket in Windows since we might subsequently get an FD_CLOSE. Instead set the timeout
 #ifdef __WINDOWS__
-    if ( true(s, PLSOCK_CONNECT) )
-    { if ( s->socket >= 0 )
-	shutdown(s->socket, SD_SEND);
-    }
+  socketIsPendingClose(s);
+#else
+  freeSocket(s);
 #endif
-
-    freeSocket(s);
   }
 
   return 0;
 }
 
-int
+NBIO_EXPORT(int)
 nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 { plsocket *s;
   va_list args;
@@ -1670,7 +1744,7 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 
       if( setsockopt(s->socket, SOL_SOCKET, SO_REUSEADDR,
 		     (const char *)&val, sizeof(val)) == -1 )
-      { nbio_error(h_errno, TCP_HERRNO);
+      { nbio_error(GET_H_ERRNO, TCP_HERRNO);
 	rc = -1;
       } else
 	rc = 0;
@@ -1686,7 +1760,7 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 #endif
       if ( setsockopt(s->socket, IPPROTO_TCP, TCP_NODELAY,
 		      (const char *)&val, sizeof(val)) == -1 )
-      { nbio_error(h_errno, TCP_HERRNO);
+      { nbio_error(GET_H_ERRNO, TCP_HERRNO);
 	rc = -1;
       } else
       { rc = 0;
@@ -1702,7 +1776,7 @@ nbio_setopt(nbio_sock_t socket, nbio_option opt, ...)
 
       if ( setsockopt(s->socket, SOL_SOCKET, SO_BROADCAST,
 		     (const char *)&val, sizeof(val)) == -1 )
-      { nbio_error(h_errno, TCP_HERRNO);
+      { nbio_error(GET_H_ERRNO, TCP_HERRNO);
 	rc = -1;
       } else
 	rc = 0;
@@ -1792,11 +1866,13 @@ nbio_get_port(term_t Port, int *port)
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Convert a term Host:Port to a socket address.  Port is either an integer
 or the name of a registered port (e.g. 'smtp').
+
+(*) TBD: Supply the port/service here too, simplifying the rest
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
 nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr)
-{ int		  port;
+{ int port;
 
   addr->sin_family = AF_INET;
   addr->sin_addr.s_addr = INADDR_ANY;
@@ -1805,20 +1881,25 @@ nbio_get_sockaddr(term_t Address, struct sockaddr_in *addr)
   { char *hostName;
     term_t arg = PL_new_term_ref();
 
-    PL_get_arg(1, Address, arg);
+    _PL_get_arg(1, Address, arg);
     if ( PL_get_atom_chars(arg, &hostName) )
-    { struct hostent *host;
+    { struct addrinfo hints;
+      struct addrinfo *res;
 
-      if( !(host = gethostbyname(hostName)) )
-	return nbio_error(h_errno, TCP_HERRNO);
-      if ( (int)sizeof(addr->sin_addr) < host->h_length )
-	return PL_warning("Oops, host address too long!");
-      memcpy(&addr->sin_addr, host->h_addr, host->h_length);
+      memset(&hints, 0, sizeof(hints));
+      hints.ai_family = AF_INET;
+      if ( getaddrinfo(hostName, NULL, &hints, &res) !=  0) /* see (*) */
+	return nbio_error(GET_H_ERRNO, TCP_HERRNO);
+      assert(res->ai_family == AF_INET);
+      memcpy(&addr->sin_addr,
+	     &((struct sockaddr_in*)res->ai_addr)->sin_addr,
+	     sizeof(addr->sin_addr));
+      freeaddrinfo(res);
     } else if ( !nbio_get_ip(arg, &addr->sin_addr) )
     { return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1, arg, "atom|ip/4");
     }
 
-    PL_get_arg(2, Address, arg);
+    _PL_get_arg(2, Address, arg);
     if ( !nbio_get_port(arg, &port) )
       return FALSE;
   } else if ( PL_is_variable(Address) )
@@ -1841,7 +1922,7 @@ nbio_get_ip(term_t ip4, struct in_addr *ip)
     term_t a = PL_new_term_ref();
 
     for(i=1; i<=4; i++)
-    { PL_get_arg(i, ip4, a);
+    { _PL_get_arg(i, ip4, a);
       if ( PL_get_integer(a, &ia) )
 	hip |= ia << ((4-i)*8);
       else
@@ -1855,7 +1936,7 @@ nbio_get_ip(term_t ip4, struct in_addr *ip)
   { term_t a = PL_new_term_ref();
     atom_t id;
 
-    PL_get_arg(1, ip4, a);
+    _PL_get_arg(1, ip4, a);
     if ( PL_get_atom(a, &id) )
     { if ( id == ATOM_any )
 	ip->s_addr = INADDR_ANY;
@@ -1897,7 +1978,8 @@ nbio_bind(nbio_sock_t socket, struct sockaddr *my_addr, size_t addrlen)
 #else
   if ( bind(s->socket, my_addr, addrlen) )
 #endif
-  { nbio_error(errno, TCP_ERRNO);
+  {
+    nbio_error(GET_ERRNO, TCP_ERRNO);
     return -1;
   }
 
@@ -1935,12 +2017,12 @@ nbio_connect(nbio_sock_t socket,
 #else /*!__WINDOWS__*/
   for(;;)
   { if ( connect(s->socket, serv_addr, addrlen) )
-    { if ( need_retry(errno) )
+    { if ( need_retry(GET_ERRNO) )
       { if ( PL_handle_signals() < 0 )
 	  return -1;
 	continue;
       }
-      nbio_error(errno, TCP_ERRNO);
+      nbio_error(GET_ERRNO, TCP_ERRNO);
       return -1;
     } else
       break;
@@ -1996,13 +2078,13 @@ nbio_accept(nbio_sock_t master, struct sockaddr *addr, socklen_t *addrlen)
     slave = accept(m->socket, addr, addrlen);
 
     if ( slave == SOCKET_ERROR )
-    { if ( need_retry(errno) )
+    { if ( need_retry(GET_ERRNO) )
       { if ( PL_handle_signals() < 0 )
 	  return -1;
 
 	continue;
       } else
-      { nbio_error(errno, TCP_ERRNO);
+      { nbio_error(GET_ERRNO, TCP_ERRNO);
 	return -1;
       }
     } else
@@ -2030,7 +2112,7 @@ nbio_listen(nbio_sock_t socket, int backlog)
     return -1;
 
   if( listen(s->socket, backlog) == -1 )
-  { nbio_error(errno, TCP_ERRNO);
+  { nbio_error(GET_ERRNO, TCP_ERRNO);
     return -1;
   }
 
@@ -2092,7 +2174,7 @@ nbio_read(int socket, char *buf, size_t bufSize)
 
     n = recv(s->socket, buf, bufSize, 0);
 
-    if ( n == -1 && need_retry(errno) )
+    if ( n == -1 && need_retry(GET_ERRNO) )
     { if ( PL_handle_signals() < 0 )
       { errno = EPLEXCEPTION;
 	return -1;
@@ -2164,7 +2246,7 @@ nbio_write(nbio_sock_t socket, char *buf, size_t bufSize)
 
     n = send(s->socket, str, len, 0);
     if ( n < 0 )
-    { if ( need_retry(errno) )
+    { if ( need_retry(GET_ERRNO) )
       { if ( PL_handle_signals() < 0 )
 	{ errno = EPLEXCEPTION;
 	  return -1;
@@ -2202,19 +2284,17 @@ nbio_close_input(nbio_sock_t socket)
     if ( (sock=s->socket) < 0 )
     { s->error = WSAECONNRESET;
       rc = -1;
-    } else if ( shutdown(sock, SD_RECEIVE) == SOCKET_ERROR )
-    { DEBUG(1, Sdprintf("shutdown(%d=%d, SD_RECEIVE) failed: %s\n",
-			socket, s->socket,
-			WinSockError(WSAGetLastError())));
-      Sseterr(s->input, SIO_FERR, WinSockError(WSAGetLastError()));
-      rc = -1;
     }
   }
 #endif
 
   s->input = NULL;
   if ( !(s->flags & (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
+#ifdef __WINDOWS__
+    return socketIsPendingClose(s);
+#else
     return freeSocket(s);
+#endif
 
   return rc;
 }
@@ -2241,20 +2321,6 @@ nbio_close_output(nbio_sock_t socket)
     if ( (sock=s->socket) < 0 )
     { s->error = WSAECONNRESET;
       rc = -1;
-    } else if ( shutdown(sock, SD_SEND) == SOCKET_ERROR )
-    { const char *msg;
-#ifdef __WINDOWS__
-      msg = WinSockError(WSAGetLastError());
-#else
-      msg = strerror(errno);
-#endif
-      Sseterr(s->output, SIO_FERR, msg);
-#ifdef O_DEBUG
-      if ( debugging )
-	Sdprintf("shutdown(%d=%d, SD_SEND) failed: %s\n",
-		 socket, s->socket, msg);
-#endif
-      rc = -1;
     }
 #endif
   }
@@ -2262,7 +2328,11 @@ nbio_close_output(nbio_sock_t socket)
   DEBUG(3, Sdprintf("%d->flags = 0x%x\n", socket, s->flags));
   s->output = NULL;
   if ( !(s->flags & (PLSOCK_INSTREAM|PLSOCK_OUTSTREAM)) )
+#ifdef __WINDOWS__
+    return socketIsPendingClose(s);
+#else
     return freeSocket(s);
+#endif
 
   return rc;
 }
@@ -2282,34 +2352,37 @@ nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
     return -1;
 
 #ifdef __WINDOWS__
-
   DEBUG(3, Sdprintf("[%d] recvfrom from socket %d\n",
 		    PL_thread_self(), socket));
 
-  n = recvfrom(s->socket, buf, (int)bufSize, flags, from, fromlen);
-  if ( n < 0 )
-  { int wsaerrno;
-
-    if ( (wsaerrno=WSAGetLastError()) == WSAEWOULDBLOCK )
-    { s->rdata.recvfrom.buffer  = buf;
-      s->rdata.recvfrom.size    = bufSize;
-      s->rdata.recvfrom.flags   = flags;
-      s->rdata.recvfrom.from    = from;
-      s->rdata.recvfrom.fromlen = fromlen;
-      placeRequest(s, REQ_RECVFROM);
-      if ( !waitRequest(s) )
-      { errno = EPLEXCEPTION;
-	return -1;
-      }
-      n = s->rdata.recvfrom.bytes;
-    }
+  { int iflen = (int)*fromlen;		/* Windows recvfrom uses int */
 
+    n = recvfrom(s->socket, buf, (int)bufSize, flags, from, &iflen);
     if ( n < 0 )
-    { s->error = wsaerrno;
-      errno = EIO;
+    { int wsaerrno;
+
+      if ( (wsaerrno=WSAGetLastError()) == WSAEWOULDBLOCK )
+      { s->rdata.recvfrom.buffer  = buf;
+	s->rdata.recvfrom.size    = bufSize;
+	s->rdata.recvfrom.flags   = flags;
+	s->rdata.recvfrom.from    = from;
+	s->rdata.recvfrom.fromlen = fromlen;
+	placeRequest(s, REQ_RECVFROM);
+	if ( !waitRequest(s) )
+	{ errno = EPLEXCEPTION;
+	  return -1;
+	}
+	n = s->rdata.recvfrom.bytes;
+      }
+
+      if ( n < 0 )
+      { s->error = wsaerrno;
+	errno = EIO;
+      }
+    } else
+    { *fromlen = iflen;
+      s->error = 0;
     }
-  } else
-  { s->error = 0;
   }
 
 #else /*__WINDOWS__*/
@@ -2322,7 +2395,7 @@ nbio_recvfrom(int socket, void *buf, size_t bufSize, int flags,
 
     n = recvfrom(s->socket, buf, bufSize, flags, from, fromlen);
 
-    if ( n == -1 && need_retry(errno) )
+    if ( n == -1 && need_retry(GET_ERRNO) )
     { if ( PL_handle_signals() < 0 )
       { errno = EPLEXCEPTION;
 	return -1;
@@ -2358,7 +2431,7 @@ nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
   DEBUG(3, Sdprintf("[%d] sending %d bytes using socket %d\n",
 		    PL_thread_self(), (int)bufSize, socket));
 
-  n = sendto(s->socket, buf, (int)bufSize, flags, to, tolen);
+  n = sendto(s->socket, buf, (int)bufSize, flags, to, (int)tolen);
   if ( n < 0 )
   { int error = WSAGetLastError();
 
@@ -2376,11 +2449,11 @@ nbio_sendto(nbio_sock_t socket, void *buf, size_t bufSize, int flags,
 
 wouldblock:
   s->rdata.sendto.buffer  = buf;
-  s->rdata.sendto.size    = bufSize;
+  s->rdata.sendto.size    = (int)bufSize;
   s->rdata.sendto.bytes   = 0;
   s->rdata.sendto.flags   = flags;
   s->rdata.sendto.to      = to;
-  s->rdata.sendto.tolen   = tolen;
+  s->rdata.sendto.tolen   = (int)tolen;
   placeRequest(s, REQ_SENDTO);
   if ( !waitRequest(s) )
   { errno = EPLEXCEPTION;		/* handled Prolog signal */
