@@ -74,7 +74,7 @@ static functor_t FUNCTOR_socket1;	/* $socket(Id) */
 		 *	     CONVERSION		*
 		 *******************************/
 
-static int
+NBIO_EXPORT(int)
 tcp_get_socket(term_t Socket, int *id)
 { IOSTREAM *s;
   int socket;
@@ -82,7 +82,7 @@ tcp_get_socket(term_t Socket, int *id)
   if ( PL_is_functor(Socket, FUNCTOR_socket1) )
   { term_t a = PL_new_term_ref();
 
-    PL_get_arg(1, Socket, a);
+    _PL_get_arg(1, Socket, a);
     if ( PL_get_integer(a, id) )
       return TRUE;
   }
@@ -109,20 +109,43 @@ tcp_unify_socket(term_t Socket, int id)
 static foreign_t
 pl_host_to_address(term_t Host, term_t Ip)
 { struct in_addr ip;
-  struct hostent *host;
   char *host_name;
 
   if ( PL_get_atom_chars(Host, &host_name) )
-  { if ( (host = gethostbyname(host_name)) )
-    { if ( sizeof(ip) == host->h_length )
-      { memcpy(&ip, host->h_addr, host->h_length);
-	return nbio_unify_ip4(Ip, ntohl(ip.s_addr));
-      } else
-	return PL_warning("tcp_host_to_address/2: length mismatch in address");
+  { struct addrinfo hints;
+    struct addrinfo *res;
+
+    memset(&hints, 0, sizeof(hints));
+    hints.ai_family = AF_INET;
+
+    if ( getaddrinfo(host_name, NULL, &hints, &res) == 0 )
+    { int rc;
+
+      switch( res->ai_family )
+      { case AF_INET:
+	{ struct sockaddr_in *addr = (struct sockaddr_in*)res->ai_addr;
+
+	  rc = nbio_unify_ip4(Ip, ntohl(addr->sin_addr.s_addr));
+	  break;
+	}
+	case AF_INET6:
+	{ rc = PL_warning("tcp_host_to_address/2: IPv6 address not supported");
+	  break;
+	}
+	default:
+	  assert(0);
+      }
+
+      freeaddrinfo(res);
+
+      return rc;
     } else
-      return nbio_error(h_errno, TCP_HERRNO);
+    { return nbio_error(h_errno, TCP_HERRNO);
+    }
   } else if ( nbio_get_ip(Ip, &ip) )
-  { if ( (host = gethostbyaddr((char *)&ip, sizeof(ip), AF_INET)) )
+  { struct hostent *host;
+
+    if ( (host = gethostbyaddr((char *)&ip, sizeof(ip), AF_INET)) )
       return PL_unify_atom_chars(Host, host->h_name);
     else
       return nbio_error(h_errno, TCP_HERRNO);
@@ -155,7 +178,7 @@ pl_setopt(term_t Socket, term_t opt)
       } else /*if ( arity == 1 )*/
       { term_t a = PL_new_term_ref();
 
-	PL_get_arg(1, opt, a);
+	_PL_get_arg(1, opt, a);
 	if ( !PL_get_bool(a, &enable) )
 	  return pl_error(NULL, 0, NULL, ERR_DOMAIN, a, "boolean");
       }
@@ -246,7 +269,7 @@ udp_receive(term_t Socket, term_t Data, term_t From, term_t options)
       int arity;
 
       if ( PL_get_name_arity(head, &name, &arity) && arity == 1 )
-      { PL_get_arg(1, head, arg);
+      { _PL_get_arg(1, head, arg);
 
 	if ( name == ATOM_as )
 	{ atom_t a;
@@ -386,7 +409,7 @@ pl_bind(term_t Socket, term_t Address)
 
     if ( getsockname(fd, (struct sockaddr *) &addr, &len) )
       return nbio_error(errno, TCP_ERRNO);
-    PL_unify_integer(Address, ntohs(addr.sin_port));
+    return PL_unify_integer(Address, ntohs(addr.sin_port));
   }
 
   return TRUE;
@@ -413,7 +436,6 @@ pl_accept(term_t Master, term_t Slave, term_t Peer)
 }
 
 
-
 static foreign_t
 pl_gethostname(term_t name)
 { static atom_t hname;
@@ -422,12 +444,18 @@ pl_gethostname(term_t name)
   { char buf[256];
 
     if ( gethostname(buf, sizeof(buf)) == 0 )
-    { struct hostent *he;
+    { struct addrinfo *res;
+      struct addrinfo hints;
 
-      if ( (he = gethostbyname(buf)) )
-	hname = PL_new_atom(he->h_name);
+      memset(&hints, 0, sizeof(hints));
+      hints.ai_flags = AI_CANONNAME;
+
+      if ( getaddrinfo(buf, NULL, &hints, &res) == 0 )
+	hname = PL_new_atom(res->ai_canonname);
       else
 	hname = PL_new_atom(buf);
+
+      freeaddrinfo(res);
     } else
     { return nbio_error(h_errno, TCP_HERRNO);
     }
@@ -570,7 +598,7 @@ tcp_select(term_t Streams, term_t Available, term_t timeout)
 
   switch(ret)
   { case -1:
-      return pl_error("tcp_select", 3, ERR_ERRNO, errno, "select", "streams", Streams);
+      return pl_error("tcp_select", 3, NULL, ERR_ERRNO, errno, "select", "streams", Streams);
 
     case 0: /* Timeout */
       break;
@@ -595,10 +623,12 @@ static foreign_t
 pl_debug(term_t val)
 { int dbg;
 
-  PL_get_integer(val, &dbg);
-  nbio_debug(dbg);
+  if ( PL_get_integer(val, &dbg) )
+  { nbio_debug(dbg);
+    return TRUE;
+  }
 
-  return TRUE;
+  return FALSE;
 }
 #endif
 
@@ -619,9 +649,6 @@ install_socket()
 
   FUNCTOR_socket1 = PL_new_functor(PL_new_atom("$socket"), 1);
 
-#ifdef O_DEBUG
-  PL_register_foreign_in_module("user", "tcp_debug", 1, pl_debug, 0);
-#endif
   PL_register_foreign("tcp_accept",           3, pl_accept,           0);
   PL_register_foreign("tcp_bind",             2, pl_bind,             0);
   PL_register_foreign("tcp_connect",          2, pl_connect,	      0);
@@ -637,6 +664,10 @@ install_socket()
   PL_register_foreign("udp_socket",           1, udp_socket,          0);
   PL_register_foreign("udp_receive",	      4, udp_receive,	      0);
   PL_register_foreign("udp_send",	      4, udp_send,	      0);
+
+#ifdef O_DEBUG
+  PL_register_foreign("tcp_debug",	      1, pl_debug,	      0);
+#endif
 }
 
 
