@@ -30,6 +30,7 @@
 */
 
 #define O_DEBUG 1			/* provides time:time_debug(+Level) */
+//#define O_SAFE 1			/* extra safety checks */
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -48,32 +49,35 @@
 #include <errno.h>
 #include <assert.h>
 
-#ifdef _REENTRANT
-#include <pthread.h>
+#ifdef SIGUSR2
+#define SIG_TIME SIGUSR2
+#else
+#define SIG_TIME SIGALRM
 #endif
 
+#ifdef O_SAFE
+#define __USE_GNU
+#endif
+
+#include <pthread.h>
+
+typedef enum
+{ TIME_ABS,
+  TIME_REL
+} time_abs_rel;
+
 #ifdef __WINDOWS__
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-The __WINDOWS__ port uses the multimedia timers.   This  module must be linked
-with winmm.lib
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-#include <windows.h>
 #include <sys/timeb.h>
-#if (_MSC_VER < 1400)
-typedef DWORD DWORD_PTR;
-#endif
+#include <malloc.h>
 
 #ifndef SIGALRM
 #define SIGALRM 14
 #endif
 
-/* Seems to be there
 struct timeval
-{ long tv_usec;
-  long tv_sec;
+{ long tv_sec;
+  long tv_usec;
 };
-*/
 
 struct timezone
 { int zone;
@@ -93,10 +97,6 @@ gettimeofday(struct timeval *tv, struct timezone *tz)
 
 #else /*__WINDOWS__*/
 
-#ifdef _REENTRANT
-#define SHARED_TABLE 1
-#endif
-
 #include <time.h>
 #include <sys/time.h>
 
@@ -157,41 +157,6 @@ print_trace (void)
 #define DEBUG(n, g) ((void)0)
 #endif /*O_DEBUG*/
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-This module defines support for timing during execution. Most of this is
-highly system dependent.
-
-Design
-======
-
-The module contains three implementations:
-
-	* Windows
-	The Windows versions is based on multimedia timer objects.
-
-	* Unix setitimer (single threaded)
-	This implementation uses setitimer() and SIGALRM.  Whenever
-	something is changed, re_schedule() is called to set the
-	alarm clock for the next wakeup.
-
-	* Unix scheduler thread
-	This implementation uses a table shared between all threads and
-	a thread that waits for the next signal to be send using
-	pthread_cond_timedwait().  The signal SIGALRM is then delivered
-	using pthread_kill() to the scheduled thread.
-
-This module keeps a double-linked  list   of  `scheduled events' that is
-tagged with and annotated using  the   absolute  time  it should happen.
-These  times  are  represented  using   the  struct  timeval,  providing
-microsecond   resolution.   If   the     SHARED_TABLE    version   (Unix
-multithreaded), there is one table for all  events. In the other designs
-each thread has its own table.
-
-The  signal  handler  uses  the  PL_SIGSYNC    option  to  be  scheduled
-synchronous with the Prolog activity  and   eb  able to throw exceptions
-(the most common alarm activity).
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
 static void	on_alarm(int sig);
 
 static module_t	   MODULE_user;
@@ -219,18 +184,9 @@ typedef struct event
   struct event  *previous;		/* idem */
   unsigned long  flags;			/* misc flags */
   long		 magic;			/* validate magic */
-  double	 time;			/* Specified time */
   struct timeval at;			/* Time to deliver */
-#ifdef SHARED_TABLE
   pthread_t	 thread_id;		/* Thread to call in */
-#ifdef O_DEBUG
   int		 pl_thread_id;		/* Prolog thread ID */
-#endif
-#endif
-#ifdef __WINDOWS__
-  UINT		 mmid;			/* MultiMedia timer id */
-  DWORD		 tid;			/* thread-id of Prolog thread */
-#endif
 } event, *Event;
 
 typedef void (*handler_t)(int);
@@ -238,52 +194,20 @@ typedef void (*handler_t)(int);
 typedef struct
 { Event first;				/* first in list */
   Event scheduled;			/* The one we scheduled for */
+  int   stop;				/* stop alarm-loop */
 } schedule;
 
-#ifdef SHARED_TABLE
 static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
 static pthread_cond_t cond   = PTHREAD_COND_INITIALIZER;
 static int scheduler_running = FALSE;	/* is scheduler running? */
 static pthread_t scheduler;		/* thread id of scheduler */
 
-#define LOCK()   pthread_mutex_lock(&mutex);
-#define UNLOCK() pthread_mutex_unlock(&mutex);
-#else
-#define LOCK()   (void)0
-#define UNLOCK() (void)0
-#endif /*SHARED_TABLE*/
+#define LOCK()   pthread_mutex_lock(&mutex)
+#define UNLOCK() pthread_mutex_unlock(&mutex)
 
-#if defined(_REENTRANT) && !defined(SHARED_TABLE)
-
-static pthread_key_t key;
-
-static schedule *
-TheSchedule()
-{ schedule *s;
-
-  if ( !(s=pthread_getspecific(key)) )
-  { s = PL_malloc(sizeof(schedule));
-    memset(s, 0, sizeof(*s));
-    pthread_setspecific(key, s);
-  }
-
-  return s;
-}
-
-static void
-free_schedule(void *closure)
-{ schedule *s = closure;
-
-  PL_free(s);
-}
-
-#else /*defined(_REENTRANT) && !defined(SHARED_TABLE)*/
-
-static schedule the_schedule;		/* the schedule */
+static schedule the_schedule = {0};	/* the schedule */
 #define TheSchedule() (&the_schedule)	/* current schedule */
 
-#endif  /*defined(_REENTRANT) && !defined(SHARED_TABLE)*/
-
 int signal_function_set = FALSE;	/* signal function is set */
 static handler_t signal_function;	/* Current signal function */
 
@@ -309,6 +233,18 @@ allocEvent()
 }
 
 
+static void
+setTimeEventAbs(Event ev, double t)
+{ struct timeval tv;
+
+  gettimeofday(&tv, NULL);
+  tv.tv_usec = (long)((t-floor(t))*1000000);
+  tv.tv_sec  = (long)t;
+
+  ev->at = tv;
+}
+
+
 static void
 setTimeEvent(Event ev, double t)
 { struct timeval tv;
@@ -322,7 +258,6 @@ setTimeEvent(Event ev, double t)
   }
 
   ev->at = tv;
-  ev->time = t;
 }
 
 
@@ -389,6 +324,8 @@ unlinkEvent(Event ev)
 
   if ( ev->next )
     ev->next->previous = ev->previous;
+
+  ev->next = ev->previous = NULL;	/* in case it's reused */
 }
 
 
@@ -405,36 +342,11 @@ freeEvent(Event ev)
 }
 
 
-#ifndef SHARED_TABLE
-static void
-callEvent(Event ev)
-{ term_t goal = PL_new_term_ref();
-
-  ev->flags |= EV_DONE;
-
-  PL_recorded(ev->goal, goal);
-  PL_call_predicate(ev->module,
-		    PL_Q_PASS_EXCEPTION,
-		    PREDICATE_call1,
-		    goal);
-}
-#endif
-
-
 static void
 cleanupHandler()
-{
-#ifndef __WINDOWS__
-  struct itimerval v;
-
-  DEBUG(1, Sdprintf("Removed timer\n"));
-  memset(&v, 0, sizeof(v));
-  setitimer(ITIMER_REAL, &v, NULL);	/* restore? */
-#endif
-
-  if ( signal_function_set )
+{ if ( signal_function_set )
   { signal_function_set = FALSE;
-    PL_signal(SIGALRM, signal_function);
+    PL_signal(SIG_TIME, signal_function);
   }
 }
 
@@ -442,130 +354,33 @@ cleanupHandler()
 static void
 installHandler()
 { if ( !signal_function_set )
-  { signal_function = PL_signal(SIGALRM|PL_SIGSYNC, on_alarm);
+  { signal_function = PL_signal(SIG_TIME|PL_SIGSYNC, on_alarm);
     signal_function_set = TRUE;
   }
 }
 
 
 static void
-cleanup()
-{ Event ev, next;
+cleanup(int rc, void *arg)
+{ Event ev;
   schedule *sched = TheSchedule();
 
-  for(ev=sched->first; ev; ev = next)
-  { next = ev->next;
-    removeEvent(ev);
+  while( (ev=sched->first) )
+  { removeEvent(ev);
   }
 
   cleanupHandler();
-}
 
+  if ( scheduler_running )
+  { sched->stop = TRUE;
+    pthread_cond_signal(&cond);
 
-#ifdef __WINDOWS__
-
-static void
-on_alarm(int sig)
-{ Event ev, next;
-
-  for(ev=TheSchedule()->first; ev; ev = next)
-  { assert(ev->magic == EV_MAGIC);
-
-    if ( ev->flags & EV_FIRED )
-    { ev->flags &= ~EV_FIRED;
-
-      callEvent(ev);
-      ev->mmid = 0;			/* TIME_ONESHOT */
-
-      next = ev->next;
-      if ( ev->flags & EV_REMOVE )
-	freeEvent(ev);
-    } else
-      next = ev->next;
+    pthread_join(scheduler, NULL);
+    scheduler_running = FALSE;
   }
 }
 
 
-static void CALLBACK
-callTimer(UINT id, UINT msg, DWORD_PTR dwuser, DWORD_PTR dw1, DWORD_PTR dw2)
-{ Event ev = (Event)dwuser;
-
-  ev->flags |= EV_FIRED;
-  PL_w32thread_raise(ev->tid, SIGALRM);
-}
-
-
-static int
-installEvent(Event ev)
-{ MMRESULT rval;
-  int rc;
-
-  LOCK();
-  rc = insertEvent(ev);
-  UNLOCK();
-  if ( rc != TRUE )
-    return rc;
-
-  rval = timeSetEvent((int)(ev->time*1000),
-		      50,			/* resolution (milliseconds) */
-		      callTimer,
-		      (DWORD_PTR)ev,
-		      TIME_ONESHOT);
-
-  if ( rval )
-  { ev->tid = GetCurrentThreadId();
-    ev->mmid = rval;
-
-    return TRUE;
-  }
-
-  return ERR_RESOURCE;
-}
-
-
-static int
-uninstallEvent(Event ev)
-{ LOCK();
-
-  if ( TheSchedule()->scheduled == ev )
-    ev->flags = EV_DONE;
-
-  if ( ev->mmid )
-  { timeKillEvent(ev->mmid);
-    ev->mmid = 0;
-  }
-
-  ev->flags &= ~(EV_DONE|EV_FIRED);
-
-  UNLOCK();
-
-  return TRUE;
-}
-
-
-static int
-removeEvent(Event ev)
-{ LOCK();
-
-  if ( TheSchedule()->scheduled == ev )
-    ev->flags |= EV_DONE;
-
-  if ( ev->mmid )
-  { timeKillEvent(ev->mmid);
-    ev->mmid = 0;
-  }
-
-  freeEvent(ev);
-  UNLOCK();
-
-  return TRUE;
-}
-
-
-#else /*__WINDOWS__*/
-
-#ifdef SHARED_TABLE
-
 static Event
 nextEvent(schedule *sched)
 { Event ev;
@@ -581,44 +396,137 @@ nextEvent(schedule *sched)
 }
 
 
+typedef struct
+{ int		*bits;
+  size_t	size;
+  size_t	high;
+} bitvector;
+
+#define BITSPERINT (8*sizeof(int))
+
+static int
+set_bit(bitvector *v, size_t bit)
+{ size_t offset = bit/BITSPERINT;
+  int bi = bit%BITSPERINT;
+
+  while ( offset >= v->size )
+  { size_t osize = v->size * sizeof(int);
+    int *newbits = realloc(v->bits, osize*2);
+
+    if ( !newbits )
+      return FALSE;
+    memset((char*)newbits+osize, 0, osize);
+    v->bits  = newbits;
+    v->size *= 2;
+  }
+
+  while ( bit > v->high )		/* TBD: zero entire ints */
+  { size_t ho = v->high/BITSPERINT;
+    int    b  = v->high%BITSPERINT;
+
+    v->bits[ho] &= ~(1<<(b-1));
+    v->high++;
+  }
+
+  v->bits[offset] |= 1<<(bi-1);
+  return TRUE;
+}
+
+
+static int
+is_set(bitvector *v, size_t bit)
+{ if ( bit <= v->high )
+  { size_t offset = bit/BITSPERINT;
+    int bi = bit%BITSPERINT;
+
+    return (v->bits[offset] & (1<<(bi-1))) != 0;
+  }
+
+  return FALSE;
+}
+
+
 static void *
 alarm_loop(void * closure)
 { schedule *sched = TheSchedule();
+  bitvector signalled;
+
+  signalled.size = 4;
+  signalled.bits = malloc(signalled.size*sizeof(int));
+  signalled.high = 0;
 
   pthread_mutex_lock(&mutex);		/* for condition variable */
+  DEBUG(1, Sdprintf("Iterating alarm_loop()\n"));
 
-  for(;;)
+  while( !sched->stop )
   { Event ev = nextEvent(sched);
+    struct timeval now;
+
+    signalled.high = 0;
+    gettimeofday(&now, NULL);
+
+    for(; ev; ev = ev->next)
+    { struct timeval left;
+
+      left.tv_sec  = ev->at.tv_sec  - now.tv_sec;
+      left.tv_usec = ev->at.tv_usec - now.tv_usec;
+      if ( left.tv_usec < 0 )
+      { left.tv_sec--;
+	left.tv_usec += 1000000;
+      }
+
+      if ( left.tv_sec < 0 ||
+	   (left.tv_sec == 0 && left.tv_usec == 0) )
+      { if ( !is_set(&signalled, ev->pl_thread_id) )
+	{ DEBUG(1, Sdprintf("Signalling (left = %ld) %d ...\n",
+			    (long)left.tv_sec,
+			    ev->pl_thread_id));
+	  set_bit(&signalled, ev->pl_thread_id);
+#ifdef __WINDOWS__
+	  PL_thread_raise(ev->pl_thread_id, SIG_TIME);
+#else
+	  pthread_kill(ev->thread_id, SIG_TIME);
+#endif
+	}
+      } else
+	break;
+    }
 
     if ( ev )
-    { struct timespec timeout;
-      int rc;
-
+    { int rc;
+      struct timespec timeout;
       timeout.tv_sec  = ev->at.tv_sec;
       timeout.tv_nsec = ev->at.tv_usec*1000;
 
+    retry_timed_wait:
       DEBUG(1, Sdprintf("Waiting ...\n"));
       rc = pthread_cond_timedwait(&cond, &mutex, &timeout);
 
       switch( rc )
       { case ETIMEDOUT:
-	  DEBUG(1, Sdprintf("Signalling %d (= %d) ...\n",
-			    ev->pl_thread_id, ev->thread_id));
-	  sched->scheduled = ev;
-	  ev->flags |= EV_FIRED;
-	  pthread_kill(ev->thread_id, SIGALRM);
-	  break;
 	case 0:
+	  continue;
 	case EINTR:
+	  goto retry_timed_wait;
+	default:
+	  Sdprintf("alarm/4: pthread_cond_timedwait(): %s\n", strerror(rc));
+	  assert(0);
+      }
+    } else
+    { int rc;
+
+    retry_wait:
+      DEBUG(1, Sdprintf("No waiting events\n"));
+      rc = pthread_cond_wait(&cond, &mutex);
+      switch(rc)
+      { case EINTR:
+	  goto retry_wait;
+	case 0:
 	  continue;
 	default:
 	  Sdprintf("alarm/4: pthread_cond_timedwait(): %s\n", strerror(rc));
+	  assert(0);
       }
-    } else
-    { int rc = pthread_cond_wait(&cond, &mutex);
-
-      if ( rc == EINTR )
-	continue;
     }
   }
 
@@ -631,42 +539,63 @@ on_alarm(int sig)
 { Event ev;
   schedule *sched = TheSchedule();
   pthread_t self = pthread_self();
-  term_t goal = 0;
-  module_t module = NULL;
 
-  DEBUG(1, Sdprintf("Signal received in %d (= %d)\n",
-		    PL_thread_self(), self));
+  DEBUG(1, Sdprintf("Signal received in %d\n",
+		    PL_thread_self()));
 #ifdef BACKTRACE
   DEBUG(10, print_trace());
 #endif
 
-  LOCK();
-  for(ev = sched->first; ev; ev=ev->next)
-  { assert(ev->magic == EV_MAGIC);
+  for(;;)
+  { struct timeval now;
+    term_t goal = 0;
+    module_t module = NULL;
 
-    if ( (ev->flags & EV_FIRED) &&
-	 pthread_equal(self, ev->thread_id) )
-    { ev->flags &= ~EV_FIRED;
+    gettimeofday(&now, NULL);
 
-      DEBUG(1, Sdprintf("Calling event\n"));
-      ev->flags |= EV_DONE;
-      module = ev->module;
-      goal = PL_new_term_ref();
-      PL_recorded(ev->goal, goal);
+    LOCK();
+    for(ev = sched->first; ev; ev=ev->next)
+    { struct timeval left;
 
-      if ( ev->flags & EV_REMOVE )
-	freeEvent(ev);
-      break;
+      assert(ev->magic == EV_MAGIC);
+
+      if ( (ev->flags & (EV_DONE|EV_FIRED)) ||
+	   !pthread_equal(self, ev->thread_id) )
+	continue;
+
+      left.tv_sec  = ev->at.tv_sec - now.tv_sec;
+      left.tv_usec = ev->at.tv_usec - now.tv_usec;
+      if ( left.tv_usec < 0 )
+      { left.tv_sec--;
+	left.tv_usec += 1000000;
+      }
+
+      if ( left.tv_sec < 0 ||
+	   (left.tv_sec == 0 && left.tv_usec == 0) )
+      { DEBUG(1, Sdprintf("Calling event\n"));
+	ev->flags |= EV_DONE;
+	module = ev->module;
+	goal = PL_new_term_ref();
+	PL_recorded(ev->goal, goal);
+
+	if ( ev->flags & EV_REMOVE )
+	  freeEvent(ev);
+	break;
+      }
     }
-  }
-  UNLOCK();
+    UNLOCK();
 
-  if ( goal )
-  { PL_call_predicate(module,
-		      PL_Q_PASS_EXCEPTION,
-		      PREDICATE_call1,
-		      goal);
+    if ( goal )
+    { PL_call_predicate(module,
+			PL_Q_PASS_EXCEPTION,
+			PREDICATE_call1,
+			goal);
+    } else
+      break;
   }
+
+  DEBUG(1, Sdprintf("Processed pending events; signalling scheduler\n"));
+  pthread_cond_signal(&cond);
 }
 
 
@@ -674,35 +603,36 @@ static int
 installEvent(Event ev)
 { int rc;
 
-  LOCK();
-
   ev->thread_id = pthread_self();
-#ifdef O_DEBUG
   ev->pl_thread_id = PL_thread_self();
-#endif
 
+  LOCK();
   if ( !scheduler_running )
   { pthread_attr_t attr;
-    int rc;
 
+    TheSchedule()->stop = FALSE;
     pthread_attr_init(&attr);
     pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
-    pthread_attr_setstacksize(&attr, 1024);
+    pthread_attr_setstacksize(&attr, 8192);
+    rc = pthread_create(&scheduler, &attr, alarm_loop, NULL);
+    pthread_attr_destroy(&attr);
 
-    if ( (rc=pthread_create(&scheduler, &attr, alarm_loop, NULL)) )
+    if ( rc != 0 )
+    { UNLOCK();
       return pl_error("alarm", 4, "Failed to start schedule thread",
 		      ERR_ERRNO, rc);
-    pthread_attr_destroy(&attr);
+    }
 
     DEBUG(1, Sdprintf("Started scheduler thread\n"));
     scheduler_running = TRUE;
   }
 
-  if ( (rc=insertEvent(ev)) == TRUE )
-    pthread_cond_signal(&cond);
-
+  rc = insertEvent(ev);
   UNLOCK();
 
+  if ( rc )
+    pthread_cond_signal(&cond);
+
   return rc;
 }
 
@@ -713,10 +643,10 @@ uninstallEvent(Event ev)
   if ( TheSchedule()->scheduled == ev )
     ev->flags |= EV_DONE;
   unlinkEvent(ev);
-  pthread_cond_signal(&cond);
   ev->flags &= ~(EV_FIRED|EV_DONE);
   UNLOCK();
 
+  pthread_cond_signal(&cond);
 
   return TRUE;
 }
@@ -728,131 +658,14 @@ removeEvent(Event ev)
   if ( TheSchedule()->scheduled == ev )
     ev->flags |= EV_DONE;
   freeEvent(ev);
-  pthread_cond_signal(&cond);
   UNLOCK();
 
-  return TRUE;
-}
-
-
-#else /*SHARED_TABLE*/
-
-static void
-re_schedule()
-{ struct itimerval v;
-  Event ev;
-  schedule *sched = TheSchedule();
-
-  for(ev=sched->first; ev; ev = ev->next)
-  { struct timeval now;
-    struct timeval left;
-
-    if ( ev->flags & EV_DONE )
-      continue;
-
-    gettimeofday(&now, NULL);
-    left.tv_sec  = ev->at.tv_sec  - now.tv_sec;
-    left.tv_usec = ev->at.tv_usec - now.tv_usec;
-    if ( left.tv_usec < 0 )
-    { left.tv_sec--;
-      left.tv_usec += 1000000;
-    }
-
-    if ( left.tv_sec < 0 ||
-	 (left.tv_sec == 0 && left.tv_usec == 0) )
-    { DEBUG(1, Sdprintf("Passed\n"));
-
-      callEvent(ev);		/* Time has passed.  What about exceptions? */
-
-      continue;
-    }
-
-    sched->scheduled = ev;	/* This is the scheduled one */
-    DEBUG(1, Sdprintf("Scheduled for %d.%06d\n", ev->at.tv_sec, ev->at.tv_usec));
-
-    v.it_value            = left;
-    v.it_interval.tv_sec  = 0;
-    v.it_interval.tv_usec = 0;
-
-    setitimer(ITIMER_REAL, &v, NULL);	/* Store old? */
-
-    return;
-  }
-}
-
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-This one is  asynchronously  called  from   the  hook  registered  using
-PL_signal(). Throwing an exception is normally   safe,  but some foreign
-code might not leave the  system   unstable  after resulting long_jmp().
-This should all nicely be protected,  but   most  likely  this isn't the
-case.
-
-The alternative is to delay the signal to a safe place using PL_raise();
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-static void
-on_alarm(int sig)
-{ Event ev;
-  schedule *sched = TheSchedule();
-
-#ifdef BACKTRACE
-  DEBUG(10, print_trace());
-#endif
-
-  if ( (ev=sched->scheduled) )
-  { assert(ev->magic == EV_MAGIC);
-    sched->scheduled = NULL;
-
-    callEvent(ev);
-
-    if ( ev->flags & EV_REMOVE )
-      freeEvent(ev);
-  }
-
-  re_schedule();
-}
-
-
-static int
-installEvent(Event ev)
-{ int rc;
-
-  if ( (rc=insertEvent(ev)) == TRUE )
-    re_schedule();
-
-  return rc;
-}
-
-
-static int
-uninstallEvent(Event ev)
-{ if ( TheSchedule()->scheduled == ev )
-  { ev->flags |= EV_DONE;
-    re_schedule();
-  }
-
-  ev->flags &= ~(EV_DONE|EV_FIRED);
+  pthread_cond_signal(&cond);
 
   return TRUE;
 }
 
 
-static int
-removeEvent(Event ev)
-{ if ( TheSchedule()->scheduled == ev )
-  { ev->flags |= EV_DONE;
-    re_schedule();
-  }
-
-  freeEvent(ev);
-
-  return TRUE;
-}
-
-#endif /*SHARED_TABLE*/
-#endif /*__WINDOWS__*/
-
-
 		 /*******************************
 		 *	PROLOG CONNECTION	*
 		 *******************************/
@@ -867,6 +680,7 @@ alarm_error(term_t alarm, int err)
 		      alarm, "install", "alarm");
     default:
       assert(0);
+      return FALSE;
   }
 }
 
@@ -888,7 +702,7 @@ get_timer(term_t t, Event *ev)
   { term_t a = PL_new_term_ref();
     void *p;
 
-    PL_get_arg(1, t, a);
+    _PL_get_arg(1, t, a);
     if ( PL_get_pointer(a, &p) )
     { Event e = p;
 
@@ -917,10 +731,10 @@ pl_get_bool_ex(term_t arg, int *val)
 
 
 static foreign_t
-alarm4(term_t time, term_t callable, term_t id, term_t options)
+alarm4_gen(time_abs_rel abs_rel, term_t time, term_t callable,
+	   term_t id, term_t options)
 { Event ev;
   double t;
-  struct timeval tv;
   module_t m = NULL;
   unsigned long flags = 0L;
 
@@ -936,7 +750,7 @@ alarm4(term_t time, term_t callable, term_t id, term_t options)
       { if ( arity == 1 )
 	{ term_t arg = PL_new_term_ref();
 
-	  PL_get_arg(1, head, arg);
+	  _PL_get_arg(1, head, arg);
 
 	  if ( name == ATOM_remove )
 	  { int t = FALSE;
@@ -964,17 +778,15 @@ alarm4(term_t time, term_t callable, term_t id, term_t options)
     return pl_error(NULL, 0, NULL, ERR_ARGTYPE, 1,
 		    time, "number");
 
-  gettimeofday(&tv, NULL);
-  tv.tv_usec += (long)((t-floor(t))*1000000);
-  tv.tv_sec  += (long)t;
-  if ( tv.tv_usec >= 1000000 )
-  { tv.tv_usec -= 1000000;
-    tv.tv_sec++;
-  }
 
   if ( !(ev = allocEvent()) )
     return FALSE;
-  setTimeEvent(ev, t);
+
+  if (abs_rel==TIME_REL)
+	  setTimeEvent(ev, t);
+  else
+	  setTimeEventAbs(ev,t);
+
   if ( !unify_timer(id, ev) )
   { freeEvent(ev);			/* not linked: no need to lock */
     return FALSE;
@@ -999,10 +811,24 @@ alarm4(term_t time, term_t callable, term_t id, term_t options)
 
 
 static foreign_t
-alarm3(term_t time, term_t callable, term_t id)
-{ return alarm4(time, callable, id, 0);
+alarm4_abs(term_t time, term_t callable, term_t id, term_t options)
+{ return alarm4_gen(TIME_ABS,time,callable,id,options);
 }
 
+static foreign_t
+alarm4_rel(term_t time, term_t callable, term_t id, term_t options)
+{ return alarm4_gen(TIME_REL,time,callable,id,options);
+}
+
+static foreign_t
+alarm3_abs(term_t time, term_t callable, term_t id)
+{ return alarm4_gen(TIME_ABS,time, callable, id, 0);
+}
+
+static foreign_t
+alarm3_rel(term_t time, term_t callable, term_t id)
+{ return alarm4_gen(TIME_REL,time, callable, id, 0);
+}
 
 static foreign_t
 install_alarm(term_t alarm)
@@ -1071,9 +897,7 @@ current_alarms(term_t time, term_t goal, term_t id, term_t status,
   term_t tail = PL_copy_term_ref(matching);
   term_t head = PL_new_term_ref();
   term_t av   = PL_new_term_refs(4);
-#ifdef SHARED_TABLE
   pthread_t self = pthread_self();
-#endif
 
   LOCK();
   ev = TheSchedule()->first;
@@ -1083,10 +907,8 @@ current_alarms(term_t time, term_t goal, term_t id, term_t status,
     double at;
     fid_t fid;
 
-#ifdef SHARED_TABLE
     if ( !pthread_equal(self, ev->thread_id) )
       continue;
-#endif
 
     fid = PL_open_foreign_frame();
 
@@ -1116,12 +938,16 @@ current_alarms(term_t time, term_t goal, term_t id, term_t status,
 
     PL_discard_foreign_frame(fid);
 
-    PL_put_float(av+0, at);		/* time */
-    PL_recorded(ev->goal, av+1);	/* goal */
-    PL_put_variable(av+2);		/* id */
-    unify_timer(av+2, ev);
-    PL_put_atom(av+3, s);		/* status */
-    PL_cons_functor_v(next, FUNCTOR_alarm4, av);
+    if ( !PL_put_float(av+0, at) ||		/* time */
+	 !PL_recorded(ev->goal, av+1) ||	/* goal */
+	 !PL_put_variable(av+2) ||		/* id */
+	 !unify_timer(av+2, ev) ||
+	 !PL_put_atom(av+3, s) ||		/* status */
+	 !PL_cons_functor_v(next, FUNCTOR_alarm4, av) )
+    { PL_close_foreign_frame(fid);
+      UNLOCK();
+      return FALSE;
+    }
 
     if ( PL_unify_list(tail, head, tail) &&
 	 PL_unify(head, next) )
@@ -1143,7 +969,7 @@ current_alarms(term_t time, term_t goal, term_t id, term_t status,
 
 
 install_t
-install()
+install_time()
 { MODULE_user	  = PL_new_module(PL_new_atom("user"));
 
   FUNCTOR_alarm1  = PL_new_functor(PL_new_atom("$alarm"), 1);
@@ -1158,8 +984,10 @@ install()
 
   PREDICATE_call1 = PL_predicate("call", 1, "user");
 
-  PL_register_foreign("alarm",          4, alarm4,         PL_FA_TRANSPARENT);
-  PL_register_foreign("alarm",          3, alarm3,         PL_FA_TRANSPARENT);
+  PL_register_foreign("alarm_at",       4, alarm4_abs,     PL_FA_TRANSPARENT);
+  PL_register_foreign("alarm",          4, alarm4_rel,     PL_FA_TRANSPARENT);
+  PL_register_foreign("alarm_at",       3, alarm3_abs,     PL_FA_TRANSPARENT);
+  PL_register_foreign("alarm",          3, alarm3_rel,     PL_FA_TRANSPARENT);
   PL_register_foreign("remove_alarm",   1, remove_alarm,   0);
   PL_register_foreign("uninstall_alarm",1, uninstall_alarm,0);
   PL_register_foreign("install_alarm",  1, install_alarm,  0);
@@ -1170,18 +998,12 @@ install()
   PL_register_foreign("time_debug",	1, pl_time_debug,  0);
 #endif
 
-#if defined(_REENTRANT) && !defined(SHARED_TABLE)
-  pthread_key_create(&key, free_schedule);
-#endif
   installHandler();
+  PL_on_halt(cleanup, NULL);
 }
 
 
 install_t
-uninstall()
-{ cleanup();
-
-#if defined(_REENTRANT) && !defined(SHARED_TABLE)
-  pthread_key_delete(key);
-#endif
+uninstall_time()
+{ cleanup(0, NULL);
 }
