@@ -1,11 +1,12 @@
-/*  $Id: atom_map.c,v 1.36 2007/01/26 11:36:35 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 2006, University of Amsterdam
+    Copyright (C): 2006-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
@@ -31,6 +32,7 @@
 #include "avl.h"
 #include "lock.h"
 #include "atom.h"
+#include "murmur.h"
 #include "debug.h"
 #include <string.h>
 #include <assert.h>
@@ -67,7 +69,7 @@ Searching is done using
 #define AM_MAGIC	0x6ab19e8e
 
 typedef struct atom_map
-{ long		magic;			/* AM_MAGIC */
+{ int		magic;			/* AM_MAGIC */
   size_t	value_count;		/* total # values */
   rwlock	lock;			/* Multi-threaded access */
   avl_tree	tree;			/* AVL tree */
@@ -82,21 +84,30 @@ typedef struct atom_set
   size_t  allocated;			/* # cells allocated */
   datum *atoms;			/* allocated cells */
 #ifdef O_SECURE
-  long	  magic;
+  int	  magic;
 #endif
 } atom_set;
 
 
 #define ND_MAGIC 0x67b49a23
+#define ND_MAGIC_EX 0x753ab3c
 
 typedef struct node_data
 { datum		key;
   atom_set     *values;
 #ifdef O_SECURE
-  long		magic;
+  int		magic;
 #endif
 } node_data;
 
+typedef struct node_data_ex
+{ node_data	data;
+  atom_info	atom;
+#ifdef O_SECURE
+  int		magic;
+#endif
+} node_data_ex;
+
 
 #define RDLOCK(map)			rdlock(&map->lock)
 #define WRLOCK(map, allowreaders)	wrlock(&map->lock, allowreaders)
@@ -152,42 +163,51 @@ init_functors()
 
 static int
 type_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_type_error2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex = PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_type_error2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 domain_error(term_t actual, const char *expected)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_domain_error2,
-		        PL_CHARS, expected,
-		        PL_TERM, actual,
-		      PL_VARIABLE);
+  if ( (ex = PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_domain_error2,
+		         PL_CHARS, expected,
+		         PL_TERM, actual,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
 static int
 resource_error(const char *what)
-{ term_t ex = PL_new_term_ref();
+{ term_t ex;
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR, FUNCTOR_resource_error1,
-		        PL_CHARS, what,
-		      PL_VARIABLE);
+  if ( (ex = PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR, FUNCTOR_resource_error1,
+		         PL_CHARS, what,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
@@ -195,12 +215,15 @@ static int
 representation_error(const char *what)
 { term_t ex = PL_new_term_ref();
 
-  PL_unify_term(ex, PL_FUNCTOR, FUNCTOR_error2,
-		      PL_FUNCTOR_CHARS, "representation_error", 1,
-		        PL_CHARS, what,
-		      PL_VARIABLE);
+  if ( (ex = PL_new_term_ref()) &&
+       PL_unify_term(ex,
+		     PL_FUNCTOR, FUNCTOR_error2,
+		       PL_FUNCTOR_CHARS, "representation_error", 1,
+		         PL_CHARS, what,
+		       PL_VARIABLE) )
+    return PL_raise_exception(ex);
 
-  return PL_raise_exception(ex);
+  return FALSE;
 }
 
 
@@ -214,15 +237,6 @@ get_atom_ex(term_t t, atom_t *a)
 }
 
 
-static int
-get_long_ex(term_t t, long *v)
-{ if ( PL_get_long(t, v) )
-    return TRUE;
-
-  return type_error(t, "integer");
-}
-
-
 static int
 get_atom_map(term_t t, atom_map **map)
 { if ( PL_is_functor(t, FUNCTOR_atom_map1) )
@@ -262,12 +276,13 @@ bits
 
 #define ATOM_TAG_BITS 7
 #define ATOM_TAG 0x1
+#define EMPTY ((datum)ATOM_TAG)
 
-#define tag(d)		((long)(d)&0x1)
-#define isAtomDatum(d)  ((long)(d)&ATOM_TAG)
+#define tag(d)		((intptr_t)(d)&0x1)
+#define isAtomDatum(d)  ((intptr_t)(d)&ATOM_TAG)
 #define isIntDatum(d)	!isAtomDatum(d)
 
-#define MAP_MIN_INT	(-(long)(1L<<(sizeof(long)*8 - 1 - 1)))
+#define MAP_MIN_INT	(-(intptr_t)((intptr_t)1<<(sizeof(intptr_t)*8 - 1 - 1)))
 #define MAP_MAX_INT	(-MAP_MIN_INT - 1)
 
 static intptr_t atom_mask;
@@ -275,14 +290,14 @@ static intptr_t atom_mask;
 static void
 init_datum_store()
 { atom_t a = PL_new_atom("[]");
-  
+
   atom_mask = a & ((1<<(ATOM_TAG_BITS-1))-1);
 }
 
 
 static inline atom_t
 atom_from_datum(datum d)
-{ unsigned long v = (unsigned long)d;
+{ uintptr_t v = (uintptr_t)d;
   atom_t a;
 
   a  = ((v&~0x1)<<(ATOM_TAG_BITS-1))|atom_mask;
@@ -291,9 +306,9 @@ atom_from_datum(datum d)
 }
 
 
-static inline long
-long_from_datum(datum d)
-{ long v = (long)d;
+static inline intptr_t
+integer_from_datum(datum d)
+{ intptr_t v = (intptr_t)d;
 
   return (v>>1);
 }
@@ -311,7 +326,7 @@ atom_to_datum(atom_t a)
 
 
 static inline datum
-long_to_datum(long v)
+integer_to_datum(intptr_t v)
 { return (datum)(v<<1);
 }
 
@@ -319,16 +334,40 @@ long_to_datum(long v)
 static int
 get_datum(term_t t, datum* d)
 { atom_t a;
-  long l;
+  intptr_t l;
 
   if ( PL_get_atom(t, &a) )
   { *d = atom_to_datum(a);
     return TRUE;
-  } else if ( PL_get_long(t, &l) )
+  } else if ( PL_get_intptr(t, &l) )
   { if ( l < MAP_MIN_INT || l > MAP_MAX_INT )
       return representation_error("integer_range");
 
-    *d = long_to_datum(l);
+    *d = integer_to_datum(l);
+    return TRUE;
+  }
+
+  return type_error(t, "atom or integer");
+}
+
+
+static int
+get_search_datum(term_t t, node_data_ex *search)
+{ atom_t a;
+  intptr_t l;
+
+  SECURE(search->magic = ND_MAGIC_EX);
+
+  if ( PL_get_atom(t, &a) )
+  { search->data.key = atom_to_datum(a);
+    search->atom.handle   = a;
+    search->atom.resolved = FALSE;
+    return TRUE;
+  } else if ( PL_get_intptr(t, &l) )
+  { if ( l < MAP_MIN_INT || l > MAP_MAX_INT )
+      return representation_error("integer_range");
+
+    search->data.key = integer_to_datum(l);
     return TRUE;
   }
 
@@ -338,18 +377,18 @@ get_datum(term_t t, datum* d)
 
 static int
 unify_datum(term_t t, datum d)
-{ unsigned long v = (unsigned long)d;
+{ uintptr_t v = (uintptr_t)d;
 
   if ( isAtomDatum(v) )
     return PL_unify_atom(t, atom_from_datum(d));
   else
-    return PL_unify_integer(t, long_from_datum(d));
+    return PL_unify_integer(t, integer_from_datum(d));
 }
 
 
 static void
 lock_datum(datum d)
-{ unsigned long v = (unsigned long)d;
+{ uintptr_t v = (uintptr_t)d;
 
   if ( isAtomDatum(v) )
     PL_register_atom(atom_from_datum(d));
@@ -358,33 +397,12 @@ lock_datum(datum d)
 
 static void
 unlock_datum(datum d)
-{ unsigned long v = (unsigned long)d;
+{ if ( d != EMPTY )
+  { uintptr_t v = (uintptr_t)d;
 
-  if ( isAtomDatum(v) )
-    PL_unregister_atom(atom_from_datum(d));
-}
-
-
-static inline int
-cmp_datum(void *p1, void *p2)
-{ datum d1 = p1;
-  datum d2 = p2;
-  int d;
-
-  if ( (d=(tag(d1)-tag(d2))) == 0 )
-  { if ( isAtomDatum(d1) )
-    { return cmp_atoms(atom_from_datum(d1), atom_from_datum(d2));
-    } else
-    { long l1 = long_from_datum(d1);
-      long l2 = long_from_datum(d2);
-      
-      DEBUG(2, Sdprintf("cmp_datum(%ld, %ld)\n", l1, l2));
-
-      return l1 > l2 ? 1 : l1 < l2 ? -1 : 0;
-    }
+    if ( isAtomDatum(v) )
+      PL_unregister_atom(atom_from_datum(d));
   }
-
-  return d;
 }
 
 
@@ -397,7 +415,7 @@ format_datum(datum d, char *buf)
 
   if ( !buf )
     buf = tmp;
-  Ssprintf(buf, "%ld", long_from_datum(d));
+  Ssprintf(buf, "%lld", (int64_t)integer_from_datum(d));
 
   return buf;
 }
@@ -408,123 +426,154 @@ format_datum(datum d, char *buf)
 		 *	     ATOM SETS		*
 		 *******************************/
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-A set of atoms (literals) is a   sorted  array of atom-handles. They are
-sorted simply by handle as we are  not   interested  in the value in the
-actual atom.  Search is implemeted as binary search.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+static int insert_atom_set(atom_set *as, datum a);
 
 #define AS_INITIAL_SIZE 4
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+A set of  datums  (atoms  or  integers)   is  a  close  hash-table.  The
+implementation is an adapted copy from   XPCE's  class hash_table, using
+closed hash-tables.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static atom_set *
 new_atom_set(datum a0)
 { atom_set *as;
 
   if ( (as = malloc(sizeof(*as))) &&
        (as->atoms = malloc(sizeof(datum)*AS_INITIAL_SIZE)) )
-  { lock_datum(a0);
-    as->size = 1;
+  { size_t i;
+
+    as->size = 0;
     as->allocated = AS_INITIAL_SIZE;
-    as->atoms[0] = a0;
-    SECURE(as->magic = S_MAGIC);
+    for(i=0; i<AS_INITIAL_SIZE; i++)
+      as->atoms[i] = EMPTY;
+
+    insert_atom_set(as, a0);
+    lock_datum(a0);
   }
-  
+
   return as;
 }
 
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-find_in_atom_set(atom_set *as, datum  a)  returns   a  pointer  to  the
-location of the atom or, if the atom  isn't there, to the first location
-*after* the atom
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
-
-static datum *
-find_in_atom_set(atom_set *as, datum a, int *found)
-{ const datum *ap = (const datum *)as->atoms;
-  const datum *ep = &ap[as->size];
-
-  SECURE(assert(as->magic == S_MAGIC));
-
-  for(;;)
-  { const datum *cp = ap+(ep-ap)/2;
-    
-    if ( a < *cp )
-    { if ( ep == cp )
-      { *found = FALSE;
-	return (datum*)cp;
-      }
-      ep = cp;
-    } else if ( a > *cp )
-    { if ( ap == cp )
-      { cp++;
-	*found = FALSE;
-	return (datum*)cp;
-      }
-      ap = cp;
-    } else
-    { *found = TRUE;
-      return (datum*)cp;
-    }
-  }
+static unsigned int
+hash_datum(datum d)
+{ return rdf_murmer_hash(&d, sizeof(d), MURMUR_SEED);
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+in_atom_set(atom_set *as, datum a) returns TRUE if datum is in the set
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static int
 in_atom_set(atom_set *as, datum a)
-{ int found;
+{ unsigned int start = hash_datum(a) % as->allocated;
+  datum *d = &as->atoms[start];
+  datum *e = &as->atoms[as->allocated];
 
-  find_in_atom_set(as, a, &found);
-
-  return found;
+  for(;;)
+  { if ( *d == a )
+      return TRUE;
+    if ( *d == EMPTY )
+      return FALSE;
+    if ( ++d == e )
+      d = as->atoms;
+  }
 }
 
 
-#define ptr_diff(p1, p2) ((char *)(p1) - (char *)(p2))
+static int
+resize_atom_set(atom_set *as, size_t size)
+{ datum *old = as->atoms;
+  size_t oldsize = as->allocated;
+
+  if ( (as->atoms = malloc(sizeof(datum)*size)) )
+  { size_t i;
+
+    as->allocated = size;
+    as->size = 0;
+
+    for(i=0; i<size; i++)
+      as->atoms[i] = EMPTY;
+
+    for(i=0; i<oldsize; i++)
+    { if ( old[i] != EMPTY )
+	insert_atom_set(as, old[i]);
+    }
+
+    free(old);
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+
+/* Adds a datum to the set.  Returns 0 if nothing was added; 1 if the
+   datum was added or -1 if we are out of memory.
+*/
 
 static int
 insert_atom_set(atom_set *as, datum a)
-{ int found;
-  datum *ap = find_in_atom_set(as, a, &found);
-  
-  if ( !found )
-  { lock_datum(a);
+{ unsigned int start;
+  datum *d, *e;
 
-    if ( as->size == as->allocated )
-    { datum *na;
-      size_t newsize = as->allocated*2;
-
-      if ( !(na = realloc(as->atoms, sizeof(datum)*newsize)) )
-	return -1;
-      ap += na-as->atoms;
-      as->atoms = na;
-      as->allocated = newsize;
-    }
-    assert(as->size < as->allocated);
-
-    memmove(ap+1, ap, ptr_diff(&as->atoms[as->size], ap));
-    as->size++;
-    *ap = a;
-
-    return 1;
+  if ( 4*as->size + 5 > 3*as->allocated )
+  { if ( !resize_atom_set(as, 2*as->size) )
+      return -1;				/* no memory */
   }
 
-  return 0;
+  start = hash_datum(a) % as->allocated;
+  d = &as->atoms[start];
+  e = &as->atoms[as->allocated];
+
+  for(;;)
+  { if ( *d == a )
+      return 0;					/* nothing added */
+    if ( *d == EMPTY )
+    { as->size++;
+      *d = a;
+      return 1;
+    }
+    if ( ++d == e )
+      d = as->atoms;
+  }
 }
 
 
 static int
 delete_atom_set(atom_set *as, datum a)
-{ int found;
-  datum *ap = find_in_atom_set(as, a, &found);
+{ unsigned int i = hash_datum(a) % as->allocated;
+  int j, r;
 
-  if ( found )
-  { unlock_datum(a);
-    as->size--;
-    memmove(ap, ap+1, ptr_diff(&as->atoms[as->size], ap));
+  while(as->atoms[i] != EMPTY && as->atoms[i] != a)
+  { if ( ++i == as->allocated )
+      i = 0;
   }
+  if ( as->atoms[i] == EMPTY )
+    return FALSE;				/* not in table */
 
-  return found;
+  as->size--;
+  as->atoms[i] = EMPTY;				/* R1 */
+  j = i;
+
+  for(;;)
+  { if ( ++i == as->allocated )
+      i = 0;
+
+    if ( as->atoms[i] == EMPTY )
+      return TRUE;
+
+    r = hash_datum(as->atoms[i]) % as->allocated;
+    if ( (i >= r && r > j) || (r > j && j > i) || (j > i && i >= r) )
+      continue;
+
+    as->atoms[j] = as->atoms[i];
+    as->atoms[i] = EMPTY;
+    j = i;
+  }
 }
 
 
@@ -532,7 +581,7 @@ static void
 destroy_atom_set(atom_set *as)
 { size_t i;
 
-  for(i=0; i<as->size; i++)
+  for(i=0; i<as->allocated; i++)
     unlock_datum(as->atoms[i]);
 
   free(as->atoms);
@@ -560,10 +609,26 @@ free_node_data(void *ptr)
 
 static int
 cmp_node_data(void *l, void *r, NODE type)
-{ node_data *d1 = l;
-  node_data *d2 = r;
+{ node_data_ex *e1 = l;
+  node_data *n2 = r;
+  datum *d1 = e1->data.key;
+  datum *d2 = n2->key;
+  int d;
 
-  return cmp_datum(d1->key, d2->key);
+  SECURE(assert(e1->magic == ND_MAGIC_EX));
+
+  if ( (d=(tag(d1)-tag(d2))) == 0 )
+  { if ( isAtomDatum(d1) )
+    { return cmp_atom_info(&e1->atom, atom_from_datum(d2));
+    } else
+    { intptr_t l1 = integer_from_datum(d1);
+      intptr_t l2 = integer_from_datum(d2);
+
+      return l1 > l2 ? 1 : l1 < l2 ? -1 : 0;
+    }
+  }
+
+  return d;
 }
 
 
@@ -618,35 +683,44 @@ destroy_atom_map(term_t handle)
 
 
 static foreign_t
-insert_atom_map(term_t handle, term_t from, term_t to)
+insert_atom_map4(term_t handle, term_t from, term_t to, term_t keys)
 { atom_map *map;
   datum a2;
-  node_data search, *data;
-
+  node_data_ex search;
+  node_data *data;
 
   if ( !get_atom_map(handle, &map) ||
-       !get_datum(from, &search.key) ||
+       !get_search_datum(from, &search) ||
        !get_datum(to, &a2) )
     return FALSE;
-  
+
   if ( !WRLOCK(map, FALSE) )
     return FALSE;
 
-  search.values = NULL;
   if ( (data=avlfind(&map->tree, &search)) )
   { int rc;
 
     SECURE(assert(data->magic == ND_MAGIC));
 
     if ( (rc=insert_atom_set(data->values, a2)) < 0 )
+    { WRUNLOCK(map);
       return resource_error("memory");
+    }
 
     if ( rc )
+    { lock_datum(a2);
       map->value_count++;
+    }
   } else
-  { if ( !(search.values = new_atom_set(a2)) )
+  { if ( keys && !PL_unify_integer(keys, map->tree.count+1) )
+    { WRUNLOCK(map);
+      return FALSE;
+    }
+    if ( !(search.data.values = new_atom_set(a2)) )
+    { WRUNLOCK(map);
       return resource_error("memory");
-    lock_datum(search.key);
+    }
+    lock_datum(search.data.key);
     SECURE(search.magic = ND_MAGIC);
 
     data = avlins(&map->tree, &search);
@@ -660,6 +734,12 @@ insert_atom_map(term_t handle, term_t from, term_t to)
 }
 
 
+static foreign_t
+insert_atom_map3(term_t handle, term_t from, term_t to)
+{ return insert_atom_map4(handle, from, to, 0);
+}
+
+
 		 /*******************************
 		 *	       DELETE		*
 		 *******************************/
@@ -667,21 +747,22 @@ insert_atom_map(term_t handle, term_t from, term_t to)
 static foreign_t
 delete_atom_map2(term_t handle, term_t from)
 { atom_map *map;
-  node_data search;
+  node_data_ex search;
   node_data *data;
 
   if ( !get_atom_map(handle, &map) ||
-       !get_datum(from, &search.key) )
+       !get_search_datum(from, &search) )
     return FALSE;
-  
+
   if ( !WRLOCK(map, TRUE) )
     return FALSE;
 
-  search.values = NULL;
+					/* TBD: Single pass? */
   if ( (data = avlfind(&map->tree, &search)) )
   { LOCKOUT_READERS(map);
     map->value_count -= data->values->size;
-    avldel(&map->tree, data);
+    search.data = *data;
+    avldel(&map->tree, &search);
     REALLOW_READERS(map);
   }
 
@@ -694,14 +775,15 @@ delete_atom_map2(term_t handle, term_t from)
 static foreign_t
 delete_atom_map3(term_t handle, term_t from, term_t to)
 { atom_map *map;
-  node_data *data, search;
+  node_data_ex search;
+  node_data *data;
   datum a2;
 
   if ( !get_atom_map(handle, &map) ||
-       !get_datum(from, &search.key) ||
+       !get_search_datum(from, &search) ||
        !get_datum(to, &a2) )
     return FALSE;
-  
+
   if ( !WRLOCK(map, TRUE) )
     return FALSE;
 
@@ -711,9 +793,11 @@ delete_atom_map3(term_t handle, term_t from, term_t to)
 
     LOCKOUT_READERS(map);
     if ( delete_atom_set(as, a2) )
-    { map->value_count--;
+    { unlock_datum(a2);
+      map->value_count--;
       if ( as->size == 0 )
-      { avldel(&map->tree, data);
+      { search.data = *data;
+	avldel(&map->tree, &search);
       }
     }
     REALLOW_READERS(map);
@@ -768,19 +852,20 @@ find_atom_map(term_t handle, term_t keys, term_t literals)
     return FALSE;
 
   while(PL_get_list(tail, head, tail))
-  { node_data *data, search;
+  { node_data *data;
+    node_data_ex search;
     int neg = FALSE;
 
     if ( PL_is_functor(head, FUNCTOR_not1) )
-    { PL_get_arg(1, head, tmp);
-      if ( !get_datum(tmp, &search.key) )
+    { _PL_get_arg(1, head, tmp);
+      if ( !get_search_datum(tmp, &search) )
 	goto failure;
       neg = TRUE;
     } else
-    { if ( !get_datum(head, &search.key) )
+    { if ( !get_search_datum(head, &search) )
 	goto failure;
     }
-    
+
     if ( (data = avlfind(&map->tree, &search)) )
     { if ( ns+1 >= MAX_SETS )
 	return resource_error("max_search_atoms");
@@ -805,18 +890,21 @@ find_atom_map(term_t handle, term_t keys, term_t literals)
   { domain_error(keys, "keywords");
     goto failure;
   }
-    
+
   s0 = as[0].set;
 
   PL_put_term(tail, literals);
 
-  for(ca=0; ca<s0->size; ca++)
+  for(ca=0; ca<s0->allocated; ca++)
   { datum a = s0->atoms[ca];
     int i;
 
+    if ( a == EMPTY )
+      continue;
+
     for(i=1; i<ns; i++)
     { if ( !as[i].neg )
-      { if ( !in_atom_set(as[i].set, a) ) 
+      { if ( !in_atom_set(as[i].set, a) )
 	{ if ( a > as[i].set->atoms[as[i].set->size-1] )
 	    goto empty;
 	  goto next;
@@ -826,7 +914,7 @@ find_atom_map(term_t handle, term_t keys, term_t literals)
 	  goto next;
       }
     }
-    
+
     if ( !PL_unify_list(tail, head, tail) ||
 	 !unify_datum(head, a) )
       goto failure;
@@ -868,33 +956,35 @@ unify_keys(term_t head, term_t tail, AVLnode *node)
     { if ( !unify_keys(head, tail, node->subtree[LEFT]) )
 	return FALSE;
     }
-    
+
     data = (node_data*)node->data;
     if ( !PL_unify_list(tail, head, tail) ||
 	 !unify_datum(head, data->key) )
       return FALSE;
-  
+
     if ( node->subtree[RIGHT] )
       return unify_keys(head, tail, node->subtree[RIGHT]);
   }
-  
+
   return TRUE;
 }
 
 
 static int
-between_keys(atom_map *map, long min, long max, term_t head, term_t tail) 
+between_keys(atom_map *map, intptr_t min, intptr_t max, term_t head, term_t tail)
 { avl_enum state;
-  node_data *data, search;
+  node_data *data;
+  node_data_ex search;
 
   DEBUG(2, Sdprintf("between %ld .. %ld\n", min, max));
 
-  search.key = long_to_datum(min);
-  search.values = NULL;
+  search.data.key = integer_to_datum(min);
+  SECURE(search.magic = ND_MAGIC_EX);
+
   if ( (data = avlfindfirst(&map->tree, &search, &state)) &&
        isIntDatum(data->key) )
   { for(;;)
-    { if ( long_from_datum(data->key) > max )
+    { if ( integer_from_datum(data->key) > max )
 	break;
 
       if ( !PL_unify_list(tail, head, tail) ||
@@ -939,15 +1029,15 @@ rdf_keys_in_literal_map(term_t handle, term_t spec, term_t keys)
       goto failure;
   } else if ( name == ATOM_key && arity == 1 )
   { term_t a = PL_new_term_ref();
-    node_data *data, search;
+    node_data *data;
+    node_data_ex search;
 
-    PL_get_arg(1, spec, a);
-    if ( !get_datum(a, &search.key) )
+    _PL_get_arg(1, spec, a);
+    if ( !get_search_datum(a, &search) )
       goto failure;
 
-    search.values = NULL;
     if ( (data = avlfind(&map->tree, &search)) )
-    { long size = (long)data->values->size;
+    { intptr_t size = (intptr_t)data->values->size;
 
       RDUNLOCK(map);
       assert(size > 0);
@@ -959,16 +1049,19 @@ rdf_keys_in_literal_map(term_t handle, term_t spec, term_t keys)
   { term_t a = PL_new_term_ref();
     atom_t prefix, first_a;
     avl_enum state;
-    node_data *data, search;
+    node_data *data;
+    node_data_ex search;
     int match = (name == ATOM_prefix ? STR_MATCH_PREFIX : STR_MATCH_EXACT);
 
-    PL_get_arg(1, spec, a);
+    _PL_get_arg(1, spec, a);
     if ( !get_atom_ex(a, &prefix) )
       goto failure;
     first_a = first_atom(prefix, STR_MATCH_PREFIX);
 
-    search.key = atom_to_datum(first_a);
-    search.values = NULL;
+    search.data.key = atom_to_datum(first_a);
+    search.atom.handle = first_a;
+    search.atom.resolved = FALSE;
+    SECURE(search.magic = ND_MAGIC_EX);
 
     for(data = avlfindfirst(&map->tree, &search, &state);
 	data;
@@ -988,28 +1081,28 @@ rdf_keys_in_literal_map(term_t handle, term_t spec, term_t keys)
     avlfinddestroy(&state);
   } else if ( (name == ATOM_ge || name == ATOM_le) && arity == 1 )
   { term_t a = PL_new_term_ref();
-    long val, min, max;
+    intptr_t val, min, max;
 
-    PL_get_arg(1, spec, a);
-    if ( !get_long_ex(a, &val) )
+    _PL_get_arg(1, spec, a);
+    if ( !PL_get_intptr_ex(a, &val) )
       goto failure;
 
     if ( name == ATOM_ge )
       min = val, max = MAP_MAX_INT;
     else
       max = val, min = MAP_MIN_INT;
-    
+
     if ( !between_keys(map, min, max, head, tail) )
       goto failure;
   } else if ( name == ATOM_between && arity == 2 )
   { term_t a = PL_new_term_ref();
-    long min, max;
+    intptr_t min, max;
 
-    PL_get_arg(1, spec, a);
-    if ( !get_long_ex(a, &min) )
+    _PL_get_arg(1, spec, a);
+    if ( !PL_get_intptr_ex(a, &min) )
       goto failure;
-    PL_get_arg(2, spec, a);
-    if ( !get_long_ex(a, &max) )
+    _PL_get_arg(2, spec, a);
+    if ( !PL_get_intptr_ex(a, &max) )
       goto failure;
 
     if ( !between_keys(map, min, max, head, tail) )
@@ -1063,14 +1156,14 @@ rdf_statistics_literal_map(term_t map, term_t key)
 
   if ( !get_atom_map(map, &m) )
     return FALSE;
-  
+
   if ( PL_is_functor(key, FUNCTOR_size2) )
   { term_t a = PL_new_term_ref();
 
-    PL_get_arg(1, key, a);
+    _PL_get_arg(1, key, a);
     if ( !PL_unify_integer(a, m->tree.count) )
       return FALSE;
-    PL_get_arg(2, key, a);
+    _PL_get_arg(2, key, a);
 
     return PL_unify_integer(a, m->value_count);
   }
@@ -1095,7 +1188,8 @@ install_atom_map()
   PRED("rdf_new_literal_map",	     1,	new_atom_map,		    0);
   PRED("rdf_destroy_literal_map",    1,	destroy_atom_map,	    0);
   PRED("rdf_reset_literal_map",	     1, rdf_reset_literal_map,	    0);
-  PRED("rdf_insert_literal_map",     3,	insert_atom_map,	    0);
+  PRED("rdf_insert_literal_map",     3,	insert_atom_map3,	    0);
+  PRED("rdf_insert_literal_map",     4,	insert_atom_map4,	    0);
   PRED("rdf_delete_literal_map",     3,	delete_atom_map3,	    0);
   PRED("rdf_delete_literal_map",     2,	delete_atom_map2,	    0);
   PRED("rdf_find_literal_map",	     3,	find_atom_map,		    0);
