@@ -25,19 +25,8 @@
 #include <string.h>
 #include <assert.h>
 #include <stdarg.h>
-#ifdef __WINDOWS__
-#  include <io.h>
-#  include <winsock2.h>
-#else
-#  include <unistd.h>
-#  include <sys/types.h>
-#  include <sys/socket.h>
-#  include <netinet/in.h>
-#  include <arpa/inet.h>
-#  include <netdb.h>
-#endif
-
 #include "ssllib.h"
+#include <openssl/rand.h>
 
 #ifdef __SWI_PROLOG__
 #include <SWI-Stream.h>
@@ -49,11 +38,6 @@
  * Remap socket related calls to nonblockio library
  */
 #include "../clib/nonblockio.h"
-#define socket          nbio_socket
-#define connect         nbio_connect
-#define bind            nbio_bind
-#define listen          nbio_listen
-#define accept          nbio_accept
 #define closesocket     nbio_closesocket
 #else   /* __SWI_PROLOG__ */
 #define Soutput         stdout
@@ -68,10 +52,10 @@ static int PL_handle_signals(void) { return 0; }
 #include <openssl/rsa.h>
 
 typedef enum
-{ SSL_SOCK_OK
-, SSL_SOCK_RETRY
-, SSL_SOCK_ERROR
-} SSL_SOCK_STATUS;
+{ SSL_PL_OK
+, SSL_PL_RETRY
+, SSL_PL_ERROR
+} SSL_PL_STATUS;
 
 #define SSL_CERT_VERIFY_MORE 0
 #define SSL_WAIT_CHILD       1
@@ -135,54 +119,47 @@ ssl_error(SSL *ssl, int ssl_ret, int code)
 	   );
 }
 
-static SSL_SOCK_STATUS
-ssl_inspect_status(SSL *ssl, int sock, int ssl_ret)
+static SSL_PL_STATUS
+ssl_inspect_status(SSL *ssl, int ssl_ret)
 {   int code;
 
-    if (ssl_ret >= 0) {
-        return SSL_SOCK_OK;
+    if (ssl_ret > 0) {
+        return SSL_PL_OK;
     }
 
     code=SSL_get_error(ssl, ssl_ret);
 
     switch (code) {
+       /* I am not sure what to do here - specifically, I am not sure if our underlying BIO
+          will block if there is not enough data to complete a handshake. If it will, we should
+          never get these return values. If it wont, then we presumably need to simply try again
+          which is why I am returning SSL_PL_RETRY
+       */
 	case SSL_ERROR_WANT_READ:
-	    if (nbio_wait(sock, REQ_READ) == 0) {
-		return SSL_SOCK_RETRY;
-	    }
-	    break;
+           return SSL_PL_RETRY;
 
 	case SSL_ERROR_WANT_WRITE:
-	    if (nbio_wait(sock, REQ_WRITE) == 0) {
-		return SSL_SOCK_RETRY;
-	    }
-	    break;
+           return SSL_PL_RETRY;
 
 #ifdef SSL_ERROR_WANT_CONNECT
 	case SSL_ERROR_WANT_CONNECT:
-	    if (nbio_wait(sock, REQ_CONNECT) == 0) {
-		return SSL_SOCK_RETRY;
-	    }
-	    break;
+           return SSL_PL_RETRY;
 #endif
 
 #ifdef SSL_ERROR_WANT_ACCEPT
 	case SSL_ERROR_WANT_ACCEPT:
-	    if (nbio_wait(sock, REQ_ACCEPT) == 0) {
-		return SSL_SOCK_RETRY;
-	    }
-	    break;
+           return SSL_PL_RETRY;
 #endif
 
 	case SSL_ERROR_ZERO_RETURN:
-	    return SSL_SOCK_OK;
+	    return SSL_PL_OK;
 
 	default:
 	    break;
     }
 
     ssl_error(ssl, ssl_ret, code);
-    return SSL_SOCK_ERROR;
+    return SSL_PL_ERROR;
 }
 
 static char *
@@ -220,10 +197,10 @@ ssl_cert_print(X509 *cert, const char *role)
 
     i = X509_get_signature_type(cert);
     ssl_deb(1, "\t signature type: %d\n", i);
-#endif
+#endif /* DEBUG */
 }
 
-#endif
+#endif /* SSL_CERT_VERIFY_MORE */
 
 #if SSL_CERT_VERIFY_MORE
 
@@ -258,7 +235,7 @@ ssl_verify_cert(PL_SSL *config)
     }
 }
 
-#endif
+#endif /* SSL_CERT_VERIFY_MORE */
 
 static PL_SSL *
 ssl_new(void)
@@ -272,6 +249,7 @@ ssl_new(void)
         new->pl_ssl_role                = PL_SSL_NONE;
 
 	new->sock			= -1;
+        new->closeparent		= 0;
 
         new->pl_ssl_peer_cert           = NULL;
         new->pl_ssl_ctx                 = NULL;
@@ -291,7 +269,7 @@ ssl_new(void)
         new->pl_ssl_cb_cert_verify_data = NULL;
         new->pl_ssl_cb_pem_passwd       = NULL;
         new->pl_ssl_cb_pem_passwd_data  = NULL;
-	new->magic 		        = SSL_CONFIG_MAGIC;
+	new->magic		        = SSL_CONFIG_MAGIC;
     }
     ssl_deb(1, "Allocated config structure\n");
 
@@ -502,8 +480,8 @@ ssl_set_cb_pem_passwd( PL_SSL *config
 BOOL
 ssl_set_cb_cert_verify( PL_SSL *config
                       , BOOL (*callback)( PL_SSL *
-                                        , const char *
-					, long
+                                        , X509 *
+                                        , X509_STORE_CTX *
                                         , const char *
                                         )
                       , void *data
@@ -526,13 +504,14 @@ ssl_cb_cert_verify(int preverify_ok, X509_STORE_CTX *ctx)
 {
     SSL    * ssl    = NULL;
     PL_SSL * config = NULL;
-
     /*
      * Get our config data
      */
     ssl = X509_STORE_CTX_get_ex_data(ctx, SSL_get_ex_data_X509_STORE_CTX_idx());
     config = SSL_get_ex_data(ssl, ssl_idx);
 
+
+
     ssl_deb(1, " ---- INIT Handling certificate verification\n");
     if (!preverify_ok) {
         X509 *cert = NULL;
@@ -544,6 +523,7 @@ ssl_cb_cert_verify(int preverify_ok, X509_STORE_CTX *ctx)
          */
         cert = X509_STORE_CTX_get_current_cert(ctx);
 
+
         /*
          * Get error specification
          */
@@ -551,22 +531,7 @@ ssl_cb_cert_verify(int preverify_ok, X509_STORE_CTX *ctx)
         error = X509_verify_cert_error_string(err);
 
         if (config->pl_ssl_cb_cert_verify) {
-            BIO * mem = NULL;
-
-            if ((mem = BIO_new(BIO_s_mem())) != NULL) {
-                long n;
-                char *p;
-
-                X509_print(mem, cert);
-                if ((n = BIO_get_mem_data(mem, &p)) > 0) {
-                    preverify_ok = ((config->pl_ssl_cb_cert_verify)(config, p, n, error) != 0);
-                } else {
-                    ssl_err("failed to print certificate\n");
-                }
-                BIO_free(mem);
-            } else {
-                ssl_err("failed to allocate BIO buffer to write certificate\n");
-            }
+           preverify_ok = ((config->pl_ssl_cb_cert_verify)(config, cert, ctx, error) != 0);
         } else {
             char  subject[256];
             char  issuer [256];
@@ -628,6 +593,16 @@ ssl_cb_pem_passwd(char *buf, int size, int rwflag, void *userdata)
     return len;
 }
 
+BOOL
+ssl_set_close_parent(PL_SSL *config, int closeparent)
+/*
+ * Should we close the parent streams?
+ */
+{
+    return config->closeparent = closeparent;
+}
+
+
 int
 ssl_close(PL_SSL_INSTANCE *instance)
 /*
@@ -635,7 +610,6 @@ ssl_close(PL_SSL_INSTANCE *instance)
  */
 {
     int ret = 0;
-
     if (instance) {
         if (instance->config->pl_ssl_role != PL_SSL_SERVER) {
             /*
@@ -649,7 +623,25 @@ ssl_close(PL_SSL_INSTANCE *instance)
         }
 
         if (instance->sock >= 0) {
-            ret = closesocket(instance->sock);
+           /* If the socket has been stored, then we ought to close it if the SSL is being closed */
+           ret = closesocket(instance->sock);
+           instance->sock = -1;
+        }
+
+        if (instance->sread != NULL) {
+           /* Indicate we are no longer filtering the stream */
+           Sset_filter(instance->sread, NULL);
+           /* Close the stream if requested */
+           if (instance->config->closeparent)
+              Sclose(instance->sread);
+        }
+
+        if (instance->swrite != NULL) {
+           /* Indicate we are no longer filtering the stream */
+           Sset_filter(instance->swrite, NULL);
+           /* Close the stream if requested */
+           if (instance->config->closeparent)
+              Sclose(instance->swrite);
         }
 
         free(instance);
@@ -669,13 +661,14 @@ ssl_exit(PL_SSL *config)
 {
     if (config)
     { if ( config->pl_ssl_role == PL_SSL_SERVER && config->sock >= 0 )
-      { closesocket(config->sock);
+      { /* If the socket has been stored, then we ought to close it if the SSL is being closed */
+        closesocket(config->sock);
 	config->sock = -1;
       }
 
       if (config->pl_ssl_ctx)
       { ssl_deb(1, "Calling SSL_CTX_free()\n");
-	SSL_CTX_free(config->pl_ssl_ctx); 	/* doesn't call free hook? */
+	SSL_CTX_free(config->pl_ssl_ctx);	/* doesn't call free hook? */
       }
       else
       { ssl_deb(1, "config without CTX encountered\n");
@@ -847,7 +840,6 @@ ssl_config(PL_SSL *config)
         }
         ssl_deb(1, "certificate installed successfully\n");
     }
-
     (void) SSL_CTX_set_verify( config->pl_ssl_ctx
                              , (config->pl_ssl_peer_cert_required)
                                ? SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT
@@ -860,13 +852,15 @@ ssl_config(PL_SSL *config)
 }
 
 PL_SSL_INSTANCE *
-ssl_instance_new(PL_SSL *config, int sock)
+ssl_instance_new(PL_SSL *config, IOSTREAM* sread, IOSTREAM* swrite)
 {
     PL_SSL_INSTANCE *new = NULL;
 
     if ((new = malloc(sizeof(PL_SSL_INSTANCE))) != NULL) {
         new->config       = config;
-        new->sock         = sock;
+        new->sock         = -1;
+        new->sread        = sread;
+        new->swrite       = swrite;
         new->close_needed = 0;
     }
     return new;
@@ -879,6 +873,9 @@ ssl_lib_init(void)
  */
 {
     SSL_load_error_strings();
+    /* This call will ensure we only end up calling RAND_poll() once
+       - preventing an ugly synchronization issue in OpenSSL */
+    RAND_status();
     (void) SSL_library_init();
 
     if ((ctx_idx = SSL_CTX_get_ex_new_index( 0
@@ -906,44 +903,174 @@ ssl_lib_init(void)
     }
 #endif
 
-    return 0;
-}
-
-int
-ssl_lib_exit(void)
-/*
- * One-time library exit calls
- */
-{
-#ifdef __SWI_PROLOG__
-    nbio_cleanup();
+#ifdef _REENTRANT
+    ssl_thread_setup();
 #endif
+
     return 0;
 }
 
-/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-(*) We use sockets from  the   ../clib/nonblockio.c.  These  sockets are
-integers like Unix file descriptors, but  not the filedescriptor itself.
-nbio_fd() returns the corresponding file/socket descriptor.
 
-Probably this is fine, surely  for   Unix.  Nevertheless,  a much better
-solution would be to wrap the nbio_* functions into an SSL BIO. See "man
-BIO_new", etc.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+/*
+ * BIO routines for SSL over streams
+ */
+
+/*
+ * Read function.  To allow for setting a timeout on the SSL stream, we
+ * use the timeout of this stream if we do not have a timeout ourselves.
+ */
+
+int bio_read(BIO* bio, char* buf, int len)
+{
+   IOSTREAM *stream = BIO_get_ex_data(bio, 0);
+   IOSTREAM *ssl_stream = stream->upstream;
+   int rc;
+
+   if ( ssl_stream &&
+	stream->timeout < 0 &&
+	ssl_stream->timeout > 0 )
+   { stream->timeout = ssl_stream->timeout;
+     rc = (int)Sfread(buf, sizeof(char), len, stream);
+     stream->timeout = -1;
+   } else
+   { rc = (int)Sfread(buf, sizeof(char), len, stream);
+   }
+
+   if ( ssl_stream && (stream->flags & SIO_TIMEOUT) )
+   { ssl_stream->flags |= (SIO_FERR|SIO_TIMEOUT);
+#ifdef ETIMEDOUT
+     errno = ETIMEDOUT;
+#endif
+     rc = -1;
+   }
+
+   return rc;
+}
+
+/*
+ * Gets function. If only OpenSSL actually had usable documentation, I might know
+ * what this was actually meant to do....
+ */
+
+int bio_gets(BIO* bio, char* buf, int len)
+{
+   IOSTREAM* stream;
+   int r = 0;
+   stream = BIO_get_app_data(bio);
+   for (r = 0; r < len; r++)
+   {
+      int c = Sgetc(stream);
+      if (c == EOF)
+         return r-1;
+      buf[r] = (char)c;
+      if (buf[r] == '\n')
+         break;
+   }
+   return r;
+}
+
+/*
+ * Write function
+ */
+
+int bio_write(BIO* bio, const char* buf, int len)
+{
+   IOSTREAM* stream;
+   int r;
+   stream  = BIO_get_ex_data(bio, 0);
+   r = (int)Sfwrite(buf, sizeof(char), len, stream);
+   /* OpenSSL expects there to be no buffering when it writes. Flush here */
+   Sflush(stream);
+   return r;
+}
+
+/*
+ * Control function
+ * (Currently only supports flush. There are several mandatory, but as-yet unsupported functions...)
+ */
+
+long bio_control(BIO* bio, int cmd, long num, void* ptr)
+{
+   IOSTREAM* stream;
+   stream  = BIO_get_ex_data(bio, 0);
+   switch(cmd)
+   {
+     case BIO_CTRL_FLUSH:
+        Sflush(stream);
+        return 1;
+   }
+   return 0;
+}
+
+/*
+ * Create function. Called when a new BIO is created
+ * It is our responsibility to set init to 1 here
+ */
+
+int bio_create(BIO* bio)
+{
+   bio->shutdown = 1;
+   bio->init = 1;
+   bio->num = -1;
+   bio->ptr = NULL;
+   return 1;
+}
+
+/*
+ * Destroy function. Called when a BIO is freed
+ */
+
+int bio_destroy(BIO* bio)
+{
+   if (bio == NULL)
+   {
+      return 0;
+   }
+   return 1;
+}
+
+/*
+ * Specify the BIO read and write function structures
+ */
+
+BIO_METHOD bio_read_functions = {BIO_TYPE_MEM,
+                                 "read",
+                                 NULL,
+                                 &bio_read,
+                                 NULL,
+                                 &bio_gets,
+                                 &bio_control,
+                                 &bio_create,
+                                 &bio_destroy};
+
+BIO_METHOD bio_write_functions = {BIO_TYPE_MEM,
+                                  "write",
+                                  &bio_write,
+                                  NULL,
+                                  NULL,
+                                  NULL,
+                                  &bio_control,
+                                  &bio_create,
+                                  &bio_destroy};
+
+/*
+ * Establish an SSL session using the given read and write streams and the role
+ */
 
 PL_SSL_INSTANCE *
-ssl_ssl(PL_SSL *config, int sock_inst)
+ssl_ssl_bio(PL_SSL *config, IOSTREAM* sread, IOSTREAM* swrite)
 /*
- * Establish the SSL layer using the supplied TCP socket descriptor.
+ * Establish the SSL layer using the supplied streams
  */
 {
     PL_SSL_INSTANCE * instance = NULL;
+    BIO* rbio = NULL;
+    BIO* wbio = NULL;
 
-    if ((instance = ssl_instance_new(config, sock_inst)) == NULL) {
+    if ((instance = ssl_instance_new(config, sread, swrite)) == NULL) {
         ssl_deb(1, "ssl instance malloc failed\n");
         return NULL;
     }
-
     /*
      * Configure SSL behaviour with install configuration parameters
      */
@@ -951,6 +1078,13 @@ ssl_ssl(PL_SSL *config, int sock_inst)
         return NULL;
     }
 
+    /*
+     * Create BIOs
+     */
+    rbio = BIO_new(&bio_read_functions);
+    BIO_set_ex_data(rbio, 0, sread);
+    wbio = BIO_new(&bio_write_functions);
+    BIO_set_ex_data(wbio, 0, swrite);
     /*
      * Prepare SSL layer
      */
@@ -963,34 +1097,24 @@ ssl_ssl(PL_SSL *config, int sock_inst)
      * Store reference to our config data in SSL
      */
     SSL_set_ex_data(instance->ssl, ssl_idx, config);
-
-    if (SSL_set_fd(instance->ssl, nbio_fd(sock_inst)) == 0) 	/* See (*) above */
-    { return NULL;
-    }
+    SSL_set_bio(instance->ssl, rbio, wbio); /* No return value */
     ssl_deb(1, "allocated ssl fd\n");
-
     switch (config->pl_ssl_role) {
         case PL_SSL_SERVER:
             ssl_deb(1, "setting up SSL server side\n");
             do {
                 int ssl_ret = SSL_accept(instance->ssl);
-                switch(ssl_inspect_status(instance->ssl, sock_inst, ssl_ret)) {
-                    case SSL_SOCK_OK:
+                switch(ssl_inspect_status(instance->ssl, ssl_ret)) {
+                    case SSL_PL_OK:
                         /* success */
                         ssl_deb(1, "established ssl server side\n");
                         return instance;
 
-                    case SSL_SOCK_RETRY:
+                    case SSL_PL_RETRY:
                         continue;
 
-                    case SSL_SOCK_ERROR:
-                       if (SSL_get_error(instance->ssl, ssl_ret) == SSL_ERROR_WANT_READ)
-                       {
-                          nbio_wait(sock_inst, REQ_READ);
-                          continue;
-                       }
-                       else
-                          return NULL;
+                    case SSL_PL_ERROR:
+                       return NULL;
                 }
             } while (1);
             break;
@@ -1000,218 +1124,30 @@ ssl_ssl(PL_SSL *config, int sock_inst)
 	   ssl_deb(1, "setting up SSL client side\n");
 	   do {
 	      int ssl_ret = SSL_connect(instance->ssl);
-	      switch(ssl_inspect_status(instance->ssl, sock_inst, ssl_ret)) {
-	         case SSL_SOCK_OK:
+	      switch(ssl_inspect_status(instance->ssl, ssl_ret)) {
+	         case SSL_PL_OK:
 	            /* success */
 	            ssl_deb(1, "established ssl client side\n");
 	            return instance;
 
-	         case SSL_SOCK_RETRY:
+	         case SSL_PL_RETRY:
 	            continue;
 
-	         case SSL_SOCK_ERROR:
-	            if (SSL_get_error(instance->ssl, ssl_ret) == SSL_ERROR_WANT_READ)
-	            {
-	               nbio_wait(sock_inst, REQ_READ);
-	               continue;
-	            }
-	            else if (SSL_get_error(instance->ssl, ssl_ret) == SSL_ERROR_WANT_WRITE)
-	            {
-	               nbio_wait(sock_inst, REQ_WRITE);
-	               continue;
-	            }
-	            else
-	            {
-	               Sdprintf("Unrecoverable error: %d\n", SSL_get_error(instance->ssl, ssl_ret));
-	               Sdprintf("Additionally, get_error returned %d\n", ERR_get_error());
-	               return NULL;
-	            }
-	      }
+	         case SSL_PL_ERROR:
+                    Sdprintf("Unrecoverable error: %d\n", SSL_get_error(instance->ssl, ssl_ret));
+                    Sdprintf("Additionally, get_error returned %d\n", ERR_get_error());
+                    return NULL;
+                    }
 	   } while (1);
 	   break;
     }
     return NULL;
 }
 
-static struct sockaddr_in *
-ssl_hostaddr(struct sockaddr_in *sa_client, const char *host, int port)
-/*
- * Initialize a socket address with supplied host and port.
- */
-{
-    memset(sa_client, '\0', sizeof(*sa_client));
-
-    if (host) {
-        struct in_addr sa_addr = {0};
-        struct hostent *hp     = NULL;
-
-        if ((hp = gethostbyname(host)) == NULL) {
-            ssl_err("Couldn't resolve host: '%s'\n", host);
-            return NULL;
-        }
-        if ((int) sizeof(sa_client->sin_addr) < hp->h_length) {
-            ssl_err("host address too long: %d > %d\n"
-                   , hp->h_length
-                   , (int) sizeof(sa_client->sin_addr)
-                   ) ;
-            return NULL;
-        }
-        memcpy(&sa_client->sin_addr, hp->h_addr_list[0], hp->h_length);
-        memcpy(&sa_addr.s_addr, hp->h_addr_list[0], hp->h_length);
-        if (strcmp(host, hp->h_name)) {
-            ssl_deb(1,  "host '%s' (%s) resolved to '%s'\n"
-                   , host
-                   , hp->h_name
-                   , inet_ntoa(sa_addr)
-                   ) ;
-        } else {
-            ssl_deb(1,  "host '%s' resolved to '%s'\n"
-                   , host
-                   , inet_ntoa(sa_addr)
-                   ) ;
-        }
-    } else {
-        sa_client->sin_addr.s_addr = INADDR_ANY;
-    }
-    sa_client->sin_family = AF_INET;
-    sa_client->sin_port   = htons((unsigned short) port);
-
-    return sa_client;
-}
-
-static int
-ssl_tcp_listen(PL_SSL *config)
-/*
- * Establish TCP server socket
- */
-{
-    struct sockaddr_in sa_server;
-    int                sock    = -1;
-
-    if ((sock = nbio_socket(AF_INET, SOCK_STREAM, 0)) < 0) {
-        perror("socket");
-        return -1;
-    }
-
-    nbio_setopt(sock, TCP_NONBLOCK);
-    if ( config->pl_ssl_reuseaddr )
-        nbio_setopt(sock, TCP_REUSEADDR);
-
-    if (!ssl_hostaddr(&sa_server, config->pl_ssl_host, config->pl_ssl_port)) {
-	closesocket(sock);
-        return -1;
-    }
-
-    if ( nbio_bind(sock,
-		  (struct sockaddr *) &sa_server,
-		  sizeof (sa_server)) < 0)
-    { nbio_closesocket(sock);
-      perror("bind");
-      return -3;
-    }
-
-    /*
-     * Receive a TCP connection.
-     */
-
-    if ( nbio_listen(sock, SSL_TCP_QUEUE_MAX) < 0)
-    { closesocket(sock);
-      perror("listen");
-      return -4;
-    }
-
-    ssl_deb(1, "established tcp server socket\n");
-
-    config->sock = sock;
-
-    return sock;
-}
-
-static int
-ssl_tcp_connect(PL_SSL *config)
-/*
- * Establish TCP client socket
- */
-{
-    int  sock;
-
-    if ( (sock = nbio_socket(AF_INET, SOCK_STREAM, 0)) < 0 )
-    { perror("socket");
-      return -1;
-    }
-    config->sock = sock;
-
-    ssl_deb(1, "established tcp client socket\n");
-
-    return sock;
-}
-
-int
-ssl_socket(PL_SSL *config)
-/*
- * Establish TCP layer and listen or connect,
- * depending on server/client mode.
- */
-{
-    int sock;
-
-    switch (config->pl_ssl_role) {
-        case PL_SSL_SERVER:
-            sock = ssl_tcp_listen(config);
-            break;
-        case PL_SSL_NONE:
-        case PL_SSL_CLIENT:
-            sock = ssl_tcp_connect(config);
-            break;
-	default:
-	    assert(0);
-	    sock = -1;
-    }
-
-    return sock;
-}
-
-int
-ssl_accept(PL_SSL *config, void *addr, socklen_t *addrlen)
-/*
- * We set the TCP layer to accept mode,
- */
-{
-    struct sockaddr_in sa_client;
-    socklen_t          client_len;
-
-    if ( !addr ) {
-	addr = (struct  sockaddr *)&sa_client;
-	client_len = sizeof(sa_client);
-	addrlen = &client_len;
-    }
-
-    return nbio_accept(config->sock, addr, addrlen);
-}
-
-int
-ssl_connect(PL_SSL *config)
-/*
- * We set the TCP layer to connect mode.
- */
-{
-    struct sockaddr_in sa_client = {0};
-    int sock = config->sock;
-
-    if (!ssl_hostaddr(&sa_client, config->pl_ssl_host, config->pl_ssl_port)) {
-        return -1;
-    }
-
-    if ( nbio_connect(sock, (struct sockaddr*) &sa_client, sizeof(sa_client)) < 0)
-      return -1;
-
-    return sock;
-}
-
 int
 ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
 /*
- * Perform read on SSL socket, establish it first if necessary.
+ * Perform read on SSL session
  */
 {
     SSL *ssl = instance->ssl;
@@ -1220,21 +1156,17 @@ ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
 
     do {
         int rbytes = SSL_read(ssl, buf, size);
-        switch(ssl_inspect_status(ssl, instance->sock, rbytes)) {
-            case SSL_SOCK_OK:
+        if (rbytes == 0) /* EOF - error, but we handle in prolog */
+          return 0;
+        switch(ssl_inspect_status(ssl, rbytes)) {
+            case SSL_PL_OK:
                 /* success */
                 return rbytes;
 
-            case SSL_SOCK_RETRY:
+            case SSL_PL_RETRY:
                 continue;
 
-            case SSL_SOCK_ERROR:
-            if (SSL_get_error(instance->ssl, rbytes) == SSL_ERROR_WANT_READ)
-            {
-               nbio_wait(instance->sock, REQ_READ);
-               continue;
-            }
-            else
+            case SSL_PL_ERROR:
                return -1;
        }
     } while (1);
@@ -1243,7 +1175,7 @@ ssl_read(PL_SSL_INSTANCE *instance, char *buf, int size)
 int
 ssl_write(PL_SSL_INSTANCE *instance, const char *buf, int size)
 /*
- * Perform write on SSL socket, establish it first if necessary.
+ * Perform write on SSL session
  */
 {
     SSL *ssl = instance->ssl;
@@ -1252,22 +1184,18 @@ ssl_write(PL_SSL_INSTANCE *instance, const char *buf, int size)
 
     do {
         int wbytes = SSL_write(ssl, buf, size);
-        switch(ssl_inspect_status(ssl, instance->sock, wbytes)) {
-            case SSL_SOCK_OK:
+        if (wbytes == 0) /* EOF - error, but we handle in prolog */
+          return 0;
+        switch(ssl_inspect_status(ssl, wbytes)) {
+            case SSL_PL_OK:
                 /* success */
                 return wbytes;
 
-            case SSL_SOCK_RETRY:
+            case SSL_PL_RETRY:
                 continue;
 
-            case SSL_SOCK_ERROR:
-               if (SSL_get_error(instance->ssl, wbytes) == SSL_ERROR_WANT_WRITE)
-               {
-                  nbio_wait(instance->sock, REQ_WRITE);
-                  continue;
-               }
-               else
-                  return -1;
+            case SSL_PL_ERROR:
+               return -1;
         }
     } while (1);
 }
@@ -1288,6 +1216,8 @@ code is based on mttest.c distributed with the OpenSSL library.
 
 static pthread_mutex_t *lock_cs;
 static long *lock_count;
+static void (*old_locking_callback)(int, int, const char*, int) = NULL;
+static unsigned long (*old_id_callback)(void) = NULL;
 
 static void
 pthreads_locking_callback(int mode, int type, const char *file, int line)
@@ -1332,6 +1262,8 @@ ssl_thread_setup(void)
     pthread_mutex_init(&(lock_cs[i]), NULL);
   }
 
+  old_id_callback = CRYPTO_get_id_callback();
+  old_locking_callback = CRYPTO_get_locking_callback();
 #ifndef __WINDOWS__
   CRYPTO_set_id_callback(pthreads_thread_id);
 #endif
@@ -1350,6 +1282,30 @@ ssl_thread_init(void)
 #endif /*_REENTRANT*/
 
 
+int
+ssl_lib_exit(void)
+/*
+ * One-time library exit calls
+ */
+{
+#ifdef __SWI_PROLOG__
+    nbio_cleanup();
+#endif
+
+/*
+ * If the module is being unloaded, we should remove callbacks pointing to
+ * our address space
+ */
+#ifdef _REENTRANT
+#ifndef __WINDOWS__
+  CRYPTO_set_id_callback(old_id_callback);
+#endif
+  CRYPTO_set_locking_callback(old_locking_callback);
+#endif
+    return 0;
+}
+
+
 /***********************************************************************
  * Warning, error and debug reporting
  ***********************************************************************/
