@@ -37,6 +37,7 @@ static status	prepareEditText(TextObj t, Name selector);
 #define Wrapped(t)      ((t)->wrap == NAME_wrap || \
 			 (t)->wrap == NAME_wrapFixedWidth)
 #define Before(x, y)    if ( valInt(x) > valInt(y) ) { Int _z = x; x=y; y=_z; }
+#define Before_i(x, y)  if ( x > y ) { intptr_t _z = x; x=y; y=_z; }
 #define MakeSel(f, t)   toInt(((valInt(t) & 0xffff) << 16) | \
 			      (valInt(f) & 0xffff))
 #define GetSel(s,f,t)	{ *(t) = ((valInt(s) >> 16) & 0xffff); \
@@ -227,7 +228,7 @@ str_format(String out, const String in, const int width, const FontObj font)
 }
 
 
-static void
+void
 str_one_line(String to, String from)
 { int n;
 
@@ -699,7 +700,7 @@ backgroundText(TextObj t, Any bg)
 
 
 static status
-underlineText(TextObj t, Bool underline)
+underlineText(TextObj t, BoolObj underline)
 { if ( t->underline != underline )
   { CHANGING_GRAPHICAL(t, assign(t, underline, underline);
 		       changedEntireImageGraphical(t));
@@ -710,14 +711,14 @@ underlineText(TextObj t, Bool underline)
 
 
 status
-transparentText(TextObj t, Bool val)
+transparentText(TextObj t, BoolObj val)
 { Any bg = (val == ON ? NIL : DEFAULT);
 
   return backgroundText(t, bg);
 }
 
 
-static Bool
+static BoolObj
 getTransparentText(TextObj t)
 { answer(isNil(t->background) ? ON : OFF);
 }
