@@ -31,7 +31,7 @@
 		********************************/
 
 #ifndef PCE_VERSION
-#define PCE_VERSION "6.6.65, July 2009"
+#define PCE_VERSION "6.6.66, July 2009"
 #endif
 
 #ifndef OS_VERSION
@@ -146,7 +146,7 @@ __pce_export PceITFSymbol getITFSymbolName(PceName);
 
 typedef union
 { char		character;		/* Value is a C char */
-  long		integer;		/* Value of PCE int or reference */
+  intptr_t	integer;		/* Value of PCE int or reference */
   double 	real;			/* Value of PCE real */
   char *	string;			/* Value is a C char * */
   void *	pointer;		/* Anonymous pointer */
