@@ -1,4 +1,4 @@
-/*  $Id: keybinding.c,v 1.35 2006/04/11 16:52:09 jan Exp $
+/*  $Id$
 
     Part of XPCE --- The SWI-Prolog GUI toolkit
 
@@ -54,7 +54,7 @@ initialiseKeyBinding(KeyBinding kb, Name name, int argc, KeyBinding *argv)
 
   if ( notDefault(name) && argc == 0 )
     initPredefinedKeyBinding(kb);
-    
+
   succeed;
 }
 
@@ -88,10 +88,10 @@ static status
 receiverKeyBinding(KeyBinding kb, Any receiver)
 { if ( getHyperedObject(kb, NAME_receiver, DEFAULT) != receiver )
     freeHypersObject(kb, NAME_receiver, DEFAULT);
-  
+
   if ( notNil(receiver) )
     newObject(ClassHyper, kb, receiver, NAME_receiver, NAME_keyBinding, EAV);
-  
+
   succeed;
 }
 
@@ -239,7 +239,7 @@ cuaKeyAsPrefixKeyBinding(KeyBinding kb, EventObj ev, Graphical receiver)
 { if ( hasSendMethodObject(receiver, NAME_cuaKeyAsPrefix) &&
        send(receiver, NAME_cuaKeyAsPrefix, ev, EAV) )
     succeed;
-    
+
   fail;
 }
 
@@ -255,7 +255,7 @@ typedKeyBinding(KeyBinding kb, Any id, Graphical receiver)
   status rval = FAIL;
   EventObj ev = id;
   int clearstatus = (notDefault(kb->argument) || kb->prefix != NAME_);
-        
+
   if ( notDefault(receiver) )
   { if ( receiver != crec )
       resetKeyBinding(kb, receiver);
@@ -268,7 +268,7 @@ typedKeyBinding(KeyBinding kb, Any id, Graphical receiver)
   DEBUG(NAME_keyBinding, writef("%O: Key = %s\n", kb, key));
   if ( instanceOfObject(id, ClassEvent) )
     id = ((EventObj)id)->id;
-    
+
   if ( kb->status == NAME_quotedInsert )
   { cmd = NAME_insertQuoted;
     reset |= RESET_STATUS;
@@ -364,7 +364,7 @@ typedKeyBinding(KeyBinding kb, Any id, Graphical receiver)
 	rval = SUCCEED;
       } else if ( cmd == NAME_quotedInsert )
       { assign(kb, status, NAME_quotedInsert );
-      } else 
+      } else
       { reset |= (RESET_ARGUMENT|RESET_COLUMN);
       }
 
@@ -614,7 +614,7 @@ initInsertKeyBinding(KeyBinding kb)
 
   succeed;
 }
-  
+
 
 static status
 initArgumentKeyBinding(KeyBinding kb)
@@ -653,12 +653,12 @@ static kbDef emacs_special[] =
 };
 
 static kbDef emacs_caret_basics[] =
-{ { "\\C-a", 		NAME_beginningOfLine },
-  { "\\C-b", 		NAME_backwardChar },
-  { "\\C-e", 		NAME_endOfLine },
-  { "\\C-f", 		NAME_forwardChar },
-  { "\\C-n", 		NAME_nextLine },
-  { "\\C-p", 		NAME_previousLine },
+{ { "\\C-a",		NAME_beginningOfLine },
+  { "\\C-b",		NAME_backwardChar },
+  { "\\C-e",		NAME_endOfLine },
+  { "\\C-f",		NAME_forwardChar },
+  { "\\C-n",		NAME_nextLine },
+  { "\\C-p",		NAME_previousLine },
 
   { "\\eb",		NAME_backwardWord },
   { "\\ef",		NAME_forwardWord },
@@ -676,11 +676,11 @@ static kbDef emacs_edit_basics[] =
 { { SUPER,		NAME_insert },
 
   { "backspace",	NAME_cutOrBackwardDeleteChar },
-  { "\\C-d", 		NAME_deleteChar },
+  { "\\C-d",		NAME_deleteChar },
   { "DEL",		NAME_cutOrDeleteChar },
-  { "\\C-t", 		NAME_transposeChars },
-  { "\\C-k", 		NAME_killLine },
-  { "\\C-o", 		NAME_openLine },
+  { "\\C-t",		NAME_transposeChars },
+  { "\\C-k",		NAME_killLine },
+  { "\\C-o",		NAME_openLine },
 
   { "\\ed",		NAME_killWord },
   { "\\ebackspace",	NAME_backwardKillWord },
@@ -757,10 +757,10 @@ static kbDef list_browser[] =
   { "TAB",		NAME_extendPrefixOrNext },
   { "\\C-w",		NAME_extendToCurrent },
   { "\\C-s",		NAME_repeatSearch },
-  { "\\C-n", 		NAME_nextLine },
+  { "\\C-n",		NAME_nextLine },
   { "cursor_down",	NAME_nextLine },
-  { "\\C-p", 		NAME_previousLine },
-  { "cursor_up", 	NAME_previousLine },
+  { "\\C-p",		NAME_previousLine },
+  { "cursor_up",	NAME_previousLine },
 
 /*{ DEFAULT_FUNCTION,	NAME_alert },*/
 
@@ -791,14 +791,15 @@ static kbDef editor[] =
   { "page_down",	NAME_cursorPageDown },
 
   { "\\C-@",		NAME_setMark },
-  { "LFD", 		NAME_newlineAndIndent },
-  { "\\C-j", 		NAME_newlineAndIndent },
-  { "\\C-r", 		NAME_isearchBackward },
-  { "\\C-s", 		NAME_isearchForward },
+  { "LFD",		NAME_newlineAndIndent },
+  { "\\C-j",		NAME_newlineAndIndent },
+  { "\\C-r",		NAME_isearchBackward },
+  { "\\C-s",		NAME_isearchForward },
+  { "key_top_3",	NAME_isearchForward },
   { "\\C-w",		NAME_killOrGrabRegion },
   { "\\C-y",		NAME_yank },
-  { "\\C-_", 		NAME_undo },
-  { "\\C-z", 		NAME_undo },		/* Windows compatibility */
+  { "\\C-_",		NAME_undo },
+  { "\\C-z",		NAME_undo },		/* Windows compatibility */
 
   { "\\e\\C-b",		NAME_backwardTerm },
   { "\\e\\C-f",		NAME_forwardTerm },
@@ -823,6 +824,7 @@ static kbDef editor[] =
   { "\\eDEL",		NAME_backwardKillWord },
 
   { "\\C-x\\C-x",	NAME_exchangePointAndMark },
+  { "\\C-xh",		NAME_markWholeBuffer },
   { "\\C-x\\C-o",	NAME_deleteBlankLines },
   { "\\C-x\\C-t",	NAME_transposeLines },
   { "\\C-x\\C-l",	NAME_downcaseRegion },
@@ -832,7 +834,7 @@ static kbDef editor[] =
 
   { DEFAULT_FUNCTION,	NAME_undefined },
 
-  { NULL, 	        NULL }
+  { NULL,	        NULL }
 };
 
 
@@ -918,7 +920,7 @@ initPredefinedKeyBinding(KeyBinding kb)
     } else
       functionKeyBinding(kb,
 			 table->key[0] == '\\' ? CtoName(table->key)
-			 		       : CtoKeyword(table->key),
+					       : CtoKeyword(table->key),
 			 table->function);
   }
 
