@@ -1,4 +1,4 @@
-/*  $Id: win32.h,v 1.36 2007/01/10 11:05:46 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -24,9 +24,8 @@
 
 #define WIN32 1
 
-#define O_GMP		1
 #define BOOTFILE	"boot32.prc"
-#define ARCH		"i386-win32"
+#define PLARCH		"i386-win32"
 #define SIZEOF_VOIDP	4
 
 #include "windows.h"
