@@ -1,4 +1,4 @@
-/*  $Id: win64.h,v 1.2 2007/01/10 14:04:18 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -25,8 +25,8 @@
 #define WIN64 1
 
 #define JMPBUF_ALIGNMENT 16		/* How to get this value? */
-#define BOOTFILE 	"boot64.prc"
-#define ARCH		"x64-win64"
+#define BOOTFILE	"boot64.prc"
+#define PLARCH		"x64-win64"
 #define SIZEOF_VOIDP	8
 
 #include "windows.h"
