@@ -133,9 +133,12 @@ unify_hsz(term_t term, HSZ hsz)
 
   if ( len == sizeof(buf)/sizeof(wchar_t)-1 )
   { if ( (len=DdeQueryStringW(ddeInst, hsz, NULL, 0, CP_WINUNICODE)) > 0 )
-    { wchar_t *b2 = malloc((len+1)*sizeof(wchar_t));
+    { wchar_t *b2;
       int rc;
 
+      if ( !(b2 = malloc((len+1)*sizeof(wchar_t))) )
+	return PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_memory);
+
       DdeQueryStringW(ddeInst, hsz, b2, len+1, CP_WINUNICODE);
       rc = PL_unify_wchars(term, PL_ATOM, len, b2);
       free(b2);
@@ -152,19 +155,22 @@ unify_hsz(term_t term, HSZ hsz)
 
 static word
 unify_hdata(term_t t, HDDEDATA data)
-{ char buf[FASTBUFSIZE];
-  int len;
+{ BYTE buf[FASTBUFSIZE];
+  DWORD len;
 
   if ( !(len=DdeGetData(data, buf, sizeof(buf), 0)) )
     return dde_warning("data handle");
 
-  DEBUG(0, Sdprintf("DdeGetData() returned %d bytes\n", len));
+  DEBUG(0, Sdprintf("DdeGetData() returned %ld bytes\n", (long)len));
 
   if ( len == sizeof(buf) )
   { if ( (len=DdeGetData(data, NULL, 0, 0)) > 0 )
-    { char *b2 = malloc(len);
+    { LPBYTE b2;
       int rval;
 
+      if ( !(b2 = malloc(len)) )
+	return PL_error(NULL, 0, NULL, ERR_RESOURCE, ATOM_memory);
+
       DdeGetData(data, b2, len, 0);
       rval = PL_unify_wchars(t, PL_ATOM, len/sizeof(wchar_t)-1, (wchar_t*)b2);
       free(b2);
@@ -233,7 +239,7 @@ findServerHandle(HCONV handle)
 static HDDEDATA CALLBACK
 DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
             HDDEDATA hData, DWORD dwData1, DWORD dwData2)
-{
+{ GET_LD
   switch(type)
   {  case XTYP_CONNECT:
      { fid_t cid = PL_open_foreign_frame();
@@ -253,10 +259,7 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
        return (void *)rval;
      }
      case XTYP_CONNECT_CONFIRM:
-     { fid_t cid = PL_open_foreign_frame();
-       term_t argv = PL_new_term_refs(3);
-       predicate_t pred = PL_pred(FUNCTOR_dde_connect_confirm3, MODULE_dde);
-       int plhandle;
+     { int plhandle;
 
        if ( (plhandle = allocServerHandle(hconv)) >= 0 )
        { fid_t cid = PL_open_foreign_frame();
@@ -318,7 +321,6 @@ DdeCallback(UINT type, UINT fmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
 	 PL_put_integer( argv+0, plhandle);
 	 unify_hsz(	 argv+1, hsz1);	/* topic */
 	 unify_hsz(      argv+2, hsz2);	/* item */
-	 PL_put_variable(argv+3);
 
 	 if ( PL_call_predicate(MODULE_dde, TRUE, pred, argv) )
 	 { wchar_t *s;
@@ -404,7 +406,8 @@ pl_dde_register_service(term_t topic, term_t onoff)
 
 word
 pl_open_dde_conversation(term_t service, term_t topic, term_t handle)
-{ UINT i;
+{ GET_LD
+  UINT i;
   HSZ Hservice, Htopic;
 
   if ( !dde_initialise() )
@@ -434,7 +437,8 @@ pl_open_dde_conversation(term_t service, term_t topic, term_t handle)
 
 static int
 get_conv_handle(term_t handle, int *theh)
-{ int h;
+{ GET_LD
+  int h;
 
   if ( !PL_get_integer(handle, &h) || h < 0 || h >= MAX_CONVERSATIONS )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_dde_handle, handle);
@@ -495,7 +499,7 @@ pl_dde_request(term_t handle, term_t item,
   DdeFreeStringHandle(ddeInst, Hitem);
 
   if ( Hvalue)
-  { char *valuedata;
+  { LPBYTE valuedata;
 
     if ( (valuedata = DdeAccessData(Hvalue, &valuelen)) )
     { DEBUG(0, Sdprintf("valuelen = %ld; format = %d\n", valuelen, fmti));
