@@ -35,33 +35,40 @@ throw(error(<Formal>, <SWI-Prolog>))
 #include <errno.h>
 #endif
 
-static void
+static int
 put_name_arity(term_t t, functor_t f)
-{ FunctorDef fdef = valueFunctor(f);
-  term_t a = PL_new_term_refs(2);
+{ GET_LD
+  FunctorDef fdef = valueFunctor(f);
+  term_t a;
 
-  PL_put_atom(a+0, fdef->name);
-  PL_put_integer(a+1, fdef->arity);
-  PL_cons_functor(t, FUNCTOR_divide2, a+0, a+1);
+  if ( (a=PL_new_term_refs(2)) )
+  { PL_put_atom(a+0, fdef->name);
+
+    return (PL_put_integer(a+1, fdef->arity) &&
+	    PL_cons_functor(t, FUNCTOR_divide2, a+0, a+1));
+  }
+
+  return FALSE;
 }
 
 
 static void
 rewrite_callable(atom_t *expected, term_t actual)
-{ term_t a = 0;
+{ GET_LD
+  term_t a = 0;
   int loops = 0;
 
   while ( PL_is_functor(actual, FUNCTOR_colon2) )
   { if ( !a )
      a = PL_new_term_ref();
 
-    PL_get_arg(1, actual, a);
+    _PL_get_arg(1, actual, a);
     if ( !PL_is_atom(a) )
     { *expected = ATOM_atom;
       PL_put_term(actual, a);
       return;
     } else
-    { PL_get_arg(2, actual, a);
+    { _PL_get_arg(2, actual, a);
       PL_put_term(actual, a);
     }
 
@@ -72,8 +79,10 @@ rewrite_callable(atom_t *expected, term_t actual)
 
 
 int
-PL_error(const char *pred, int arity, const char *msg, int id, ...)
-{ Definition caller;
+PL_error(const char *pred, int arity, const char *msg, PL_error_code id, ...)
+{ GET_LD
+  char msgbuf[50];
+  Definition caller;
   term_t except, formal, swi;
   va_list args;
   int do_throw = FALSE;
@@ -95,7 +104,11 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     msg = OsError();
   }
 
-  fid    = PL_open_foreign_frame();
+  LD->exception.processing = TRUE;	/* allow using spare stack */
+
+  if ( !(fid = PL_open_foreign_frame()) )
+    goto nomem;
+
   except = PL_new_term_ref();
   formal = PL_new_term_ref();
   swi    = PL_new_term_ref();
@@ -105,38 +118,48 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
   switch(id)
   { case ERR_INSTANTIATION:
       err_instantiation:
-      PL_unify_atom(formal, ATOM_instantiation_error);
+      rc = PL_unify_atom(formal, ATOM_instantiation_error);
       break;
-    case ERR_MUST_BE_VAR:
+    case ERR_UNINSTANTIATION:
     { int argn = va_arg(args, int);
-      char buf[50];
-      /*term_t bound =*/ va_arg(args, term_t);
+      term_t bound = va_arg(args, term_t);
 
-      if ( !msg )
-      { Ssprintf(buf, "%d-%s argument",
+      if ( !msg && argn > 0 )
+      { Ssprintf(msgbuf, "%d-%s argument",
 		 argn, argn == 1 ? "st" : argn == 2 ? "nd" : "th");
-	msg = buf;
+	msg = msgbuf;
       }
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_representation_error1,
-		      PL_ATOM, ATOM_variable);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_uninstantiation_error1,
+			   PL_TERM, bound);
       break;
     }
     case ERR_TYPE:			/* ERR_INSTANTIATION if var(actual) */
     { atom_t expected = va_arg(args, atom_t);
       term_t actual   = va_arg(args, term_t);
 
+    case_type_error:
       if ( expected == ATOM_callable )
 	rewrite_callable(&expected, actual);
       if ( PL_is_variable(actual) && expected != ATOM_variable )
 	goto err_instantiation;
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_ATOM, expected,
-		      PL_TERM, actual);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_type_error2,
+			   PL_ATOM, expected,
+			   PL_TERM, actual);
       break;
+    case ERR_PTR_TYPE:			/* atom_t, Word */
+      { Word ptr;
+
+	expected = va_arg(args, atom_t);
+	ptr      = va_arg(args, Word);
+	actual   = PL_new_term_ref();
+
+	*valTermRef(actual) = *ptr;
+	goto case_type_error;
+      }
     }
     case ERR_CHARS_TYPE:		/* ERR_INSTANTIATION if var(actual) */
     { const char *expected = va_arg(args, const char*);
@@ -145,10 +168,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       if ( PL_is_variable(actual) && !streq(expected, "variable") )
 	goto err_instantiation;
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_CHARS, expected,
-		      PL_TERM, actual);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_type_error2,
+			   PL_CHARS, expected,
+			   PL_TERM, actual);
       break;
     }
     case ERR_AR_TYPE:			/* arithmetic type error */
@@ -156,29 +179,41 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       Number num      = va_arg(args, Number);
       term_t actual   = PL_new_term_ref();
 
-      _PL_put_number(actual, num);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_ATOM, expected,
-		      PL_TERM, actual);
+      rc = (_PL_put_number(actual, num) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_type_error2,
+			    PL_ATOM, expected,
+			    PL_TERM, actual));
+      break;
+    }
+    case ERR_AR_DOMAIN:
+    { atom_t domain = va_arg(args, atom_t);
+      Number num    = va_arg(args, Number);
+      term_t actual = PL_new_term_ref();
+
+      rc = (_PL_put_number(actual, num) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_domain_error2,
+			    PL_ATOM, domain,
+			    PL_TERM, actual));
       break;
     }
     case ERR_AR_UNDEF:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_undefined);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_undefined);
       break;
     }
     case ERR_AR_OVERFLOW:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_float_overflow);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_float_overflow);
       break;
     }
     case ERR_AR_UNDERFLOW:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_float_underflow);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_float_underflow);
       break;
     }
     case ERR_DOMAIN:			/*  ERR_INSTANTIATION if var(arg) */
@@ -188,42 +223,44 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       if ( PL_is_variable(arg) )
 	goto err_instantiation;
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_domain_error2,
-		      PL_ATOM, domain,
-		      PL_TERM, arg);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_domain_error2,
+			   PL_ATOM, domain,
+			   PL_TERM, arg);
       break;
     }
     case ERR_REPRESENTATION:
     { atom_t what = va_arg(args, atom_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_representation_error1,
-		      PL_ATOM, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_representation_error1,
+			   PL_ATOM, what);
       break;
     }
     case ERR_MODIFY_STATIC_PROC:
     { Procedure proc = va_arg(args, Procedure);
       term_t pred = PL_new_term_ref();
 
-      unify_definition(pred, proc->definition, 0, GP_NAMEARITY|GP_HIDESYSTEM);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_permission_error3,
-		      PL_ATOM, ATOM_modify,
-		      PL_ATOM, ATOM_static_procedure,
-		      PL_TERM, pred);
+      rc = (unify_definition(MODULE_user, pred, proc->definition, 0,
+			     GP_NAMEARITY|GP_HIDESYSTEM) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_permission_error3,
+			    PL_ATOM, ATOM_modify,
+			    PL_ATOM, ATOM_static_procedure,
+			    PL_TERM, pred));
       break;
     }
     case ERR_MODIFY_THREAD_LOCAL_PROC:
     { Procedure proc = va_arg(args, Procedure);
       term_t pred = PL_new_term_ref();
 
-      unify_definition(pred, proc->definition, 0, GP_NAMEARITY|GP_HIDESYSTEM);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_permission_error3,
-		      PL_ATOM, ATOM_modify,
-		      PL_ATOM, ATOM_thread_local_procedure,
-		      PL_TERM, pred);
+      rc = (unify_definition(MODULE_user, pred, proc->definition, 0,
+			     GP_NAMEARITY|GP_HIDESYSTEM) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_permission_error3,
+			    PL_ATOM, ATOM_modify,
+			    PL_ATOM, ATOM_thread_local_procedure,
+			    PL_TERM, pred));
       break;
     }
     case ERR_UNDEFINED_PROC:
@@ -234,11 +271,11 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       if ( clr )
 	caller = clr;
 
-      unify_definition(pred, def, 0, GP_NAMEARITY);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_existence_error2,
-		      PL_ATOM, ATOM_procedure,
-		      PL_TERM, pred);
+      rc = (unify_definition(MODULE_user, pred, def, 0, GP_NAMEARITY) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_existence_error2,
+			    PL_ATOM, ATOM_procedure,
+			    PL_TERM, pred));
       break;
     }
     case ERR_PERMISSION_PROC:
@@ -247,71 +284,70 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       predicate_t pred = va_arg(args, predicate_t);
       term_t pi = PL_new_term_ref();
 
-      PL_unify_predicate(pi, pred, GP_NAMEARITY|GP_HIDESYSTEM);
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_permission_error3,
-		    PL_ATOM, op,
-		    PL_ATOM, type,
-		    PL_TERM, pi);
+      rc = ( PL_unify_predicate(pi, pred, GP_NAMEARITY|GP_HIDESYSTEM) &&
+	     PL_unify_term(formal,
+			   PL_FUNCTOR, FUNCTOR_permission_error3,
+			     PL_ATOM, op,
+			     PL_ATOM, type,
+			     PL_TERM, pi));
       break;
     }
     case ERR_NOT_IMPLEMENTED_PROC:
     { const char *name = va_arg(args, const char *);
       int arity = va_arg(args, int);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_not_implemented2,
-		      PL_ATOM, ATOM_procedure,
-		      PL_FUNCTOR, FUNCTOR_divide2,
-		        PL_CHARS, name,
-		        PL_INT, arity);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_not_implemented2,
+			   PL_ATOM, ATOM_procedure,
+			   PL_FUNCTOR, FUNCTOR_divide2,
+			     PL_CHARS, name,
+			     PL_INT, arity);
       break;
     }
     case ERR_FAILED:
     { term_t goal = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_failure_error1,
-		      PL_TERM, goal);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_failure_error1,
+			   PL_TERM, goal);
 
       break;
     }
     case ERR_EVALUATION:
     { atom_t what = va_arg(args, atom_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, what);
       break;
     }
     case ERR_NOT_EVALUABLE:
     { functor_t f = va_arg(args, functor_t);
       term_t actual = PL_new_term_ref();
 
-      put_name_arity(actual, f);
-
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_type_error2,
-		      PL_ATOM, ATOM_evaluable,
-		      PL_TERM, actual);
+      rc = (put_name_arity(actual, f) &&
+	    PL_unify_term(formal,
+			  PL_FUNCTOR, FUNCTOR_type_error2,
+			    PL_ATOM, ATOM_evaluable,
+			    PL_TERM, actual));
       break;
     }
     case ERR_DIV_BY_ZERO:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_evaluation_error1,
-		      PL_ATOM, ATOM_zero_divisor);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_evaluation_error1,
+			   PL_ATOM, ATOM_zero_divisor);
       break;
     }
     case ERR_PERMISSION:
-    { atom_t type = va_arg(args, atom_t);
-      atom_t op   = va_arg(args, atom_t);
+    { atom_t op   = va_arg(args, atom_t);
+      atom_t type = va_arg(args, atom_t);
       term_t obj  = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_permission_error3,
-			  PL_ATOM, type,
-			  PL_ATOM, op,
-			  PL_TERM, obj);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_permission_error3,
+			   PL_ATOM, op,
+			   PL_ATOM, type,
+			   PL_TERM, obj);
 
       break;
     }
@@ -319,10 +355,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { Word p1  = va_arg(args, Word);
       Word p2  = va_arg(args, Word);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_occurs_check2,
-			  PL_TERM, wordToTermRef(p1),
-			  PL_TERM, wordToTermRef(p2));
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_occurs_check2,
+			   PL_TERM, wordToTermRef(p1),
+			   PL_TERM, wordToTermRef(p2));
 
       break;
     }
@@ -330,10 +366,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { atom_t op   = va_arg(args, atom_t);
       term_t obj  = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_timeout_error2,
-			  PL_ATOM, op,
-			  PL_TERM, obj);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_timeout_error2,
+			   PL_ATOM, op,
+			   PL_TERM, obj);
 
       break;
     }
@@ -341,10 +377,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { atom_t type = va_arg(args, atom_t);
       term_t obj  = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_existence_error2,
-			  PL_ATOM, type,
-			  PL_TERM, obj);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_existence_error2,
+			   PL_ATOM, type,
+			   PL_TERM, obj);
 
       break;
     }
@@ -354,18 +390,21 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
       term_t file   = va_arg(args, term_t);
 
       switch(errno)
-      { case EACCES:
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_permission_error3,
-			  PL_ATOM, action,
-			  PL_ATOM, type,
-			  PL_TERM, file);
+      { case EAGAIN:
+	  action = ATOM_lock;		/* Hack for file-locking*/
+	  /*FALLTHROUGH*/
+	case EACCES:
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_permission_error3,
+			       PL_ATOM, action,
+			       PL_ATOM, type,
+			       PL_TERM, file);
 	  break;
 	case EMFILE:
 	case ENFILE:
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_resource_error1,
-			  PL_ATOM, ATOM_max_files);
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_resource_error1,
+			       PL_ATOM, ATOM_max_files);
 	  break;
 #ifdef EPIPE
 	case EPIPE:
@@ -374,10 +413,10 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 	  /*FALLTHROUGH*/
 #endif
 	default:			/* what about the other cases? */
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_existence_error2,
-			  PL_ATOM, type,
-			  PL_TERM, file);
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_existence_error2,
+			       PL_ATOM, type,
+			       PL_TERM, file);
 	  break;
       }
 
@@ -387,61 +426,61 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { atom_t action = va_arg(args, atom_t);
       term_t stream = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_io_error2,
-		      PL_ATOM, action,
-		      PL_TERM, stream);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_io_error2,
+			   PL_ATOM, action,
+			   PL_TERM, stream);
       break;
     }
     case ERR_DDE_OP:
     { const char *op  = va_arg(args, const char *);
       const char *err = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_dde_error2,
-		      PL_CHARS, op,
-		      PL_CHARS, err);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_dde_error2,
+			   PL_CHARS, op,
+			   PL_CHARS, err);
       break;
     }
     case ERR_SHARED_OBJECT_OP:
     { atom_t action = va_arg(args, atom_t);
       const char *err = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_shared_object2,
-		      PL_ATOM,  action,
-		      PL_CHARS, err);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_shared_object2,
+			   PL_ATOM,  action,
+			   PL_CHARS, err);
       break;
     }
     case ERR_NOT_IMPLEMENTED:		/* non-ISO */
     { const char *what = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_not_implemented2,
-		          PL_ATOM, ATOM_feature,
-			  PL_CHARS, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_not_implemented2,
+			   PL_ATOM, ATOM_feature,
+			   PL_CHARS, what);
       break;
     }
     case ERR_RESOURCE:
     { atom_t what = va_arg(args, atom_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_resource_error1,
-		      PL_ATOM, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_resource_error1,
+			   PL_ATOM, what);
       break;
     }
     case ERR_SYNTAX:
     { const char *what = va_arg(args, const char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_syntax_error1,
-		      PL_CHARS, what);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_syntax_error1,
+			   PL_CHARS, what);
       break;
     }
     case ERR_NOMEM:
-    { PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_resource_error1,
-		      PL_ATOM, ATOM_no_memory);
+    { rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_resource_error1,
+			   PL_ATOM, ATOM_no_memory);
 
       break;
     }
@@ -453,12 +492,12 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
 
       switch(errno)
       { case ENOMEM:
-	  PL_unify_term(formal,
-			PL_FUNCTOR, FUNCTOR_resource_error1,
-			  PL_ATOM, ATOM_no_memory);
+	  rc = PL_unify_term(formal,
+			     PL_FUNCTOR, FUNCTOR_resource_error1,
+			       PL_ATOM, ATOM_no_memory);
 	  break;
 	default:
-	  PL_unify_atom(formal, ATOM_system_error);
+	  rc = PL_unify_atom(formal, ATOM_system_error);
 	  break;
       }
 
@@ -467,40 +506,40 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     case ERR_SHELL_FAILED:
     { term_t cmd = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_shell2,
-		      PL_ATOM, ATOM_execute,
-		      PL_TERM, cmd);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_shell2,
+			   PL_ATOM, ATOM_execute,
+			   PL_TERM, cmd);
       break;
     }
     case ERR_SHELL_SIGNALLED:
     { term_t cmd = va_arg(args, term_t);
       int sig = va_arg(args, int);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_shell2,
-		      PL_FUNCTOR, FUNCTOR_signal1,
-		        PL_INT, sig,
-		      PL_TERM, cmd);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_shell2,
+			   PL_FUNCTOR, FUNCTOR_signal1,
+			     PL_INT, sig,
+			 PL_TERM, cmd);
       break;
     }
     case ERR_SIGNALLED:
     { int   sig     = va_arg(args, int);
       char *signame = va_arg(args, char *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_signal2,
-			PL_CHARS,   signame,
-		        PL_INT, sig);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_signal2,
+			   PL_CHARS, signame,
+			   PL_INT, sig);
       break;
     }
     case ERR_CLOSED_STREAM:
     { IOSTREAM *s = va_arg(args, IOSTREAM *);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR, FUNCTOR_existence_error2,
-		    PL_ATOM, ATOM_stream,
-		    PL_POINTER, s);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR, FUNCTOR_existence_error2,
+			   PL_ATOM, ATOM_stream,
+			   PL_POINTER, s);
       do_throw = TRUE;
       break;
     }
@@ -508,25 +547,25 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
     { atom_t type  = va_arg(args, atom_t);
       term_t mutex = va_arg(args, term_t);
 
-      PL_unify_term(formal, PL_FUNCTOR, FUNCTOR_busy2, type, mutex);
+      rc = PL_unify_term(formal, PL_FUNCTOR, FUNCTOR_busy2, type, mutex);
       break;
     }
     case ERR_FORMAT:
     { const char *s = va_arg(args, const char*);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR_CHARS, "format", 1,
-		      PL_CHARS, s);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR_CHARS, "format", 1,
+			   PL_CHARS, s);
       break;
     }
     case ERR_FORMAT_ARG:
     { const char *s = va_arg(args, const char*);
       term_t arg = va_arg(args, term_t);
 
-      PL_unify_term(formal,
-		    PL_FUNCTOR_CHARS, "format_argument_type", 2,
-		      PL_CHARS, s,
-		      PL_TERM, arg);
+      rc = PL_unify_term(formal,
+			 PL_FUNCTOR_CHARS, "format_argument_type", 2,
+			   PL_CHARS, s,
+			   PL_TERM, arg);
       break;
     }
     default:
@@ -535,34 +574,40 @@ PL_error(const char *pred, int arity, const char *msg, int id, ...)
   va_end(args);
 
 					/* build SWI-Prolog context term */
-  if ( pred || msg || caller )
+  if ( rc && (pred || msg || caller) )
   { term_t predterm = PL_new_term_ref();
     term_t msgterm  = PL_new_term_ref();
 
     if ( pred )
-    { PL_unify_term(predterm,
-		    PL_FUNCTOR, FUNCTOR_divide2,
-		      PL_CHARS, pred,
-		      PL_INT, arity);
+    { rc = PL_unify_term(predterm,
+			 PL_FUNCTOR, FUNCTOR_divide2,
+			   PL_CHARS, pred,
+			   PL_INT, arity);
     } else if ( caller )
-    { unify_definition(predterm, caller, 0, GP_NAMEARITY);
+    { rc = unify_definition(MODULE_user, predterm, caller, 0, GP_NAMEARITY);
     }
 
-    if ( msg )
-    { PL_put_atom_chars(msgterm, msg);
+    if ( rc && msg )
+    { rc = PL_put_atom_chars(msgterm, msg);
     }
 
-    PL_unify_term(swi,
-		  PL_FUNCTOR, FUNCTOR_context2,
-		    PL_TERM, predterm,
-		    PL_TERM, msgterm);
+    if ( rc )
+      rc = PL_unify_term(swi,
+			 PL_FUNCTOR, FUNCTOR_context2,
+			   PL_TERM, predterm,
+			   PL_TERM, msgterm);
   }
 
-  PL_unify_term(except,
-		PL_FUNCTOR, FUNCTOR_error2,
-		  PL_TERM, formal,
-		  PL_TERM, swi);
+  if ( rc )
+    rc = PL_unify_term(except,
+		       PL_FUNCTOR, FUNCTOR_error2,
+		         PL_TERM, formal,
+		         PL_TERM, swi);
 
+  if ( !rc )
+  { nomem:
+    fatalError("Cannot report error: no memory");
+  }
 
   if ( do_throw )
     rc = PL_throw(except);
@@ -587,6 +632,75 @@ tostr(char *buf, const char *fmt, ...)
 }
 
 
+		 /*******************************
+		 *	  TYPICAL ERRORS	*
+		 *******************************/
+
+int
+PL_instantiation_error(term_t actual)
+{ return PL_error(NULL, 0, NULL, ERR_INSTANTIATION);
+}
+
+int
+PL_representation_error(const char *resource)
+{ atom_t r = PL_new_atom(resource);
+  int rc = PL_error(NULL, 0, NULL, ERR_RESOURCE, r);
+  PL_unregister_atom(r);
+
+  return rc;
+}
+
+
+int
+PL_type_error(const char *expected, term_t actual)
+{ return PL_error(NULL, 0, NULL, ERR_CHARS_TYPE, expected, actual);
+}
+
+
+int
+PL_domain_error(const char *expected, term_t actual)
+{ atom_t a = PL_new_atom(expected);
+  int rc = PL_error(NULL, 0, NULL, ERR_DOMAIN, a, actual);
+  PL_unregister_atom(a);
+
+  return rc;
+}
+
+
+int
+PL_existence_error(const char *type, term_t actual)
+{ atom_t a = PL_new_atom(type);
+  int rc = PL_error(NULL, 0, NULL, ERR_EXISTENCE, a, actual);
+  PL_unregister_atom(a);
+
+  return rc;
+}
+
+
+int
+PL_permission_error(const char *op, const char *type, term_t obj)
+{ atom_t t = PL_new_atom(type);
+  atom_t o = PL_new_atom(op);
+  int rc = PL_error(NULL, 0, NULL, ERR_PERMISSION, o, t, obj);
+
+  PL_unregister_atom(t);
+  PL_unregister_atom(o);
+
+  return rc;
+}
+
+
+int
+PL_resource_error(const char *resource)
+{ atom_t r = PL_new_atom(resource);
+  int rc = PL_error(NULL, 0, NULL, ERR_RESOURCE, r);
+
+  PL_unregister_atom(r);
+
+  return rc;
+}
+
+
 		 /*******************************
 		 *	PRINTING MESSAGES	*
 		 *******************************/
@@ -599,44 +713,47 @@ PL_unify_term(). This predicate saves possible   pending  exceptions and
 restores them to make the call from B_THROW possible.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-void
+#define OK_RECURSIVE 10
+
+int
 printMessage(atom_t severity, ...)
-{ fid_t fid;
-  term_t ex, av;
+{ GET_LD
+  wakeup_state wstate;
+  term_t av;
   predicate_t pred = PROCEDURE_print_message2;
   va_list args;
+  int rc;
+
+  if ( ++LD->in_print_message >= OK_RECURSIVE*3 )
+    fatalError("printMessage(): recursive call\n");
+  if ( !saveWakeup(&wstate, TRUE PASS_LD) )
+  { LD->in_print_message--;
+    return FALSE;
+  }
 
-  blockGC(PASS_LD1);			/* sometimes called from dangerous */
-					/* places */
-  fid = PL_open_foreign_frame();
   av = PL_new_term_refs(2);
-
-  if ( exception_term )
-  { ex = PL_copy_term_ref(exception_term);
-    exception_term = 0;
-  } else
-    ex = 0;
-
   va_start(args, severity);
   PL_put_atom(av+0, severity);
-  PL_unify_termv(av+1, args);
+  rc = PL_unify_termv(av+1, args);
   va_end(args);
 
-  if ( isDefinedProcedure(pred) )
-    PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_CATCH_EXCEPTION, pred, av);
-  else
-  { Sfprintf(Serror, "Message: ");
-    PL_write_term(Serror, av+1, 1200, 0);
-    Sfprintf(Serror, "\n");
+  if ( rc )
+  { if ( isDefinedProcedure(pred) && LD->in_print_message <= OK_RECURSIVE )
+    { rc = PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_CATCH_EXCEPTION,
+			     pred, av);
+    } else if ( LD->in_print_message <= OK_RECURSIVE*2 )
+    { Sfprintf(Serror, "Message: ");
+      rc = PL_write_term(Serror, av+1, 1200, 0);
+      Sfprintf(Serror, "\n");
+    } else				/* in_print_message == 2 */
+    { Sfprintf(Serror, "printMessage(): recursive call\n");
+    }
   }
 
-  if ( ex )
-  { PL_put_term(exception_bin, ex);
-    exception_term = exception_bin;
-  }
+  restoreWakeup(&wstate PASS_LD);
+  LD->in_print_message--;
 
-  PL_discard_foreign_frame(fid);
-  unblockGC(PASS_LD1);
+  return rc;
 }
 
 
@@ -645,19 +762,7 @@ printMessage(atom_t severity, ...)
 		 *******************************/
 
 int
-PL_get_nchars_ex(term_t t, size_t *len, char **s, unsigned int flags)
-{ return PL_get_nchars(t, len, s, flags|CVT_EXCEPTION);
-}
-
-
-int
-PL_get_chars_ex(term_t t, char **s, unsigned int flags)
-{ return PL_get_nchars(t, NULL, s, flags|CVT_EXCEPTION);
-}
-
-
-int
-PL_get_atom_ex(term_t t, atom_t *a)
+PL_get_atom_ex__LD(term_t t, atom_t *a ARG_LD)
 { if ( PL_get_atom(t, a) )
     succeed;
 
@@ -665,9 +770,21 @@ PL_get_atom_ex(term_t t, atom_t *a)
 }
 
 
+#undef PL_get_atom_ex
+int
+PL_get_atom_ex(term_t t, atom_t *a)
+{ GET_LD
+
+  return PL_get_atom_ex__LD(t, a PASS_LD);
+}
+#define PL_get_atom_ex(t, a)	PL_get_atom_ex__LD(t, a PASS_LD)
+
+
 int
 PL_get_integer_ex(term_t t, int *i)
-{ if ( PL_get_integer(t, i) )
+{ GET_LD
+
+  if ( PL_get_integer(t, i) )
     succeed;
 
   if ( PL_is_integer(t) )
@@ -679,7 +796,9 @@ PL_get_integer_ex(term_t t, int *i)
 
 int
 PL_get_long_ex(term_t t, long *i)
-{ if ( PL_get_long(t, i) )
+{ GET_LD
+
+  if ( PL_get_long(t, i) )
     succeed;
 
   if ( PL_is_integer(t) )
@@ -691,7 +810,9 @@ PL_get_long_ex(term_t t, long *i)
 
 int
 PL_get_int64_ex(term_t t, int64_t *i)
-{ if ( PL_get_int64(t, i) )
+{ GET_LD
+
+  if ( PL_get_int64(t, i) )
     succeed;
 
   if ( PL_is_integer(t) )
@@ -728,7 +849,7 @@ PL_get_size_ex(term_t t, size_t *i)
 #endif
 #endif
 
-  *i = val;
+  *i = (size_t)val;
 
   return TRUE;
 }
@@ -761,9 +882,21 @@ PL_get_char_ex(term_t t, int *p, int eof)
 }
 
 
+int
+PL_get_pointer_ex(term_t t, void **addrp)
+{ GET_LD
+  if ( PL_get_pointer(t, addrp) )
+    succeed;
+
+  return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_address, t);
+}
+
+
 int
 PL_unify_list_ex(term_t l, term_t h, term_t t)
-{ if ( PL_unify_list(l, h, t) )
+{ GET_LD
+
+  if ( PL_unify_list(l, h, t) )
     succeed;
 
   if ( PL_get_nil(l) )
@@ -787,7 +920,9 @@ PL_unify_nil_ex(term_t l)
 
 int
 PL_get_list_ex(term_t l, term_t h, term_t t)
-{ if ( PL_get_list(l, h, t) )
+{ GET_LD
+
+  if ( PL_get_list(l, h, t) )
     succeed;
 
   if ( PL_get_nil(l) )
@@ -808,8 +943,9 @@ PL_get_nil_ex(term_t l)
 }
 
 int
-PL_unify_bool_ex(term_t t, bool val)
-{ bool v;
+PL_unify_bool_ex(term_t t, int val)
+{ GET_LD
+  bool v;
 
   if ( PL_is_variable(t) )
     return PL_unify_atom(t, val ? ATOM_true : ATOM_false);
@@ -825,9 +961,11 @@ PL_unify_bool_ex(term_t t, bool val)
 
 int
 PL_get_arg_ex(int n, term_t term, term_t arg)
-{ if ( PL_get_arg(n, term, arg) )
-    succeed;
-  else
+{ GET_LD
+
+  if ( PL_get_arg(n, term, arg) )
+  { succeed;
+  } else
   { term_t a = PL_new_term_ref();
 
     PL_put_integer(a, n);
