@@ -24,7 +24,7 @@
 
 /*#define O_DEBUG 1*/			/* include crash/0 */
 #include "pl-incl.h"
-#include "pl-ctype.h"
+#include "os/pl-ctype.h"
 
 #if O_DEBUG
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -69,8 +69,6 @@ static const PL_extension foreigns[] = {
 #if O_DEBUG
   FRG("crash",			0, pl_crash,			0),
 #endif
-  FRG("nl",			0, pl_nl,			ISO),
-  FRG("expand_file_name",	2, pl_expand_file_name,		0),
 #ifdef __WINDOWS__
   FRG("win_exec",		2, pl_win_exec,			0),
   FRG("win_module_file",	2, pl_win_module_file,		0),
@@ -80,7 +78,6 @@ static const PL_extension foreigns[] = {
   FRG("getenv",			2, pl_getenv,			0),
   FRG("setenv",			2, pl_setenv,			0),
   FRG("unsetenv",		1, pl_unsetenv,			0),
-  FRG("wildcard_match",		2, pl_wildcard_match,		0),
   FRG("$apropos_match",		2, pl_apropos_match,		0),
   FRG("sub_atom",		5, pl_sub_atom,		 NDET|ISO),
   FRG("sleep",			1, pl_sleep,			0),
@@ -89,8 +86,8 @@ static const PL_extension foreigns[] = {
   FRG("notrace",		1, pl_notrace1,		     META),
 
   FRG("write_canonical",	1, pl_write_canonical,	      ISO),
-  FRG("write_term",		2, pl_write_term,	      ISO),
-  FRG("write_term",		3, pl_write_term3,	      ISO),
+  FRG("write_term",		2, pl_write_term,	 META|ISO),
+  FRG("write_term",		3, pl_write_term3,	 META|ISO),
   FRG("write",			1, pl_write,		      ISO),
   FRG("writeq",			1, pl_writeq,		      ISO),
   FRG("print",			1, pl_print,			0),
@@ -100,23 +97,13 @@ static const PL_extension foreigns[] = {
   FRG("read",			1, pl_read,		      ISO),
   FRG("$raw_read",		1, pl_raw_read,			0),
   FRG("$raw_read",		2, pl_raw_read2,		0),
-  FRG("current_op",		3, pl_current_op,	NDET|META|ISO),
-  FRG("$local_op",		3, pl_local_op,	        NDET|META),
-  FRG("$builtin_op",		3, pl_builtin_op,	     NDET),
   FRG("current_functor",	2, pl_current_functor,	     NDET),
   FRG("$complete_atom",		3, pl_complete_atom,		0),
   FRG("$atom_completions",	2, pl_atom_completions,		0),
-  FRG("op",			3, pl_op,		     META|ISO),
   FRG("char_conversion",	2, pl_char_conversion,	      ISO),
   FRG("current_char_conversion",2, pl_current_char_conversion, NDET|ISO),
 
   FRG("!",			0, pl_metacut,		      ISO),
-  FRG("$e_free_variables",	2, pl_e_free_variables,		0),
-
-  FRG("$open_wic",		1, pl_open_wic,			0),
-  FRG("$close_wic",		0, pl_close_wic,		0),
-  FRG("$add_directive_wic",	1, pl_add_directive_wic,	0),
-  FRG("$import_wic",		2, pl_import_wic,		0),
 
   FRG("$rc_handle",		1, pl_rc_handle,		0),
   FRG("$rc_members",		2, pl_rc_members,		0),
@@ -126,34 +113,15 @@ static const PL_extension foreigns[] = {
   FRG("$rc_save_archive",	2, pl_rc_save_archive,		0),
   FRG("$rc_append_file",	5, pl_rc_append_file,		0),
 
-  FRG("$qlf_start_module",	1, pl_qlf_start_module,		0),
-  FRG("$qlf_start_sub_module",	1, pl_qlf_start_sub_module,	0),
-  FRG("$qlf_start_file",	1, pl_qlf_start_file,		0),
-  FRG("$qlf_end_part",		0, pl_qlf_end_part,		0),
-  FRG("$qlf_open",		1, pl_qlf_open,			0),
-  FRG("$qlf_close",		0, pl_qlf_close,		0),
-  FRG("$qlf_assert_clause",	2, pl_qlf_assert_clause,	0),
-
-  FRG("abolish",    		1, pl_abolish1,		     META|ISO),
-  FRG("abolish",    		2, pl_abolish,		     META),
-  FRG("clause",    		2, pl_clause2,	        NDET|META|CREF|ISO),
-  FRG("clause",    		3, pl_clause3,	        NDET|META|CREF),
-  FRG("$clause",	        4, pl_clause4,	        NDET|META|CREF),
-  FRG("nth_clause", 		3, pl_nth_clause,       NDET|META|CREF),
+  FRG("abolish",		1, pl_abolish1,		     META|ISO),
+  FRG("abolish",		2, pl_abolish,		     META),
+  FRG("nth_clause",		3, pl_nth_clause,       NDET|META|CREF),
   FRG("retractall",		1, pl_retractall,	 META|ISO),
 #ifdef O_MAINTENANCE
   FRG("$list_generations",	1, pl_list_generations,	     META),
   FRG("$check_procedure",	1, pl_check_procedure,	     META),
 #endif
 
-  FRG("recorda",		3, pl_recorda,			0),
-  FRG("recordz",		3, pl_recordz,			0),
-  FRG("recorded",		3, pl_recorded,		     NDET),
-  FRG("erase",			1, pl_erase,			0),
-  FRG("$term_complexity",	3, pl_term_complexity,		0),
-  FRG("redefine_system_predicate", 1, pl_redefine_system_predicate,
-							     META),
-
   FRG("$c_current_predicate",	2, pl_current_predicate,  NDET|META),
   FRG("current_predicate",	1, pl_current_predicate1, NDET|META|ISO),
   FRG("$set_predicate_attribute", 3, pl_set_predicate_attribute,META),
@@ -169,7 +137,6 @@ static const PL_extension foreigns[] = {
   FRG("fail",			0, pl_fail,		      ISO),
   FRG("true",			0, pl_true,		      ISO),
   FRG("$fail",			0, pl_fail,		  NOTRACE),
-  FRG("abort",			0, pl_abort,			0),
 
   FRG("trace",			0, pl_trace,		  NOTRACE),
   FRG("notrace",		0, pl_notrace,		  NOTRACE),
@@ -185,9 +152,6 @@ static const PL_extension foreigns[] = {
 #endif /* COUNTING */
 
   FRG("prolog_current_frame",	1, pl_prolog_current_frame,	0),
-  FRG("prolog_frame_attribute",	3, pl_prolog_frame_attribute,	0),
-  FRG("prolog_choice_attribute",3, pl_prolog_choice_attribute,	0),
-  FRG("prolog_skip_level",	2, pl_skip_level,	  NOTRACE),
 
   FRG("dwim_match",		3, pl_dwim_match,		0),
   FRG("$dwim_predicate",	2, pl_dwim_predicate,	     NDET),
@@ -225,8 +189,6 @@ static const PL_extension foreigns[] = {
   FRG("$atom_hashstat",		2, pl_atom_hashstat,		0),
   FRG("$current_prolog_flag",	5, pl_prolog_flag5,	     NDET),
   FRG("current_prolog_flag",	2, pl_prolog_flag,	 NDET|ISO),
-  FRG("set_prolog_flag",	2, pl_set_prolog_flag,	      ISO),
-  FRG("trim_stacks",		0, pl_trim_stacks,		0),
   FRG("$garbage_collect",	1, pl_garbage_collect,		0),
 #ifdef O_ATOMGC
   FRG("garbage_collect_atoms",	0, pl_garbage_collect_atoms,	0),
@@ -235,10 +197,8 @@ static const PL_extension foreigns[] = {
   FRG("track_atom",		2, pl_track_atom,		0),
 #endif
 #endif
-  FRG("current_key",		1, pl_current_key,	     NDET),
   FRG("current_flag",		1, pl_current_flag,	     NDET),
 
-  FRG("nl",			1, pl_nl1,		      ISO),
   FRG("read",			2, pl_read2,		      ISO),
   FRG("write",			2, pl_write2,		      ISO),
   FRG("writeq",			2, pl_writeq2,		      ISO),
@@ -254,10 +214,6 @@ static const PL_extension foreigns[] = {
   FRG("current_format_predicate", 2, pl_current_format_predicate,
 						        META|NDET),
   FRG("get_time",		1, pl_get_time,			0),
-#if O_PROLOG_FUNCTIONS
-  FRG("current_arithmetic_function", 1, pl_current_arithmetic_function,
-							NDET|META),
-#endif
 
 #ifdef O_PLMT
   FRG("thread_create",		3, pl_thread_create,	 META|ISO),
@@ -420,6 +376,11 @@ DECL_PLIST(flag);
 DECL_PLIST(list);
 DECL_PLIST(module);
 DECL_PLIST(prims);
+DECL_PLIST(variant);
+DECL_PLIST(copyterm);
+DECL_PLIST(prologflag);
+DECL_PLIST(trace);
+DECL_PLIST(pro);
 DECL_PLIST(read);
 DECL_PLIST(thread);
 DECL_PLIST(profile);
@@ -429,6 +390,7 @@ DECL_PLIST(gvar);
 DECL_PLIST(win);
 DECL_PLIST(file);
 DECL_PLIST(files);
+DECL_PLIST(glob);
 DECL_PLIST(btree);
 DECL_PLIST(ctype);
 DECL_PLIST(tai);
@@ -438,6 +400,9 @@ DECL_PLIST(proc);
 DECL_PLIST(write);
 DECL_PLIST(dlopen);
 DECL_PLIST(system);
+DECL_PLIST(op);
+DECL_PLIST(rec);
+DECL_PLIST(termhash);
 
 void
 initBuildIns(void)
@@ -453,12 +418,18 @@ initBuildIns(void)
   REG_PLIST(list);
   REG_PLIST(module);
   REG_PLIST(prims);
+  REG_PLIST(variant);
+  REG_PLIST(copyterm);
+  REG_PLIST(prologflag);
+  REG_PLIST(trace);
+  REG_PLIST(pro);
   REG_PLIST(read);
   REG_PLIST(thread);
   REG_PLIST(profile);
   REG_PLIST(wic);
   REG_PLIST(file);
   REG_PLIST(files);
+  REG_PLIST(glob);
   REG_PLIST(btree);
   REG_PLIST(ctype);
   REG_PLIST(tai);
@@ -468,6 +439,9 @@ initBuildIns(void)
   REG_PLIST(write);
   REG_PLIST(dlopen);
   REG_PLIST(system);
+  REG_PLIST(op);
+  REG_PLIST(rec);
+  REG_PLIST(termhash);
 #ifdef O_ATTVAR
   REG_PLIST(attvar);
 #endif
@@ -482,13 +456,13 @@ initBuildIns(void)
 	GD->procedures.name = lookupProcedure(FUNCTOR_ ## name, m);
 
   LOOKUPPROC(dgarbage_collect1);
-  LOOKUPPROC(block3);
   LOOKUPPROC(catch3);
   LOOKUPPROC(true0);
   LOOKUPPROC(fail0);
   LOOKUPPROC(equals2);
   LOOKUPPROC(is2);
   LOOKUPPROC(strict_equal2);
+  LOOKUPPROC(not_strict_equal2);
   LOOKUPPROC(print_message2);
   LOOKUPPROC(dcall1);
   LOOKUPPROC(setup_call_catcher_cleanup4);
@@ -505,7 +479,25 @@ initBuildIns(void)
 	PL_predicate("prolog_exception_hook", 4, "user");
 					/* allow debugging in call/1 */
   clear(PROCEDURE_dcall1->definition, HIDE_CHILDS|TRACE_ME);
-  set(PROCEDURE_dcall1->definition, DYNAMIC);
+  set(PROCEDURE_dcall1->definition, DYNAMIC|SYSTEM);
+
+  PL_meta_predicate(PL_predicate("assert",         1, "system"), MA_META);
+  PL_meta_predicate(PL_predicate("asserta",        1, "system"), MA_META);
+  PL_meta_predicate(PL_predicate("assertz",        1, "system"), MA_META);
+  PL_meta_predicate(PL_predicate("assert",         2, "system"), MA_META, MA_VAR);
+  PL_meta_predicate(PL_predicate("asserta",        2, "system"), MA_META, MA_VAR);
+  PL_meta_predicate(PL_predicate("assertz",        2, "system"), MA_META, MA_VAR);
+  PL_meta_predicate(PL_predicate("retract",        1, "system"), MA_META);
+  PL_meta_predicate(PL_predicate("retractall",     1, "system"), MA_META);
+
+  PL_meta_predicate(PL_predicate("notrace",        1, "system"), 0);
+  PL_meta_predicate(PL_predicate("with_mutex",     2, "system"), MA_NONVAR, 0);
+  PL_meta_predicate(PL_predicate("with_output_to", 2, "system"), MA_NONVAR, 0);
+#ifdef O_PLMT
+  PL_meta_predicate(PL_predicate("thread_create",  3, "system"), 0, MA_ANY, MA_NONVAR);
+  PL_meta_predicate(PL_predicate("thread_at_exit", 1, "system"), 0);
+  PL_meta_predicate(PL_predicate("thread_signal",  2, "system"), MA_NONVAR, 0);
+#endif
 
   for( ecell = ext_head; ecell; ecell = ecell->next )
     bindExtensions(ecell->module, ecell->extensions);
