@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2008, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -65,9 +65,10 @@ typedef struct
 
 typedef struct
 { atom_t	file;			/* current source file */
-  int	  	line;			/* current line */
+  int		line;			/* current line */
   int		linepos;		/* position in the line */
   int64_t	character;		/* current character location */
+  int64_t	byte;			/* byte offset of location */
 } source_location;
 
 		 /*******************************
@@ -75,11 +76,9 @@ typedef struct
 		 *******************************/
 
 struct PL_global_data
-{ char *top_of_heap;			/* highest allocated heap address */
-  char *base_of_heap;			/* lowest allocated heap address */
-  uintptr_t rounded_heap_base;		/* heap-base rounded downwards */
-  pl_defaults_t	    defaults;		/* system default settings */
-  pl_options_t	    options;		/* command-line options */
+{ uintptr_t	heap_base;		/* heap-base rounded downwards */
+  pl_defaults_t	defaults;		/* system default settings */
+  pl_options_t	options;		/* command-line options */
   State		stateList;		/* list of loaded states */
   int		initialised;		/* Heap is initialised */
   int		io_initialised;		/* I/O system has been initialised */
@@ -92,7 +91,7 @@ struct PL_global_data
   sig_handler sig_handlers[MAXSIGNAL];	/* How Prolog preceives signals */
 #endif
 #ifdef O_LOGICAL_UPDATE
-  uintptr_t generation;		/* generation of the database */
+  uintptr_t generation;			/* generation of the database */
 #endif
 
   struct
@@ -115,6 +114,7 @@ struct PL_global_data
   { size_t	heap;			/* heap in use */
     size_t	atoms;			/* No. of atoms defined */
     size_t	atomspace;		/* # bytes used to store atoms */
+    size_t	stack_space;		/* # bytes on stacks */
 #ifdef O_ATOMGC
     size_t	atomspacefreed;		/* Freed atom-space */
 #endif
@@ -122,6 +122,9 @@ struct PL_global_data
     int		predicates;		/* No. of predicates defined */
     int		modules;		/* No. of modules in the system */
     intptr_t	codes;			/* No. of byte codes generated */
+    double	start_time;		/* When Prolog was started */
+    double	user_cputime;		/* User CPU time (whole process) */
+    double	system_cputime;		/* Kernel CPU time (whole process) */
 #ifdef O_PLMT
     int		threads_created;	/* # threads created */
     int		threads_finished;	/* # finished threads */
@@ -136,12 +139,17 @@ struct PL_global_data
 
   struct
   { Table	modules;		/* atom --> module */
-    Table	record_lists;		/* Available record lists */
   } tables;
 
   struct
-  { buffer	functions;		/* index --> function */
-    ArithFunction table[ARITHHASHSIZE];	/* functor --> function table */
+  { Table	record_lists;		/* Available record lists */
+    RecordList	head;			/* first record list */
+    RecordList	tail;			/* last record list */
+  } recorded_db;
+
+  struct
+  { ArithF     *functions;		/* index --> function */
+    size_t	functions_allocated;	/* Size of above array */
   } arith;
 
   struct
@@ -160,6 +168,7 @@ struct PL_global_data
     size_t	non_garbage;		/* # atoms for after last AGC */
     int		gc;			/* # atom garbage collections */
     int64_t	collected;		/* # collected atoms */
+    size_t	unregistered;		/* # candidate GC atoms */
     double	gc_time;		/* Time spent on atom-gc */
     PL_agc_hook_t gc_hook;		/* Current hook */
 #endif
@@ -184,6 +193,7 @@ struct PL_global_data
 
     InitialiseHandle initialise_head;	/* PL_initialise_hook() */
     InitialiseHandle initialise_tail;
+    PL_dispatch_hook_t dispatch_events; /* PL_dispatch_hook() */
 
     int		  _loaded;		/* system extensions are loaded */
   } foreign;
@@ -218,8 +228,7 @@ struct PL_global_data
   } exceptions;
 
   struct
-  { TempFile		_tmpfile_head;
-    TempFile		_tmpfile_tail;
+  { Table		tmp_files;	/* Known temporary files */
     CanonicalDir	_canonical_dirlist;
     char *		myhome;		/* expansion of ~ */
     char *		fred;		/* last expanded ~user */
@@ -228,19 +237,19 @@ struct PL_global_data
     int			halting;	/* process is shutting down */
     int			gui_app;	/* Win32: Application is a gui app */
     IOFUNCTIONS		iofunctions;	/* initial IO functions */
-    IOFUNCTIONS 	org_terminal;	/* IO+Prolog terminal functions */
+    IOFUNCTIONS		org_terminal;	/* IO+Prolog terminal functions */
     IOFUNCTIONS		rl_functions;	/* IO+Terminal+Readline functions */
   } os;
 
   struct
   { Procedure	dgarbage_collect1;
-    Procedure 	block3;
     Procedure	catch3;
     Procedure	true0;
     Procedure	fail0;
     Procedure	equals2;		/* =/2 */
     Procedure	is2;			/* is/2 */
     Procedure	strict_equal2;		/* ==/2 */
+    Procedure	not_strict_equal2;	/* \==/2 */
     Procedure	event_hook1;
     Procedure	exception_hook4;
     Procedure	print_message2;
@@ -248,6 +257,7 @@ struct PL_global_data
     Procedure	prolog_trace_interception4;
     Procedure	portray;		/* portray/1 */
     Procedure   dcall1;			/* $call/1 */
+    Procedure   call3;			/* call/3*/
     Procedure	setup_call_catcher_cleanup4; /* setup_call_catcher_cleanup/4 */
     Procedure	undefinterc4;		/* $undefined_procedure/4 */
     Procedure   dthread_init0;		/* $thread_init/0 */
@@ -290,14 +300,16 @@ struct PL_global_data
 
   struct
   { struct _at_exit_goal *exit_goals;	/* Global thread_at_exit/1 goals */
-    int		    	enabled;	/* threads are enabled */
+    int			enabled;	/* threads are enabled */
     Table		mutexTable;	/* Name --> mutex table */
     int			mutex_next_id;	/* next id for anonymous mutexes */
     struct pl_mutex*	MUTEX_load;	/* The $load mutex */
 #ifdef __WINDOWS__
-    HINSTANCE	    	instance;	/* Win32 process instance */
+    HINSTANCE		instance;	/* Win32 process instance */
 #endif
     counting_mutex     *mutexes;	/* Registered mutexes */
+    int			thread_max;	/* Maximum # threads */
+    PL_thread_info_t  **threads;	/* Pointers to thread-info */
   } thread;
 #endif /*O_PLMT*/
 };
@@ -314,28 +326,40 @@ struct PL_local_data
   LocalFrame    environment;		/* Current local frame */
   Choice	choicepoints;		/* Choice-point chain */
   FliFrame      foreign_environment;	/* Current foreign context */
+  QueryFrame    query;			/* Currently open query */
   Word		mark_bar;		/* Mark globals > this one */
 #ifdef O_GVAR
   Word		frozen_bar;		/* Frozen part of the global stack */
 #endif
   pl_stacks_t   stacks;			/* Prolog runtime stacks */
   uintptr_t	bases[STG_MASK+1];	/* area base addresses */
+#if defined(O_SECURE) || defined(SECURE_GC)
+  unsigned int  incr_seed;		/* Seed for random stack increments */
+#endif
   int		alerted;		/* Special mode. See updateAlerted() */
-  int64_t	pending_signals;	/* PL_raise() pending signals */
-  record_t	pending_exception;	/* Pending exception from signal */
-  int		current_signal;		/* Currently handled signal */
-  int		sync_signal;		/* Current signal is synchronous */
   int		critical;		/* heap is being modified */
-  int		aborted;		/* thread asked for abort */
+  abort_type	aborted;		/* !ABORT_NONE: abort in Critical */
   Stack		outofstack;		/* thread is out of stack */
   int		trim_stack_requested;	/* perform a trim-stack */
 #ifdef O_PLMT
   int		exit_requested;		/* Thread is asked to exit */
 #endif
   int		in_arithmetic;		/* doing arithmetic */
+  int		in_print_message;	/* Inside printMessage() */
   int		autoload_nesting;	/* Nesting level in autoloader */
   void *	glob_info;		/* pl-glob.c */
   IOENC		encoding;		/* default I/O encoding */
+  ClauseRef	freed_clauses;		/* List of pending freeable clauses */
+
+  struct
+  { int64_t	pending;		/* PL_raise() pending signals */
+    int		current;		/* currently processing signal */
+    int		is_sync;		/* current signal is synchronous */
+    record_t	exception;		/* Pending exception from signal */
+#ifdef O_PLMT
+    simpleMutex sig_lock;		/* lock delivery and processing */
+#endif
+  } signal;
 
   struct
   { int		active;			/* doing pipe I/O */
@@ -350,7 +374,6 @@ struct PL_local_data
   } prompt;
 
   source_location read_source;		/* file, line, char of last term */
-  const char   *float_format;		/* floating point format */
 
   struct
   { term_t	term;			/* exception term */
@@ -359,6 +382,7 @@ struct PL_local_data
     term_t	tmp;			/* tmp for errors */
     term_t	pending;		/* used by the debugger */
     int		in_hook;		/* inside exception_hook() */
+    int		processing;		/* processing an exception */
     exception_frame *throw_environment;	/* PL_throw() environments */
   } exception;
 
@@ -373,6 +397,11 @@ struct PL_local_data
   { term_t	dummy;			/* see trimStacks() */
   } trim;
 
+  struct
+  { term_t	h[TMP_PTR_SIZE];	/* temporary handles. See unify_ptrs */
+    int		top;			/* Top-of-stack index */
+  } tmp;
+
 #ifdef O_GVAR
   struct
   { Table	nb_vars;		/* atom --> value */
@@ -384,7 +413,8 @@ struct PL_local_data
   { int64_t	inferences;		/* inferences in this thread */
     uintptr_t	last_cputime;		/* milliseconds last CPU time */
     uintptr_t	last_systime;		/* milliseconds last SYSTEM time */
-    uintptr_t	last_walltime;		/* milliseconds last Wall time */
+    uintptr_t	last_real_time;		/* Last Real Time (seconds since Epoch) */
+    double	last_walltime;		/* Last Wall time (m-secs since start) */
     double	user_cputime;		/* User saved CPU time */
     double	system_cputime;		/* Kernel saved CPU time */
   } statistics;
@@ -420,7 +450,7 @@ struct PL_local_data
   } modules;
 
   struct
-  { intptr_t 	generator;		/* See PL_atom_generator() */
+  { intptr_t	generator;		/* See PL_atom_generator() */
   } atoms;
 
   struct
@@ -445,7 +475,8 @@ struct PL_local_data
 
 #if O_CYCLIC
   struct
-  { segstack stack;			/* stack to find cycles */
+  { segstack lstack;			/* Stack for cycle-links */
+    segstack vstack;			/* Stack for visited marks */
   } cycle;
 #endif
 
@@ -479,19 +510,24 @@ struct PL_local_data
   { AbortHandle	_abort_head;		/* PL_abort_hook() */
     AbortHandle _abort_tail;
 
-    PL_dispatch_hook_t _dispatch_events; /* PL_dispatch_hook() */
-
     buffer	_discardable_buffer;	/* PL_*() character buffers */
     buffer	_buffer_ring[BUFFER_RING_SIZE];
     int		_current_buffer_id;
+
+    int		SP_state;		/* For SICStus interface */
   } fli;
 
   struct				/* Local IO stuff */
   { IOSTREAM *streams[6];		/* handles for standard streams */
     struct input_context *input_stack;	/* maintain input stream info */
     struct output_context *output_stack; /* maintain output stream info */
+    st_check stream_type_check;		/* Check bin/text streams? */
   } IO;
 
+  struct
+  { fid_t	numbervars_frame;	/* Numbervars choice-point */
+  } var_names;
+
 #ifdef O_LIMIT_DEPTH
   struct
   { uintptr_t limit;
@@ -500,24 +536,27 @@ struct PL_local_data
 #endif
 
   struct
-  { intptr_t _total_marked;			/* # marked global cells */
-    intptr_t _trailcells_deleted;		/* # garbage trailcells */
-    intptr_t _relocation_chains;		/* # relocation chains (debugging) */
+  { intptr_t _total_marked;		/* # marked global cells */
+    intptr_t _trailcells_deleted;	/* # garbage trailcells */
+    intptr_t _relocation_chains;	/* # relocation chains (debugging) */
     intptr_t _relocation_cells;		/* # relocation cells */
     intptr_t _relocated_cells;		/* # relocated cells */
     intptr_t _needs_relocation;		/* # cells that need relocation */
-    intptr_t _local_marked;			/* # marked local -> global ptrs */
-    intptr_t _marks_swept;			/* # marks swept */
+    intptr_t _local_marked;		/* # marked local -> global ptrs */
+    intptr_t _marks_swept;		/* # marks swept */
     intptr_t _marks_unswept;		/* # marks swept */
-    intptr_t _alien_relocations;		/* # alien_into_relocation_chain() */
-    intptr_t _local_frames;			/* frame count for debugging */
-    intptr_t _choice_count;			/* choice-point count for debugging */
+    intptr_t _alien_relocations;	/* # alien_into_relocation_chain() */
+    intptr_t _local_frames;		/* frame count for debugging */
+    intptr_t _choice_count;		/* choice-point count for debugging */
+    int  *_start_map;			/* bitmap with legal global starts */
+    sigset_t saved_sigmask;		/* Saved signal mask */
 #if defined(O_SECURE) || defined(SECURE_GC)
     intptr_t _trailtops_marked;		/* # marked trailtops */
     Word *_mark_base;			/* Array of marked cells addresses */
     Word *_mark_top;			/* Top of this array */
     Table _check_table;			/* relocation address table */
     Table _local_table;			/* marked local variables */
+    int  _relocated_check;		/* Verify relocated addresses? */
 #endif
     int64_t inferences;			/* #inferences at last GC */
 
@@ -525,13 +564,12 @@ struct PL_local_data
 #ifdef O_CALL_RESIDUE
     int			marked_attvars;	/* do not GC attvars */
 #endif
+    int active;				/* GC is running in this thread */
   } gc;
 
-#ifdef O_SHIFT_STACKS
-  pl_shift_status_t	shift_status;	/* Stack shifter status */
-#endif
-
+  pl_shift_status_t shift_status;	/* Stack shifter status */
   pl_debugstatus_t _debugstatus;	/* status of the debugger */
+  struct btrace *btrace_store;		/* C-backtraces */
 
 #ifdef O_PLMT
   struct
@@ -557,31 +595,19 @@ GLOBAL PL_local_data_t  PL_local_data;
 GLOBAL PL_local_data_t *PL_current_engine_ptr;
 #endif
 
-#if !defined(O_PLMT) && !defined(O_MULTIPLE_ENGINES)
-#define GET_LD
-#define ARG_LD
-#define ARG1_LD void
-#define PASS_LD
-#define PASS_LD1
-#define LOCAL_LD  (&PL_local_data)
-#define GLOBAL_LD (&PL_local_data)
-#define LD	  GLOBAL_LD
-#endif /*O_PLMT && O_MULTIPLE_ENGINES*/
 #define GD (&PL_global_data)
 #define CD (&PL_code_data)
 
-#define hTop			(GD->top_of_heap)
-#define hBase			(GD->base_of_heap)
-#define heap_base		(GD->rounded_heap_base)
 #define functor_array		(GD->functors.array)
 #define systemDefaults		(GD->defaults)
 
-#define environment_frame 	(LD->environment)
-#define fli_context	  	(LD->foreign_environment)
+#define environment_frame	(LD->environment)
+#define fli_context		(LD->foreign_environment)
 #define source_file_name	(LD->read_source.file)
 #define source_line_no		(LD->read_source.line)
 #define source_line_pos		(LD->read_source.linepos)
 #define source_char_no		(LD->read_source.character)
+#define source_byte_no		(LD->read_source.byte)
 #define exception_term		(LD->exception.term)
 #define exception_bin		(LD->exception.bin)
 #define exception_printed	(LD->exception.printed)
@@ -596,7 +622,7 @@ GLOBAL PL_local_data_t *PL_current_engine_ptr;
 #define Scurin			(LD->IO.streams[3])
 #define Scurout			(LD->IO.streams[4])
 #define Sprotocol		(LD->IO.streams[5])
-#define Sdin			Suser_input 		/* not used for now */
+#define Sdin			Suser_input		/* not used for now */
 #define Sdout			Suser_output
 
 #ifdef VMCODE_IS_ADDRESS
