@@ -1,11 +1,11 @@
-/*  $Id: pl-gvar.c,v 1.5 2007/01/24 14:29:16 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        jan@swi.psy.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2002, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -26,9 +26,6 @@
 #include "pl-incl.h"
 #ifdef O_GVAR
 
-#undef LD
-#define LD LOCAL_LD
-
 
 		 /*******************************
 		 * NON-BACKTRACKABLE GLOBAL VARS*
@@ -36,7 +33,7 @@
 
 void
 freezeGlobal(ARG1_LD)
-{ LD->frozen_bar = LD->mark_bar = gTop;  
+{ LD->frozen_bar = LD->mark_bar = gTop;
   DEBUG(2, Sdprintf("*** frozen bar to %p at freezeGlobal()\n",
 		    LD->frozen_bar));
 }
@@ -76,6 +73,9 @@ Assign  a  global  variable.  For    backtrackable   variables  we  need
 TrailAssignment(), but we can only call that  on addresses on the global
 stack. Therefore we must make  a  reference   to  the  real value if the
 variable is not already a reference.
+
+SHIFT-SAFE: TrailAssignment() takes at most g+t=1+2.  One more Trail and
+	    2 more allocGlobal(1) makes g+t<3+3
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 static int
@@ -93,7 +93,13 @@ setval(term_t var, term_t value, int backtrackable ARG_LD)
     LD->gvar.nb_vars->free_symbol = free_nb_linkval_symbol;
   }
 
-  requireStack(global, sizeof(word));
+  if ( !hasGlobalSpace(3) )		/* also ensures trail for */
+  { int rc;				/* TrailAssignment() */
+
+    if ( (rc=ensureGlobalSpace(3, ALLOW_GC)) != TRUE )
+      return raiseStackOverflow(rc);
+  }
+
   p = valTermRef(value);
   deRef(p);
   w = *p;
@@ -104,7 +110,7 @@ setval(term_t var, term_t value, int backtrackable ARG_LD)
 
       setVar(*p2);
       w = *p = makeRef(p2);
-      Trail(p);
+      LTrail(p);
     } else
     { w = makeRef(p);
     }
@@ -124,21 +130,21 @@ setval(term_t var, term_t value, int backtrackable ARG_LD)
     PL_unregister_atom(old);
 
   if ( backtrackable )
-  { if ( isRef(old) )
-    { Word p = unRef(old);
+  { Word p;
 
-      TrailAssignment(p);
-      *p = w;
+    if ( isRef(old) )
+    { p = unRef(old);
     } else
-    { Word p = allocGlobal(1);
+    { p = allocGlobal(1);
       *p = old;
-      freezeGlobal(PASS_LD1);		/* Why is this? */
-      if ( storage(old) != STG_GLOBAL )
+      freezeGlobal(PASS_LD1);		/* The value location must be */
+      if ( storage(old) != STG_GLOBAL )	/* preserved */
 	LD->gvar.grefs++;
       s->value = (void*)makeRefG(p);
-      TrailAssignment(p);
-      *p = w;
     }
+
+    TrailAssignment(p);
+    *p = w;
   } else
   { if ( storage(old) == STG_GLOBAL )
       LD->gvar.grefs--;
@@ -167,13 +173,16 @@ static gvar_action
 auto_define_gvar(atom_t name)
 { GET_LD
   static predicate_t pred;
-  fid_t fid = PL_open_foreign_frame();
-  term_t av = PL_new_term_refs(3);
+  fid_t fid;
+  term_t av;
   gvar_action rc = gvar_error;
 
   if ( !pred )
     pred = PL_predicate("exception", 3, "user");
 
+  if ( !(fid = PL_open_foreign_frame()) )
+    return gvar_error;
+  av = PL_new_term_refs(3);
   PL_put_atom(av+0, ATOM_undefined_global_variable);
   PL_put_atom(av+1, name);
 
@@ -202,15 +211,24 @@ getval(term_t var, term_t value ARG_LD)
 
   if ( !PL_get_atom_ex(var, &name) )
     fail;
+  if ( !hasGlobalSpace(0) )
+  { int rc;
+
+    if ( (rc=ensureGlobalSpace(0, ALLOW_GC)) != TRUE )
+      return raiseStackOverflow(rc);
+  }
+
 
   for(i=0; i<2; i++)
   { if ( LD->gvar.nb_vars )
     { Symbol s = lookupHTable(LD->gvar.nb_vars, (void*)name);
-    
+
       if ( s )
-      { word w = (word)s->value;
-	
-	return unify_ptrs(valTermRef(value), &w PASS_LD);
+      { term_t tmp = PL_new_term_ref();
+	word w = (word)s->value;
+
+	*valTermRef(tmp) = w;
+	return PL_unify(value, tmp);
       }
     }
 
@@ -243,7 +261,7 @@ PRED_IMPL("nb_linkval", 2, nb_linkval, 0)
 static
 PRED_IMPL("nb_getval", 2, nb_getval, 0)
 { PRED_LD
-  
+
   return getval(A1, A2 PASS_LD);
 }
 
@@ -258,7 +276,7 @@ PRED_IMPL("b_setval", 2, b_setval, 0)
 static
 PRED_IMPL("b_getval", 2, b_getval, 0)
 { PRED_LD
-  
+
   return getval(A1, A2 PASS_LD);
 }
 
@@ -273,7 +291,7 @@ PRED_IMPL("nb_delete", 1, nb_delete, 0)
 
   if ( LD->gvar.nb_vars )
   { Symbol s = lookupHTable(LD->gvar.nb_vars, (void*)name);
-    
+
     if ( s )
     { free_nb_linkval_symbol(s);
       deleteSymbolHTable(LD->gvar.nb_vars, s);
@@ -311,13 +329,16 @@ PRED_IMPL("nb_current", 2, nb_current, PL_FA_NONDETERMINISTIC)
       fail;
   }
 
-  fid = PL_open_foreign_frame();
+  if ( !(fid = PL_open_foreign_frame()) )
+  { freeTableEnum(e);
+    return FALSE;
+  }
   while( (s=advanceTableEnum(e)) )
   { atom_t name = (atom_t)s->name;
     word   val = (word)s->value;
 
     if ( PL_unify_atom(A1, name) &&
-	 unify_ptrs(valTermRef(A2), &val PASS_LD) )
+	 unify_ptrs(valTermRef(A2), &val, 0 PASS_LD) )
     { PL_close_foreign_frame(fid);
       ForeignRedoPtr(e);
     } else
