@@ -32,20 +32,31 @@
 #include <winsock2.h>
 #endif
 
+#ifdef __MINGW32__
+#define _WIN32_IE 0x0400
+#include <shlobj.h>
+/* FIXME: these are copied from SWI-Prolog.h. */
+#define PL_MSG_EXCEPTION_RAISED -1
+#define PL_MSG_IGNORED 0
+#define PL_MSG_HANDLED 1
+#endif
+
 #include "pl-incl.h"
-#include "pl-utf8.h"
-#include <crtdbg.h>
+#include "os/pl-utf8.h"
 #include <process.h>
-#include "pl-ctype.h"
+#include "os/pl-ctype.h"
 #include <stdio.h>
 #include <stdarg.h>
 #include "SWI-Stream.h"
 #include <process.h>
 #include <winbase.h>
+#ifdef HAVE_CRTDBG_H
+#include <crtdbg.h>
+#endif
 
 
 		 /*******************************
-		 *	    MESSAGE BOX		*
+		 *	       CONSOLE		*
 		 *******************************/
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -73,6 +84,38 @@ hasConsole(void)
 }
 
 
+int
+PL_wait_for_console_input(void *handle)
+{ BOOL rc;
+  HANDLE hConsole = handle;
+
+  for(;;)
+  { rc = MsgWaitForMultipleObjects(1,
+				   &hConsole,
+				   FALSE,	/* wait for either event */
+				   INFINITE,
+				   QS_ALLINPUT);
+
+    if ( rc == WAIT_OBJECT_0+1 )
+    { MSG msg;
+
+      while( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
+      { TranslateMessage(&msg);
+	DispatchMessage(&msg);
+      }
+    } else if ( rc == WAIT_OBJECT_0 )
+    { return TRUE;
+    } else
+    { Sdprintf("MsgWaitForMultipleObjects(): 0x%x\n", rc);
+    }
+  }
+}
+
+
+		 /*******************************
+		 *	    MESSAGE BOX		*
+		 *******************************/
+
 void
 PlMessage(const char *fm, ...)
 { va_list(args);
@@ -104,7 +147,7 @@ WinError()
 { int id = GetLastError();
   char *msg;
   static WORD lang;
-  static lang_initialised = 0;
+  static int lang_initialised = 0;
 
   if ( !lang_initialised )
     lang = MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_UK);
@@ -191,6 +234,27 @@ Pause(double t)
   }
 }
 
+
+		 /*******************************
+		 *	  SET FILE SIZE		*
+		 *******************************/
+
+#ifndef HAVE_FTRUNCATE
+
+int
+ftruncate(int fileno, int64_t length)
+{ errno_t e;
+
+  if ( (e=_chsize_s(fileno, length)) == 0 )
+    return 0;
+
+  errno = e;
+  return -1;
+}
+
+#endif
+
+
 		 /*******************************
 		 *	 QUERY CPU TIME		*
 		 *******************************/
@@ -214,6 +278,9 @@ CpuTime(cputime_kind which)
       case CPU_SYSTEM:
 	p = &kerneltime;
         break;
+      default:
+	assert(0);
+        return 0.0;
     }
     t = (double)p->dwHighDateTime * (4294967296.0 * ntick nano);
     t += (double)p->dwLowDateTime  * (ntick nano);
@@ -289,12 +356,12 @@ get_showCmd(term_t show, int *cmd)
 { char *s;
   showtype *st;
   static showtype types[] =
-  { { "hide", 		 SW_HIDE },
-    { "maximize", 	 SW_MAXIMIZE },
-    { "minimize", 	 SW_MINIMIZE },
-    { "restore", 	 SW_RESTORE },
-    { "show", 		 SW_SHOW },
-    { "showdefault", 	 SW_SHOWDEFAULT },
+  { { "hide",		 SW_HIDE },
+    { "maximize",	 SW_MAXIMIZE },
+    { "minimize",	 SW_MINIMIZE },
+    { "restore",	 SW_RESTORE },
+    { "show",		 SW_SHOW },
+    { "showdefault",	 SW_SHOWDEFAULT },
     { "showmaximized",   SW_SHOWMAXIMIZED },
     { "showminimized",   SW_SHOWMINIMIZED },
     { "showminnoactive", SW_SHOWMINNOACTIVE },
@@ -302,8 +369,8 @@ get_showCmd(term_t show, int *cmd)
     { "shownoactive",    SW_SHOWNOACTIVATE },
     { "shownormal",      SW_SHOWNORMAL },
 					/* compatibility */
-    { "normal", 	 SW_SHOWNORMAL },
-    { "iconic", 	 SW_MINIMIZE },
+    { "normal",		 SW_SHOWNORMAL },
+    { "iconic",		 SW_MINIMIZE },
     { NULL, 0 },
   };
 
@@ -312,7 +379,7 @@ get_showCmd(term_t show, int *cmd)
     succeed;
   }
 
-  if ( !PL_get_chars_ex(show, &s, CVT_ATOM) )
+  if ( !PL_get_chars(show, &s, CVT_ATOM|CVT_EXCEPTION) )
     fail;
   for(st=types; st->name; st++)
   { if ( streq(st->name, s) )
@@ -329,7 +396,8 @@ get_showCmd(term_t show, int *cmd)
 
 static int
 win_exec(size_t len, const wchar_t *cmd, UINT show)
-{ STARTUPINFOW startup;
+{ GET_LD
+  STARTUPINFOW startup;
   PROCESS_INFORMATION info;
   int rval;
   wchar_t *wcmd;
@@ -362,8 +430,9 @@ win_exec(size_t len, const wchar_t *cmd, UINT show)
   } else
   { term_t tmp = PL_new_term_ref();
 
-    PL_unify_wchars(tmp, PL_ATOM, len, cmd);
-    return PL_error(NULL, 0, WinError(), ERR_SHELL_FAILED, tmp);
+    return ( PL_unify_wchars(tmp, PL_ATOM, len, cmd) &&
+	     PL_error(NULL, 0, WinError(), ERR_SHELL_FAILED, tmp)
+	   );
   }
 }
 
@@ -464,7 +533,7 @@ static const shell_error se_errors[] =
   { SE_ERR_DDETIMEOUT,	    "DDE request timed out" },
   { SE_ERR_DLLNOTFOUND,	    "DLL not found" },
   { SE_ERR_FNF,		    "File not found (FNF)" },
-  { SE_ERR_NOASSOC, 	    "No association" },
+  { SE_ERR_NOASSOC,	    "No association" },
   { SE_ERR_OOM,		    "Not enough memory" },
   { SE_ERR_PNF,		    "Path not found (PNF)" },
   { SE_ERR_SHARE,	    "Sharing violation" },
@@ -518,7 +587,7 @@ pl_win_module_file(term_t module, term_t file)
   char *m;
   char *f;
 
-  if ( !PL_get_chars_ex(module, &m, CVT_ALL) )
+  if ( !PL_get_chars(module, &m, CVT_ALL|CVT_EXCEPTION) )
     fail;
   if ( (f = findExecutable(m, buf)) )
     return PL_unify_atom_chars(file, f);
@@ -620,7 +689,9 @@ dlclose(void *handle)
 		 *	      FOLDERS		*
 		 *******************************/
 
+#ifdef HAVE_SHLOBJ_H
 #include <Shlobj.h>
+#endif
 
 typedef struct folderid
 { int csidl;
@@ -683,7 +754,8 @@ unify_csidl_path(term_t t, int csidl)
 
 static
 PRED_IMPL("win_folder", 2, win_folder, PL_FA_NONDETERMINISTIC)
-{ int n;
+{ GET_LD
+  int n;
 
   switch( CTX_CNTRL )
   { case FRG_FIRST_CALL:
@@ -798,7 +870,8 @@ reg_open_key(const wchar_t *which, int create)
 
 static
 PRED_IMPL("win_registry_get_value", 3, win_registry_get_value, 0)
-{ DWORD type;
+{ GET_LD
+  DWORD type;
   BYTE  data[MAXREGSTRLEN];
   DWORD len = sizeof(data);
   size_t klen, namlen;
@@ -849,7 +922,6 @@ static struct regdef
 { { "localSize",    &GD->defaults.local },
   { "globalSize",   &GD->defaults.global },
   { "trailSize",    &GD->defaults.trail },
-  { "argumentSize", &GD->defaults.argument },
   { NULL,           NULL }
 };
 
@@ -885,11 +957,6 @@ getDefaultsFromRegistry()
   { setStacksFromKey(key);
     RegCloseKey(key);
   }
-
-#ifdef O_CYCLIC
-  if ( GD->defaults.argument == DEFARGUMENT )
-    GD->defaults.argument = GD->defaults.global/4;
-#endif
 }
 
 		 /*******************************
