@@ -38,7 +38,7 @@
 #include <pthread.h>
 #endif
 
-#include "pl-utf8.c"			/* we're not in the libpl.dll module */
+#include "os/pl-utf8.c"			/* we're not in the libpl.dll module */
 
 #ifndef streq
 #define streq(s,q) (strcmp((s), (q)) == 0)
@@ -50,7 +50,7 @@ X11-xterm like features: scrollback for a   predefined  number of lines,
 cut/paste and the GNU readline library for command-line editing.
 
 This module combines libpl.dll and plterm.dll  with some glue to produce
-the final executable plwin.exe.
+the final executable swipl-win.exe.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 __declspec(dllexport) rlc_console	PL_current_console();
@@ -178,7 +178,7 @@ Srlc_read(void *handle, char *buffer, size_t size)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-The user streams for plwin.exe  run   using  the  'wchar' encoding. This
+The user streams for swipl-win.exe run  using the 'wchar' encoding. This
 means size must be a multiple of   sizeof(wchar_t),  but not if the user
 cheats and either switches the encoding   or uses put_byte/1 or similar.
 The flushing code will remember `half'   characters  and re-send them as
@@ -273,6 +273,10 @@ rlc_bind_terminal(rlc_console c)
   Sinput->encoding  = ENC_WCHAR;
   Soutput->encoding = ENC_WCHAR;
   Serror->encoding  = ENC_WCHAR;
+
+  Sinput->flags  &= ~SIO_FILE;
+  Soutput->flags &= ~SIO_FILE;
+  Serror->flags  &= ~SIO_FILE;
 }
 
 
@@ -335,7 +339,7 @@ pl_win_open_console(term_t title, term_t input, term_t output, term_t error,
   create_prolog_hidden_window(c);	/* for sending messages */
   registerConsole(c);
 
-#define STREAM_COMMON (SIO_TEXT|	/* text-stream */ 		\
+#define STREAM_COMMON (SIO_TEXT|	/* text-stream */		\
 		       SIO_NOCLOSE|	/* do no close on abort */	\
 		       SIO_ISATTY|	/* terminal */			\
 		       SIO_NOFEOF)	/* reset on end-of-file */
