@@ -25,12 +25,14 @@
 /*#define O_SECURE 1*/
 /*#define O_DEBUG 1*/
 #include "pl-incl.h"
+#include "pl-dbref.h"
+#include "pl-termwalk.c"
 
 #define WORDS_PER_PLINT (sizeof(int64_t)/sizeof(word))
 
 static RecordList lookupRecordList(word);
 static RecordList isCurrentRecordList(word);
-static void freeRecordRef(RecordRef r);
+static void freeRecordRef(RecordRef r ARG_LD);
 
 #define RECORDA 0
 #define RECORDZ 1
@@ -43,7 +45,7 @@ static void freeRecordRef(RecordRef r);
 
 void
 initRecords(void)
-{ GD->tables.record_lists = newHTable(8);
+{ GD->recorded_db.record_lists = newHTable(8);
 }
 
 
@@ -54,7 +56,7 @@ static RecordList
 lookupRecordList(word key)
 { Symbol s;
 
-  if ( (s = lookupHTable(GD->tables.record_lists, (void *)key)) )
+  if ( (s = lookupHTable(GD->recorded_db.record_lists, (void *)key)) )
   { return s->value;
   } else
   { GET_LD
@@ -68,7 +70,14 @@ lookupRecordList(word key)
     l->references = 0;
     l->flags = 0;
     l->firstRecord = l->lastRecord = NULL;
-    addHTable(GD->tables.record_lists, (void *)key, l);
+    l->next = NULL;
+    addHTable(GD->recorded_db.record_lists, (void *)key, l);
+    if ( !GD->recorded_db.head )
+    { GD->recorded_db.head = GD->recorded_db.tail = l;
+    } else
+    { GD->recorded_db.tail->next = l;
+      GD->recorded_db.tail = l;
+    }
 
     return l;
   }
@@ -79,7 +88,7 @@ static RecordList
 isCurrentRecordList(word key)
 { Symbol s;
 
-  if ( (s = lookupHTable(GD->tables.record_lists, (void *)key)) )
+  if ( (s = lookupHTable(GD->recorded_db.record_lists, (void *)key)) )
     return s->value;
 
   return NULL;
@@ -90,16 +99,16 @@ isCurrentRecordList(word key)
 */
 
 static void
-cleanRecordList(RecordList rl)
+cleanRecordList(RecordList rl ARG_LD)
 { RecordRef *p;
   RecordRef r, prev=NULL;
 
   for(p = &rl->firstRecord; (r=*p); )
-  { if ( true(r->record, ERASED) )
+  { if ( true(r->record, R_ERASED) )
     { *p = r->next;
       if ( r == rl->lastRecord )
 	rl->lastRecord = prev;
-      freeRecordRef(r);
+      freeRecordRef(r PASS_LD);
     } else
     { prev = r;
       p = &r->next;
@@ -139,10 +148,10 @@ typedef struct
 
 #define	PL_TYPE_VARIABLE	(1)	/* variable */
 #define PL_TYPE_ATOM		(2)	/* atom */
-#define PL_TYPE_INTEGER	  	(3)	/* big integer */
+#define PL_TYPE_INTEGER		(3)	/* big integer */
 #define PL_TYPE_TAGGED_INTEGER  (4)	/* tagged integer */
-#define PL_TYPE_FLOAT	  	(5)	/* double */
-#define PL_TYPE_STRING	  	(6)	/* string */
+#define PL_TYPE_FLOAT		(5)	/* double */
+#define PL_TYPE_STRING		(6)	/* string */
 #define PL_TYPE_COMPOUND	(7)	/* compound term */
 #define PL_TYPE_CONS		(8)	/* list-cell */
 
@@ -157,7 +166,9 @@ typedef struct
 #define addUnalignedBuf(b, ptr, type) \
 	do \
 	{ if ( (b)->top + sizeof(type) > (b)->max ) \
-	    growBuffer((Buffer)b, sizeof(type)); \
+	  { if ( !growBuffer((Buffer)b, sizeof(type)) ) \
+	      outOfCore(); \
+	  } \
 	  memcpy((b)->top, ptr, sizeof(type)); \
 	  (b)->top += sizeof(type); \
 	} while(0)
@@ -332,155 +343,157 @@ typedef struct
 #define isAttVarP(p)  ((word)(p) & 0x1)
 #define valAttVarP(p) ((Word)((word)(p) & ~0x1L))
 
-static void
-compile_term_to_heap(Word p, CompileInfo info ARG_LD)
-{ word w;
+static int
+compile_term_to_heap(term_agenda *agenda, CompileInfo info ARG_LD)
+{ Word p;
 
-right_recursion:
-  w = *p;
+  while( (p=nextTermAgenda(agenda)) )
+  { word w;
 
-  switch(tag(w))
-  { case TAG_VAR:
-    { intptr_t n = info->nvars++;
+  again:
+    w = *p;
 
-      *p = (n<<7)|TAG_ATOM|STG_GLOBAL;
-      addBuffer(&info->vars, p, Word);
-      addOpCode(info, PL_TYPE_VARIABLE);
-      addSizeInt(info, n);
-
-      return;
-    }
-#if O_ATTVAR
-    case TAG_ATTVAR:
-    { intptr_t n = info->nvars++;
-      Word ap = valPAttVar(w);
-
-      if ( isEmptyBuffer(&info->code) )
-      { addOpCode(info, PL_REC_ALLOCVAR); 	/* only an attributed var */
-	info->size++;
-      }
-
-      addBuffer(&info->vars, *p, word);		/* save value */
-      *p = (n<<7)|TAG_ATOM|STG_GLOBAL;
-      addBuffer(&info->vars, mkAttVarP(p), Word);
-      addOpCode(info, PL_TYPE_ATTVAR);
-      addSizeInt(info, n);
-      info->size++;
-      DEBUG(9, Sdprintf("Added attvar %d\n", n));
-
-      p = ap;
-      goto right_recursion;
-    }
-#endif
-    case TAG_ATOM:
-    { if ( storage(w) == STG_GLOBAL )	/* this is a variable */
-      { intptr_t n = ((intptr_t)(w) >> 7);
+    switch(tag(w))
+    { case TAG_VAR:
+      { intptr_t n = info->nvars++;
 
+	*p = (n<<7)|TAG_ATOM|STG_GLOBAL;
+	addBuffer(&info->vars, p, Word);
 	addOpCode(info, PL_TYPE_VARIABLE);
 	addSizeInt(info, n);
-	DEBUG(9, Sdprintf("Added var-link %d\n", n));
-      } else
-      { addAtom(info, w);
-	DEBUG(9, Sdprintf("Added '%s'\n", stringAtom(w)));
+
+	continue;
       }
+  #if O_ATTVAR
+      case TAG_ATTVAR:
+      { intptr_t n = info->nvars++;
+	Word ap = valPAttVar(w);
 
-      return;
-    }
-    case TAG_INTEGER:
-    { int64_t val;
-
-      if ( isTaggedInt(w) )
-      { val = valInt(w);
-	addOpCode(info, PL_TYPE_TAGGED_INTEGER);
-	addInt64(info, val);
-      } else
-      { number n;
-
-	info->size += wsizeofIndirect(w) + 2;
-
-	get_integer(w, &n);
-	switch(n.type)
-	{ case V_INTEGER:
-	    addOpCode(info, PL_TYPE_INTEGER);
-	    addInt64(info, n.value.i);
-	    break;
-#ifdef O_GMP
-	  case V_MPZ:
-	    addOpCode(info, PL_REC_MPZ);
-	    addMPZToBuffer((Buffer)&info->code, n.value.mpz);
-	    break;
-#endif
-  	  default:
-	    assert(0);
+	if ( isEmptyBuffer(&info->code) )
+	{ addOpCode(info, PL_REC_ALLOCVAR);	/* only an attributed var */
+	  info->size++;
 	}
+
+	addBuffer(&info->vars, *p, word);		/* save value */
+	*p = (n<<7)|TAG_ATOM|STG_GLOBAL;
+	addBuffer(&info->vars, mkAttVarP(p), Word);
+	addOpCode(info, PL_TYPE_ATTVAR);
+	addSizeInt(info, n);
+	info->size++;
+	DEBUG(9, Sdprintf("Added attvar %d\n", n));
+
+	p = ap;
+	deRef(p);
+	goto again;
       }
+  #endif
+      case TAG_ATOM:
+      { if ( storage(w) == STG_GLOBAL )	/* this is a variable */
+	{ intptr_t n = ((intptr_t)(w) >> 7);
 
-      return;
-    }
-    case TAG_STRING:
-    { Word f     = addressIndirect(w);
-      size_t n   = wsizeofInd(*f);
-      size_t pad = padHdr(*f);		/* see also getCharsString() */
-      size_t l   = n*sizeof(word)-pad;
+	  addOpCode(info, PL_TYPE_VARIABLE);
+	  addSizeInt(info, n);
+	  DEBUG(9, Sdprintf("Added var-link %d\n", n));
+	} else
+	{ addAtom(info, w);
+	  DEBUG(9, Sdprintf("Added '%s'\n", stringAtom(w)));
+	}
 
-      info->size += n+2;
-      addOpCode(info, PL_TYPE_STRING);
-      addChars(info, l, (const char *)(f+1)); /* +1 to skip header */
-
-      return;
-    }
-    case TAG_FLOAT:
-    { info->size += WORDS_PER_DOUBLE + 2;
-      addFloat(info, valIndirectP(w));
-
-      return;
-    }
-    case TAG_COMPOUND:
-    { Functor f = valueTerm(w);
-      int arity;
-      word functor;
-
-#if O_CYCLIC
-      if ( isInteger(f->definition) )
-      { addOpCode(info, PL_REC_CYCLE);
-	addSizeInt(info, valInt(f->definition));
-
-	DEBUG(1, Sdprintf("Added cycle for offset = %d\n",
-			  valInt(f->definition)));
-
-	return;
-      } else
-      { cycle_mark mark;
-
-	arity   = arityFunctor(f->definition);
-	functor = f->definition;
-
-	requireStack(argument, sizeof(Word)*2);
-	mark.term = f;
-	mark.fdef = f->definition;
-	pushSegStack(&LD->cycle.stack, &mark);
-	f->definition = (functor_t)consInt(info->size);
-	assert(valInt(f->definition) == (intptr_t)info->size); /* overflow test */
+	continue;
       }
-#endif
+      case TAG_INTEGER:
+      { int64_t val;
 
-      info->size += arity+1;
-      addFunctor(info, functor);
-      DEBUG(9, if ( GD->io_initialised )
-	         Sdprintf("Added %s/%d\n",
-			  stringAtom(valueFunctor(functor)->name),
-			  arityFunctor(functor)));
-      p = f->arguments;
-      for(; --arity > 0; p++)
-	compile_term_to_heap(p, info PASS_LD);
-      goto right_recursion;
+	if ( isTaggedInt(w) )
+	{ val = valInt(w);
+	  addOpCode(info, PL_TYPE_TAGGED_INTEGER);
+	  addInt64(info, val);
+	} else
+	{ number n;
+
+	  info->size += wsizeofIndirect(w) + 2;
+
+	  get_integer(w, &n);
+	  switch(n.type)
+	  { case V_INTEGER:
+	      addOpCode(info, PL_TYPE_INTEGER);
+	      addInt64(info, n.value.i);
+	      break;
+  #ifdef O_GMP
+	    case V_MPZ:
+	      addOpCode(info, PL_REC_MPZ);
+	      addMPZToBuffer((Buffer)&info->code, n.value.mpz);
+	      break;
+  #endif
+	    default:
+	      assert(0);
+	  }
+	}
+
+	continue;
+      }
+      case TAG_STRING:
+      { Word f     = addressIndirect(w);
+	size_t n   = wsizeofInd(*f);
+	size_t pad = padHdr(*f);		/* see also getCharsString() */
+	size_t l   = n*sizeof(word)-pad;
+
+	info->size += n+2;
+	addOpCode(info, PL_TYPE_STRING);
+	addChars(info, l, (const char *)(f+1)); /* +1 to skip header */
+
+	continue;
+      }
+      case TAG_FLOAT:
+      { info->size += WORDS_PER_DOUBLE + 2;
+	addFloat(info, valIndirectP(w));
+
+	continue;
+      }
+      case TAG_COMPOUND:
+      { Functor f = valueTerm(w);
+	int arity;
+	word functor;
+
+  #if O_CYCLIC
+	if ( isInteger(f->definition) )
+	{ addOpCode(info, PL_REC_CYCLE);
+	  addSizeInt(info, valInt(f->definition));
+
+	  DEBUG(1, Sdprintf("Added cycle for offset = %d\n",
+			    valInt(f->definition)));
+
+	  continue;
+	} else
+	{ cycle_mark mark;
+
+	  arity   = arityFunctor(f->definition);
+	  functor = f->definition;
+
+	  mark.term = f;
+	  mark.fdef = f->definition;
+	  pushSegStack(&LD->cycle.lstack, mark, cycle_mark);
+	  f->definition = (functor_t)consInt(info->size);
+					  /* overflow test */
+	  assert(valInt(f->definition) == (intptr_t)info->size);
+	}
+  #endif
+
+	info->size += arity+1;
+	addFunctor(info, functor);
+	DEBUG(9, if ( GD->io_initialised )
+		   Sdprintf("Added %s/%d\n",
+			    stringAtom(valueFunctor(functor)->name),
+			    arityFunctor(functor)));
+	pushWorkAgenda(agenda, arity, f->arguments);
+	continue;
+      }
+      default:
+	assert(0);
     }
-    case TAG_REFERENCE:
-      p = unRef(w);
-      goto right_recursion;
-    default:
-      assert(0);
   }
+
+  return TRUE;
 }
 
 
@@ -488,7 +501,7 @@ right_recursion:
 
 static void
 init_cycle(ARG1_LD)
-{ LD->cycle.stack.unit_size = sizeof(cycle_mark);
+{ LD->cycle.lstack.unit_size = sizeof(cycle_mark);
 }
 
 
@@ -496,7 +509,7 @@ static void
 unvisit(ARG1_LD)
 { cycle_mark mark;
 
-  while( popSegStack(&LD->cycle.stack, &mark) )
+  while( popSegStack(&LD->cycle.lstack, &mark, cycle_mark) )
   { mark.term->definition = mark.fdef;
   }
 }
@@ -508,28 +521,12 @@ static void unvisit(ARG1_LD) {}
 
 #endif
 
+static void
+restoreVars(compile_info *info)
+{ Word *p = topBuffer(&info->vars, Word);
+  Word *b = baseBuffer(&info->vars, Word);
 
-Record
-compileTermToHeap__LD(term_t t, int flags ARG_LD)
-{ compile_info info;
-  Record record;
-  Word *p;
-  size_t size;
-  size_t rsize = SIZERECORD(flags);
-  init_cycle(PASS_LD1);
-
-  SECURE(checkData(valTermRef(t)));
-
-  initBuffer(&info.code);
-  initBuffer(&info.vars);
-  info.size = 0;
-  info.nvars = 0;
-  info.external = (flags & R_EXTERNAL);
-  info.lock = !(info.external || (flags&R_NOLOCK));
-
-  compile_term_to_heap(valTermRef(t), &info PASS_LD);
-  p = topBuffer(&info.vars, Word);
-  while(p > baseBuffer(&info.vars, Word))
+  while(p > b)
   { p--;
     if (isAttVarP(*p) )
     { *valAttVarP(*p) = (word)p[-1];
@@ -537,12 +534,39 @@ compileTermToHeap__LD(term_t t, int flags ARG_LD)
     } else
       setVar(**p);
   }
-  discardBuffer(&info.vars);
+  discardBuffer(&info->vars);
+}
 
+
+Record
+compileTermToHeap__LD(term_t t, int flags ARG_LD)
+{ compile_info info;
+  Record record;
+  size_t size;
+  size_t rsize = SIZERECORD(flags);
+  term_agenda agenda;
+
+  SECURE(checkData(valTermRef(t)));
+
+  init_cycle(PASS_LD1);
+  initBuffer(&info.code);
+  initBuffer(&info.vars);
+  info.size = 0;
+  info.nvars = 0;
+  info.external = (flags & R_EXTERNAL);
+  info.lock = !(info.external || (flags&R_NOLOCK));
+
+  initTermAgenda(&agenda, 1, valTermRef(t));
+  compile_term_to_heap(&agenda, &info PASS_LD);
+  clearTermAgenda(&agenda);
+  restoreVars(&info);
   unvisit(PASS_LD1);
 
   size = rsize + sizeOfBuffer(&info.code);
   record = allocHeap(size);
+#ifdef REC_MAGIC
+  record->magic = REC_MAGIC;
+#endif
   record->gsize = (unsigned int)info.size; /* only 28-bit */
   record->nvars = info.nvars;
   record->size  = (int)size;
@@ -586,12 +610,12 @@ char *
 PL_record_external(term_t t, size_t *len)
 { GET_LD
   compile_info info;
-  Word p, *vp;
-  int n;
+  Word p;
   tmp_buffer hdr;
   int scode, shdr;
   char *rec;
   int first = REC_HDR;
+  term_agenda agenda;
 
   SECURE(checkData(valTermRef(t)));
   p = valTermRef(t);
@@ -619,30 +643,28 @@ PL_record_external(term_t t, size_t *len)
     discardBuffer(&info.code);
     *len = scode;
     return rec;
-  } else if ( isAtom(*p) )
+  } else if ( isAtom(*p) )		/* atom-only record */
   { first |= (REC_ATOM|REC_GROUND);
     addOpCode(&info, first);
     addAtomValue(&info, *p);
     goto ret_primitive;
   }
 
+					/* the real stuff */
   initBuffer(&info.vars);
   info.size = 0;
   info.nvars = 0;
   info.external = TRUE;
   info.lock = FALSE;
 
-  compile_term_to_heap(p, &info PASS_LD);
-  n = info.nvars;
-  vp = (Word *)info.vars.base;
-  while(--n >= 0)
-    setVar(**vp++);
-  discardBuffer(&info.vars);
-  scode = (int)sizeOfBuffer(&info.code);
+  initTermAgenda(&agenda, 1, p);
+  compile_term_to_heap(&agenda, &info PASS_LD);
+  clearTermAgenda(&agenda);
   if ( info.nvars == 0 )
     first |= REC_GROUND;
-
+  restoreVars(&info);
   unvisit(PASS_LD1);
+  scode = (int)sizeOfBuffer(&info.code);
 
   initBuffer(&hdr);
   addBuffer(&hdr, first, uchar);		/* magic code */
@@ -674,10 +696,10 @@ typedef struct
   const char   *base;			/* start of data */
   Word	       *vars;
   Word		gbase;			/* base of term on global stack */
-  Word 		gstore;			/* current storage location */
+  Word		gstore;			/* current storage location */
 					/* for se_record() */
-  uint  	nvars;			/* Variables seen */
-  TmpBuffer 	avars;			/* Values stored for attvars */
+  uint		nvars;			/* Variables seen */
+  TmpBuffer	avars;			/* Values stored for attvars */
 } copy_info, *CopyInfo;
 
 
@@ -718,7 +740,7 @@ other stacks as scratch-area.
 #define fetchMultipleBuf(b, var, times, type) \
 		do \
 		{ memcpy(var, (b)->data, times*sizeof(type)); \
- 		  (b)->data +=  times*sizeof(type); \
+		  (b)->data +=  times*sizeof(type); \
 		} while(0)
 #define skipBuf(b, type) \
 		((b)->data += sizeof(type))
@@ -813,181 +835,202 @@ fetchChars(CopyInfo b, unsigned len, Word to)
 }
 
 
-static void
+static int
 copy_record(Word p, CopyInfo b ARG_LD)
-{ intptr_t tag;
+{ term_agenda agenda;
+  int is_compound = FALSE;
+  int tag;
 
-right_recursion:
-  switch( (tag = fetchOpCode(b)) )
-  { case PL_TYPE_VARIABLE:
-    { intptr_t n = fetchSizeInt(b);
+  do
+  {
+  right_recursion:
+    switch( (tag = fetchOpCode(b)) )
+    { case PL_TYPE_VARIABLE:
+      { intptr_t n = fetchSizeInt(b);
 
-      if ( b->vars[n] )
-      { if ( p > b->vars[n] )		/* ensure the reference is in the */
-	  *p = makeRef(b->vars[n]);	/* right direction! */
-	else
-	{ *p = *b->vars[n];		/* wrong way.  make sure b->vars[n] */
-	  *b->vars[n] = makeRef(p);	/* stays at the real variable */
-	  b->vars[n] = p;		/* NOTE: also links attvars! */
+	if ( b->vars[n] )
+	{ if ( p > b->vars[n] )		/* ensure the reference is in the */
+	    *p = makeRef(b->vars[n]);	/* right direction! */
+	  else
+	  { *p = *b->vars[n];		/* wrong way.  make sure b->vars[n] */
+	    *b->vars[n] = makeRef(p);	/* stays at the real variable */
+	    b->vars[n] = p;		/* NOTE: also links attvars! */
+	  }
+	} else
+	{ setVar(*p);
+	  b->vars[n] = p;
 	}
-      } else
-      {	setVar(*p);
-	b->vars[n] = p;
+
+	continue;
+      }
+      case PL_REC_ALLOCVAR:
+      { setVar(*b->gstore);
+	*p = makeRefG(b->gstore);
+	p = b->gstore++;
+	goto right_recursion;
       }
-
-      return;
-    }
-    case PL_REC_ALLOCVAR:
-    { setVar(*b->gstore);
-      *p = makeRefG(b->gstore);
-      p = b->gstore++;
-      goto right_recursion;
-    }
 #if O_ATTVAR
-    case PL_TYPE_ATTVAR:
-    { intptr_t n = fetchSizeInt(b);
+      case PL_TYPE_ATTVAR:
+      { intptr_t n = fetchSizeInt(b);
 
-      *p = consPtr(b->gstore, TAG_ATTVAR|STG_GLOBAL);
-      b->vars[n] = p;
-      p = b->gstore++;
-      goto right_recursion;
-    }
+	*p = consPtr(b->gstore, TAG_ATTVAR|STG_GLOBAL);
+	b->vars[n] = p;
+	p = b->gstore++;
+	goto right_recursion;
+      }
 #endif
-    case PL_TYPE_ATOM:
-    { *p = fetchWord(b);
+      case PL_TYPE_ATOM:
+      { *p = fetchWord(b);
+	continue;
+      }
+      case PL_TYPE_EXT_ATOM:
+      { fetchAtom(b, p);
+	PL_unregister_atom(*p);
+	continue;
+      }
+      case PL_TYPE_TAGGED_INTEGER:
+      { int64_t val = fetchInt64(b);
+	*p = consInt(val);
+	continue;
+      }
+      case PL_TYPE_INTEGER:
+      { size_t i;
+	union
+	{ int64_t i64;
+	  word    w[WORDS_PER_PLINT];
+	} val;
 
-      return;
-    }
-    case PL_TYPE_EXT_ATOM:
-    { fetchAtom(b, p);
-      PL_unregister_atom(*p);
-      return;
-    }
-    case PL_TYPE_TAGGED_INTEGER:
-    { int64_t val = fetchInt64(b);
+	val.i64 = fetchInt64(b);
 
-      *p = consInt(val);
-
-      return;
-    }
-    case PL_TYPE_INTEGER:
-    { size_t i;
-      union
-      { int64_t i64;
-	word    w[WORDS_PER_PLINT];
-      } val;
-
-      val.i64 = fetchInt64(b);
-
-      *p = consPtr(b->gstore, TAG_INTEGER|STG_GLOBAL);
-      *b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
-      for(i=0; i<WORDS_PER_PLINT; i++)
-	*b->gstore++ = val.w[i];
-      *b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
-
-      return;
-    }
+	*p = consPtr(b->gstore, TAG_INTEGER|STG_GLOBAL);
+	*b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
+	for(i=0; i<WORDS_PER_PLINT; i++)
+	  *b->gstore++ = val.w[i];
+	*b->gstore++ = mkIndHdr(WORDS_PER_PLINT, TAG_INTEGER);
+	continue;
+      }
 #ifdef O_GMP
-    case PL_REC_MPZ:
-    { b->data = loadMPZFromCharp(b->data, p, &b->gstore);
-
-      return;
-    }
+      case PL_REC_MPZ:
+      { b->data = loadMPZFromCharp(b->data, p, &b->gstore);
+	continue;
+      }
 #endif
-    case PL_TYPE_FLOAT:
-    case PL_TYPE_EXT_FLOAT:
-    { *p = consPtr(b->gstore, TAG_FLOAT|STG_GLOBAL);
-      *b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
-      if ( tag == PL_TYPE_FLOAT )
-	fetchFloat(b, b->gstore);
-      else
-	fetchExtFloat(b, b->gstore);
-      b->gstore += WORDS_PER_DOUBLE;
-      *b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
+      case PL_TYPE_FLOAT:
+      case PL_TYPE_EXT_FLOAT:
+      { *p = consPtr(b->gstore, TAG_FLOAT|STG_GLOBAL);
+	*b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
+	if ( tag == PL_TYPE_FLOAT )
+	  fetchFloat(b, b->gstore);
+	else
+	  fetchExtFloat(b, b->gstore);
+	b->gstore += WORDS_PER_DOUBLE;
+	*b->gstore++ = mkIndHdr(WORDS_PER_DOUBLE, TAG_FLOAT);
+	continue;
+      }
+      case PL_TYPE_STRING:
+      { unsigned len = fetchSizeInt(b);
+	int lw, pad;
+	word hdr;
 
-      return;
-    }
-    case PL_TYPE_STRING:
-    { unsigned len = fetchSizeInt(b);
-      int lw, pad;
-      word hdr;
-
-      lw = (len+sizeof(word))/sizeof(word); /* see globalNString() */
-      pad = (lw*sizeof(word) - len);
-      *p = consPtr(b->gstore, TAG_STRING|STG_GLOBAL);
-      *b->gstore++ = hdr = mkStrHdr(lw, pad);
-      b->gstore[lw-1] = 0L;		/* zero-padding */
-      fetchChars(b, len, b->gstore);
-      b->gstore += lw;
-      *b->gstore++ = hdr;
-
-      return;
-    }
+	lw = (len+sizeof(word))/sizeof(word); /* see globalNString() */
+	pad = (lw*sizeof(word) - len);
+	*p = consPtr(b->gstore, TAG_STRING|STG_GLOBAL);
+	*b->gstore++ = hdr = mkStrHdr(lw, pad);
+	b->gstore[lw-1] = 0L;		/* zero-padding */
+	fetchChars(b, len, b->gstore);
+	b->gstore += lw;
+	*b->gstore++ = hdr;
+	continue;
+      }
 #ifdef O_CYCLIC
-    case PL_REC_CYCLE:
-    { unsigned offset = fetchSizeInt(b);
-      Word ct = b->gbase+offset;
+      case PL_REC_CYCLE:
+      { unsigned offset = fetchSizeInt(b);
+	Word ct = b->gbase+offset;
 
-      *p = consPtr(ct, TAG_COMPOUND|STG_GLOBAL);
-
-      return;
-    }
+	*p = consPtr(ct, TAG_COMPOUND|STG_GLOBAL);
+	continue;
+      }
 #endif
-  { word fdef;
-    int arity;
-    case PL_TYPE_COMPOUND:
+    { word fdef;
+      int arity;
+      case PL_TYPE_COMPOUND:
 
-      fdef = fetchWord(b);
-      arity = arityFunctor(fdef);
+	fdef = fetchWord(b);
+	arity = arityFunctor(fdef);
 
-    compound:
-      *p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
-      *b->gstore++ = fdef;
-      p = b->gstore;
-      b->gstore += arity;
-      for(; --arity > 0; p++)
-	copy_record(p, b PASS_LD);
-      goto right_recursion;
-    case PL_TYPE_EXT_COMPOUND:
-    { atom_t name;
+      compound:
+	*p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
+	*b->gstore++ = fdef;
+	p = b->gstore;
+	b->gstore += arity;
+	if ( !is_compound )
+	{ is_compound = TRUE;
+	  initTermAgenda(&agenda, arity, p);
+	} else
+	{ if ( !pushWorkAgenda(&agenda, arity, p) )
+	    return MEMORY_OVERFLOW;
+	}
+	continue;
+      case PL_TYPE_EXT_COMPOUND:
+      { atom_t name;
 
-      arity = (int)fetchSizeInt(b);
-      fetchAtom(b, &name);
-      fdef = lookupFunctorDef(name, arity);
-
-      goto compound;
+	arity = (int)fetchSizeInt(b);
+	fetchAtom(b, &name);
+	fdef = lookupFunctorDef(name, arity);
+	goto compound;
+      }
     }
-  }
-    case PL_TYPE_CONS:
-    { *p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
-      *b->gstore++ = FUNCTOR_dot2;
-      p = b->gstore;
-      b->gstore += 2;
-      copy_record(p, b PASS_LD);
-      p++;
-      goto right_recursion;
+      case PL_TYPE_CONS:
+      { *p = consPtr(b->gstore, TAG_COMPOUND|STG_GLOBAL);
+	*b->gstore++ = FUNCTOR_dot2;
+	p = b->gstore;
+	b->gstore += 2;
+	if ( !is_compound )
+	{ is_compound = TRUE;
+	  initTermAgenda(&agenda, 2, p);
+	} else
+	{ if ( !pushWorkAgenda(&agenda, 2, p) )
+	    return MEMORY_OVERFLOW;
+	}
+	continue;
+      }
+      default:
+	assert(0);
     }
-    default:
-      assert(0);
-  }
+  } while ( is_compound && (p=nextTermAgendaNoDeRef(&agenda)) );
+
+  return TRUE;
 }
 
 
-void
-copyRecordToGlobal(term_t copy, Record r ARG_LD)
+int
+copyRecordToGlobal(term_t copy, Record r, int flags ARG_LD)
 { copy_info b;
+  int rc;
 
   DEBUG(3, Sdprintf("PL_recorded(%p)\n", r));
 
+#ifdef REC_MAGIC
+  assert(r->magic == REC_MAGIC);
+#endif
+  if ( !hasGlobalSpace(r->gsize) )
+  { if ( (rc=ensureGlobalSpace(r->gsize, flags)) != TRUE )
+      return rc;
+  }
   b.base = b.data = dataRecord(r);
-  b.gbase = b.gstore = allocGlobal(r->gsize);
+  b.gbase = b.gstore = gTop;
+  gTop += r->gsize;
 
   INITCOPYVARS(b, r->nvars);
-  copy_record(valTermRef(copy), &b PASS_LD);
+  rc = copy_record(valTermRef(copy), &b PASS_LD);
   FREECOPYVARS(b, r->nvars);
+  if ( rc != TRUE )
+    return rc;
 
   assert(b.gstore == gTop);
   SECURE(checkData(valTermRef(copy)));
+
+  return TRUE;
 }
 
 
@@ -1024,79 +1067,78 @@ skipLong(CopyInfo b)
 
 static void
 scanAtomsRecord(CopyInfo b, void (*func)(atom_t a))
-{
-right_recursion:
+{ size_t work = 0;
 
-  switch( fetchOpCode(b) )
-  { case PL_TYPE_VARIABLE:
-    case PL_REC_CYCLE:
-    { skipSizeInt(b);
-      return;
-    }
-    case PL_REC_ALLOCVAR:
-      goto right_recursion;
+  do
+  { switch( fetchOpCode(b) )
+    { case PL_TYPE_VARIABLE:
+      case PL_REC_CYCLE:
+      { skipSizeInt(b);
+	continue;
+      }
+      case PL_REC_ALLOCVAR:
+	work++;
+	continue;
 #ifdef O_ATTVAR
-    case PL_TYPE_ATTVAR:
-    { skipSizeInt(b);
-      goto right_recursion;
-    }
+      case PL_TYPE_ATTVAR:
+      { skipSizeInt(b);
+	work++;
+	continue;
+      }
 #endif
-    case PL_TYPE_ATOM:
-    { atom_t a = fetchWord(b);
+      case PL_TYPE_ATOM:
+      { atom_t a = fetchWord(b);
 
-      (*func)(a);
-      return;
-    }
-    case PL_TYPE_EXT_ATOM:
-    { skipAtom(b);
-      return;
-    }
-    case PL_TYPE_TAGGED_INTEGER:
-    case PL_TYPE_INTEGER:
-    { skipLong(b);
-      return;
-    }
+	(*func)(a);
+	continue;
+      }
+      case PL_TYPE_EXT_ATOM:
+      { skipAtom(b);
+	continue;
+      }
+      case PL_TYPE_TAGGED_INTEGER:
+      case PL_TYPE_INTEGER:
+      { skipLong(b);
+	continue;
+      }
 #ifdef O_GMP
-    case PL_REC_MPZ:
-      b->data = skipMPZOnCharp(b->data);
-      return;
+      case PL_REC_MPZ:
+	b->data = skipMPZOnCharp(b->data);
+	continue;
 #endif
-    case PL_TYPE_FLOAT:
-    case PL_TYPE_EXT_FLOAT:
-    { skipBuf(b, double);
-      return;
-    }
-    case PL_TYPE_STRING:
-    { uint len = fetchSizeInt(b);
+      case PL_TYPE_FLOAT:
+      case PL_TYPE_EXT_FLOAT:
+      { skipBuf(b, double);
+	continue;
+      }
+      case PL_TYPE_STRING:
+      { uint len = fetchSizeInt(b);
+	b->data += len;
+	continue;
+      }
+      case PL_TYPE_COMPOUND:
+      { word fdef = fetchWord(b);
+	int arity;
 
-      b->data += len;
+	arity = arityFunctor(fdef);
+	work += arity;
+	continue;
+      }
+      case PL_TYPE_EXT_COMPOUND:
+      { intptr_t arity = fetchSizeInt(b);
 
-      return;
+	skipAtom(b);
+	work += arity;
+	continue;
+      }
+      case PL_TYPE_CONS:
+      { work += 2;
+	continue;
+      }
+      default:
+	assert(0);
     }
-    case PL_TYPE_COMPOUND:
-    { word fdef = fetchWord(b);
-      int arity;
-
-      arity = arityFunctor(fdef);
-      while(--arity > 0)
-	scanAtomsRecord(b, func);
-      goto right_recursion;
-    }
-    case PL_TYPE_EXT_COMPOUND:
-    { intptr_t arity = fetchSizeInt(b);
-
-      skipAtom(b);
-      while(--arity > 0)
-	scanAtomsRecord(b, func);
-      goto right_recursion;
-    }
-    case PL_TYPE_CONS:
-    { scanAtomsRecord(b, func);
-      goto right_recursion;
-    }
-    default:
-      assert(0);
-  }
+  } while ( work-- );
 }
 
 #endif /*O_ATOMGC*/
@@ -1374,11 +1416,21 @@ freeRecord__LD(Record record ARG_LD)
 }
 
 
+void
+unallocRecordRef(RecordRef r ARG_LD)
+{ freeHeap(r, sizeof(*r));
+}
+
+
 static void
-freeRecordRef(RecordRef r)
-{ GET_LD
+freeRecordRef(RecordRef r ARG_LD)
+{ int reclaim_now = false(r->record, R_DBREF);
+
   freeRecord(r->record);
-  freeHeap(r, sizeof(*r));
+  if ( reclaim_now )
+    freeHeap(r, sizeof(*r));
+  else
+    r->record = NULL;
 }
 
 
@@ -1478,8 +1530,7 @@ real atom.
 
 bool
 unifyKey(term_t key, word val)
-{ if ( (isAtom(val) && storage(val) != STG_GLOBAL) ||
-       isTaggedInt(val) )
+{ if ( isAtom(val) || isTaggedInt(val) )
     return _PL_unify_atomic(key, val);
 
   return PL_unify_functor(key, (functor_t) val);
@@ -1502,37 +1553,48 @@ getKeyEx(term_t key, word *w ARG_LD)
 }
 
 
-word
-pl_current_key(term_t k, control_t h)
-{ TableEnum e;
-  Symbol s;
+static
+PRED_IMPL("current_key", 1, current_key, PL_FA_NONDETERMINISTIC)
+{ PRED_LD
   fid_t fid;
+  RecordList rl = NULL;
+  word k = 0L;
 
-  switch( ForeignControl(h) )
+  switch( CTX_CNTRL )
   { case FRG_FIRST_CALL:
-      e = newTableEnum(GD->tables.record_lists);
-      break;
+    { if ( PL_is_variable(A1) )
+      { rl = GD->recorded_db.head;
+	break;
+      } else if ( getKeyEx(A1, &k PASS_LD) &&
+		  isCurrentRecordList(k) )
+	succeed;
+
+      fail;
+    }
     case FRG_REDO:
-      e = ForeignContextPtr(h);
+      rl = CTX_PTR;
       break;
     case FRG_CUTTED:
     default:				/* fool gcc */
-      e = ForeignContextPtr(h);
-      freeTableEnum(e);
       succeed;
   }
 
-  fid = PL_open_foreign_frame();
-  while( (s=advanceTableEnum(e)) )
-  { RecordList l = s->value;
+  if ( !(fid = PL_open_foreign_frame()) )
+    return FALSE;
 
-    if ( l->firstRecord && unifyKey(k, l->key) )
-      ForeignRedoPtr(e);
+  for( ; rl; rl = rl->next )
+  { if ( rl->firstRecord && unifyKey(A1, rl->key) )
+    { PL_close_foreign_frame(fid);
+      if ( rl->next )
+	ForeignRedoPtr(rl->next);
+      else
+	succeed;
+    }
 
     PL_rewind_foreign_frame(fid);
   }
 
-  freeTableEnum(e);
+  PL_close_foreign_frame(fid);
   fail;
 }
 
@@ -1551,13 +1613,14 @@ record(term_t key, term_t term, term_t ref, int az)
 
   if ( !getKeyEx(key, &k PASS_LD) )
     fail;
-  if ( !PL_is_variable(ref) )
+  if ( ref && !PL_is_variable(ref) )
     return PL_error(NULL, 0, NULL, ERR_TYPE, ATOM_variable, ref);
 
   copy = compileTermToHeap(term, 0);
   r = allocHeap(sizeof(*r));
   r->record = copy;
-  PL_unify_pointer(ref, r);
+  if ( ref && !PL_unify_recref(ref, r) )
+    return FALSE;
 
   LOCK();
   l = lookupRecordList(k);
@@ -1581,58 +1644,74 @@ record(term_t key, term_t term, term_t ref, int az)
 }
 
 
-word
-pl_recorda(term_t key, term_t term, term_t ref)
-{ return record(key, term, ref, RECORDA);
+static
+PRED_IMPL("recorda", va, recorda, 0)
+{ return record(A1, A2, CTX_ARITY == 3 ? A3 : 0, RECORDA);
 }
 
 
-word
-pl_recordz(term_t key, term_t term, term_t ref)
-{ return record(key, term, ref, RECORDZ);
+static
+PRED_IMPL("recordz", va, recordz, 0)
+{ return record(A1, A2, CTX_ARITY == 3 ? A3 : 0, RECORDZ);
 }
 
 
-word
-pl_recorded(term_t key, term_t term, term_t ref, control_t h)
-{ RecordList rl;
+static
+PRED_IMPL("recorded", va, recorded, PL_FA_NONDETERMINISTIC)
+{ PRED_LD
+  RecordList rl = NULL;			/* make compiler happy */
   RecordRef record;
   word k = 0L;
-  term_t copy;
   word rval;
+  fid_t fid;
+  int varkey = FALSE;			/* make compiler happy */
 
-  switch( ForeignControl(h) )
+  term_t key  = A1;
+  term_t term = A2;
+  term_t ref  = (CTX_ARITY == 3 ? A3 : 0);
+
+  switch( CTX_CNTRL )
   { case FRG_FIRST_CALL:
-    { GET_LD
-      void *ptr;
+    { if ( ref && !PL_is_variable(ref) )		/* recorded(?,?,+) */
+      { if ( PL_get_recref(ref, &record) )
+	{ LOCK();
+	  if ( unifyKey(key, record->list->key) )
+	  { int rc;
+	    term_t copy = PL_new_term_ref();
 
-      if ( PL_get_pointer(ref, &ptr) )
-      { record = ptr;
-	LOCK();
-	if ( isRecordRef(record) )
-	{ if ( unifyKey(key, record->list->key) )
-	  { GET_LD
-	    copy = PL_new_term_ref();
-	    copyRecordToGlobal(copy, record->record PASS_LD);
-	    rval = PL_unify(term, copy);
+	    if ( (rc=copyRecordToGlobal(copy, record->record,
+					ALLOW_GC PASS_LD)) < 0 )
+	      rval = raiseStackOverflow(rc);
+	    else
+	      rval = PL_unify(term, copy);
 	  } else
 	    rval = FALSE;
-	} else
-	  rval = PL_error("recorded", 3, NULL, ERR_TYPE, ATOM_record, record);
-	UNLOCK();
-	return rval;
+	  UNLOCK();
+
+	  return rval;
+	}
+	return FALSE;
+      }
+      if ( PL_is_variable(key) )
+      { if ( !(rl = GD->recorded_db.head) )
+	  fail;
+	varkey = TRUE;
+      } else if ( getKeyEx(key, &k PASS_LD) )
+      { if ( !(rl = isCurrentRecordList(k)) )
+	  fail;
+	varkey = FALSE;
+      } else
+      { return FALSE;
       }
-      if ( !getKeyEx(key, &k PASS_LD) ||
-	   !(rl = isCurrentRecordList(k)) )
-	fail;
       LOCK();
       rl->references++;
       record = rl->firstRecord;
       break;
     }
     case FRG_REDO:
-    { record = ForeignContextPtr(h);
+    { record = CTX_PTR;
       rl = record->list;
+      varkey = PL_is_variable(key);
 
       assert(rl->references > 0);
 
@@ -1640,14 +1719,14 @@ pl_recorded(term_t key, term_t term, term_t ref, control_t h)
       break;
     }
     case FRG_CUTTED:
-    { record = ForeignContextPtr(h);
+    { record = CTX_PTR;
 
       if ( record )
       { rl = record->list;
 
 	LOCK();
-	if ( --rl->references == 0 && true(rl, R_DIRTY) )
-	  cleanRecordList(rl);
+	if ( --rl->references == 0 && true(rl, RL_DIRTY) )
+	  cleanRecordList(rl PASS_LD);
 	UNLOCK();
       }
     }
@@ -1656,60 +1735,144 @@ pl_recorded(term_t key, term_t term, term_t ref, control_t h)
       succeed;
   }
 
-{ GET_LD
-  fid_t fid;
+  if ( !(fid = PL_open_foreign_frame()) )
+  { UNLOCK();
+    return FALSE;
+  }
 
-  copy = PL_new_term_ref();
-  fid = PL_open_foreign_frame();
+  while( rl )
+  { for( ; record; record = record->next )
+    { int rc;
+      term_t copy;
 
-  for( ; record; record = record->next )
-  { if ( true(record->record, ERASED) )
-      continue;
+      if ( true(record->record, R_ERASED) )
+	continue;
 
-    copyRecordToGlobal(copy, record->record PASS_LD);
-    if ( PL_unify(term, copy) && PL_unify_pointer(ref, record) )
-    { PL_close_foreign_frame(fid);
-
-      if ( !record->next )
-      { if ( --rl->references == 0 && true(rl, R_DIRTY) )
-	  cleanRecordList(rl);
-	UNLOCK();
-	succeed;
-      } else
+      copy = PL_new_term_ref();
+      if ( (rc=copyRecordToGlobal(copy, record->record, ALLOW_GC PASS_LD)) < 0 )
       { UNLOCK();
-	ForeignRedoPtr(record->next);
+	return raiseStackOverflow(rc);
+      }
+      if ( PL_unify(term, copy) &&
+	   (!ref || PL_unify_recref(ref, record)) )
+      { PL_close_foreign_frame(fid);
+
+	if ( varkey && !unifyKey(key, rl->key) )	/* stack overflow */
+	{ UNLOCK();
+	  fail;
+	}
+
+	if ( record->next )
+	{ UNLOCK();
+	  ForeignRedoPtr(record->next);
+	} else
+	{ if ( --rl->references == 0 && true(rl, RL_DIRTY) )
+	    cleanRecordList(rl PASS_LD);
+
+	  if ( varkey )
+	  { for( rl=rl->next; rl; rl=rl->next )
+	    { if ( rl->firstRecord )
+	      { rl->references++;
+		UNLOCK();
+		ForeignRedoPtr(rl->firstRecord);
+	      }
+	    }
+	  }
+
+	  UNLOCK();
+	  succeed;
+	}
+      }
+
+      PL_rewind_foreign_frame(fid);
+    }
+
+    if ( --rl->references == 0 && true(rl, RL_DIRTY) )
+      cleanRecordList(rl PASS_LD);
+
+    if ( varkey )
+    { if ( rl->next )
+      { rl = rl->next;
+	rl->references++;
+	record = rl->firstRecord;
+
+	continue;
       }
     }
 
-    PL_rewind_foreign_frame(fid);
+    break;
   }
 
   PL_close_foreign_frame(fid);
-}
-
-  if ( --rl->references == 0 && true(rl, R_DIRTY) )
-    cleanRecordList(rl);
 
   UNLOCK();
   fail;
 }
 
 
-word
-pl_erase(term_t ref)
-{ GET_LD
+/** instance(+Ref, -Term)
+*/
+
+static
+PRED_IMPL("instance", 2, instance, 0)
+{ PRED_LD
+  void *ptr;
+  db_ref_type type;
+
+  term_t ref  = A1;
+  term_t term = A2;
+
+  if ( !(ptr=PL_get_dbref(ref, &type)) )
+    return FALSE;
+
+  if ( type == DB_REF_CLAUSE )
+  { Clause clause = ptr;
+    uintptr_t generation = environment_frame->generation;
+
+    if ( true(clause, GOAL_CLAUSE) ||
+	 !visibleClause(clause, generation) )
+      return FALSE;
+
+    if ( true(clause, UNIT_CLAUSE) )
+    { term_t head = PL_new_term_ref();
+
+      return ( decompile(clause, head, 0) &&
+	       PL_unify_term(term,
+			     PL_FUNCTOR, FUNCTOR_prove2,
+			       PL_TERM, head,
+			       PL_ATOM, ATOM_true) );
+    } else
+    { return decompile(clause, term, 0);
+    }
+  } else
+  { RecordRef rref = ptr;
+    term_t t = PL_new_term_ref();
+
+    if ( copyRecordToGlobal(t, rref->record, ALLOW_GC PASS_LD) == TRUE )
+      return PL_unify(term, t);
+  }
+
+  return FALSE;
+}
+
+
+
+static
+PRED_IMPL("erase", 1, erase, 0)
+{ PRED_LD
   void *ptr;
-  RecordRef record;
   RecordRef prev, r;
   RecordList l;
   word rval;
+  db_ref_type type;
 
-  if ( !PL_get_pointer(ref, &ptr) || !inCore(ptr))
-    return PL_error("erase", 1, NULL, ERR_TYPE, ATOM_db_reference, ref);
+  term_t ref = A1;
 
-  record = ptr;
-  if ( isClause(record) )
-  { Clause clause = (Clause) record;
+  if ( !(ptr=PL_get_dbref(ref, &type)) )
+    return FALSE;
+
+  if ( type == DB_REF_CLAUSE )
+  { Clause clause = ptr;
     Definition def = getProcDefinition(clause->procedure);
 
     if ( !true(def, DYNAMIC) )
@@ -1717,24 +1880,23 @@ pl_erase(term_t ref)
 		      ATOM_clause, ATOM_erase, ref);
 
     return retractClauseDefinition(def, clause PASS_LD);
-  }
+  } else
+  { RecordRef record = ptr;
 
-  LOCK();
-  if ( isRecordRef(record) )
-  {
 #if O_DEBUGGER
-    callEventHook(PLEV_ERASED, record);
+    callEventHook(PLEV_ERASED_RECORD, record);
 #endif
 
+    LOCK();
     l = record->list;
     if ( l->references )		/* a recorded has choicepoints */
-    { set(record->record, ERASED);
-      set(l, R_DIRTY);
+    { set(record->record, R_ERASED);
+      set(l, RL_DIRTY);
     } else if ( record == l->firstRecord )
     { if ( !record->next )
 	l->lastRecord = NULL;
       l->firstRecord = record->next;
-      freeRecordRef(record);
+      freeRecordRef(record PASS_LD);
     } else
     { prev = l->firstRecord;
       r = prev->next;
@@ -1745,82 +1907,33 @@ pl_erase(term_t ref)
 	    l->lastRecord = prev;
 	  }
 	  prev->next = r->next;
-	  freeRecordRef(r);
+	  freeRecordRef(r PASS_LD);
 	  goto ok;
 	}
       }
-      goto nok;
+      assert(0);
     }
 
   ok:
+    UNLOCK();
     rval = TRUE;
-  } else
-  { nok:
-    rval = PL_error("erase", 1, NULL, ERR_DOMAIN, ATOM_db_reference, ref);
   }
-  UNLOCK();
 
   return rval;
 }
 
 		 /*******************************
-		 *	     COMPLEXITY		*
+		 *      PUBLISH PREDICATES	*
 		 *******************************/
 
-static int
-count_term(Word t, int left ARG_LD)
-{ int count = 0;
-
-right_recursion:
-  deRef(t);
-
-  if ( --left < 0 )
-    return -1;
-  count++;
-
-  if ( isAttVar(*t) )
-  { Word p = valPAttVar(*t);
-
-    assert(onStackArea(global, p));
-    t = p;
-    goto right_recursion;
-  } else if ( isTerm(*t) )
-  { int arity = arityTerm(*t);
-    int me;
-
-    for(t = argTermP(*t, 0); arity-- > 0; count += me, t++ )
-    { if ( arity == 0 )
-	goto right_recursion;
-
-      me = count_term(t, left PASS_LD);
-      if ( me < 0 )
-	return me;
-      left -= me;
-      if ( left < 0 )
-	return -1;
-    }
-  }
-
-  return count+1;
-}
-
-
-#ifndef INT_MAX
-#define INT_MAX	    ((int)(((unsigned int)1<<(sizeof(int)*8-1))-1))
-#define INT_MIN     (-(INT_MIN)-1)
-#endif
-
-word
-pl_term_complexity(term_t t, term_t mx, term_t count)
-{ GET_LD
-  int c, m;
-
-  if ( !PL_get_integer(mx, &m) )
-    m = INT_MAX;
-
-  c = count_term(valTermRef(t), m PASS_LD);
-  if ( c < 0 || c > m )
-    fail;
-
-  return PL_unify_integer(count, c);
-}
+BeginPredDefs(rec)
+  PRED_SHARE("recorded", 2, recorded, PL_FA_NONDETERMINISTIC)
+  PRED_SHARE("recorded", 3, recorded, PL_FA_NONDETERMINISTIC)
+  PRED_SHARE("recordz", 2, recordz, 0)
+  PRED_SHARE("recordz", 3, recordz, 0)
+  PRED_SHARE("recorda", 2, recorda, 0)
+  PRED_SHARE("recorda", 3, recorda, 0)
+  PRED_DEF("erase", 1, erase, 0)
+  PRED_DEF("instance", 2, instance, 0)
+  PRED_DEF("current_key", 1, current_key, PL_FA_NONDETERMINISTIC)
+EndPredDefs
