@@ -62,7 +62,7 @@ initSegStack(segstack *stack, size_t unit_size, size_t len, void *data)
 
 
 int
-pushSegStack(segstack *stack, void *data)
+pushSegStack_(segstack *stack, void *data)
 { if ( stack->top + stack->unit_size <= stack->max )
   { memcpy(stack->top, data, stack->unit_size);
     stack->top += stack->unit_size;
@@ -101,13 +101,35 @@ pushSegStack(segstack *stack, void *data)
 }
 
 
+int
+pushRecordSegStack(segstack *stack, Record r)
+{ if ( stack->top + sizeof(r) <= stack->max )
+  { Record *rp = (Record*)stack->top;
+
+    *rp++ = r;
+    stack->top = (char*)rp;
+    stack->count++;
+
+    return TRUE;
+  } else
+  { int rc;
+
+    PL_LOCK(L_AGC);
+    rc = pushSegStack(stack, r, Record);
+    PL_UNLOCK(L_AGC);
+
+    return rc;
+  }
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Pop data. Note that we leave the first chunk associated with the stack
 to speedup frequent small usage.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 int
-popSegStack(segstack *stack, void *data)
+popSegStack_(segstack *stack, void *data)
 { again:
 
   if ( stack->top >= stack->base + stack->unit_size )
@@ -139,6 +161,51 @@ popSegStack(segstack *stack, void *data)
 }
 
 
+void *
+topOfSegStack(segstack *stack)
+{ segchunk *chunk;
+
+  if ( stack->top >= stack->base + stack->unit_size )
+  { return stack->top - stack->unit_size;
+  } else if ( stack->last && (chunk=stack->last->previous) )
+  { assert(chunk->top - stack->unit_size >= chunk->data);
+    return chunk->top - stack->unit_size;
+  }
+
+  return NULL;
+}
+
+
+void
+popTopOfSegStack(segstack *stack)
+{ again:
+
+  if ( stack->top >= stack->base + stack->unit_size )
+  { stack->top -= stack->unit_size;
+    stack->count--;
+  } else
+  { segchunk *chunk = stack->last;
+
+    if ( chunk )
+    { if ( chunk->previous )
+      { stack->last = chunk->previous;
+	stack->last->next = NULL;
+	if ( chunk->allocated )
+	  PL_free(chunk);
+
+	chunk = stack->last;
+	stack->base = chunk->data;
+	stack->max  = addPointer(chunk, chunk->size);
+	stack->top  = chunk->top;
+	goto again;
+      }
+    }
+
+    assert(0);
+  }
+}
+
+
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 scanSegStack(segstack *stack, void (*func)(void *cell))
 Walk along all living cells on the stack and call func on them.  The stack
@@ -168,15 +235,30 @@ scanSegStack(segstack *stack, void (*func)(void *cell))
 
 void
 clearSegStack(segstack *s)
-{ segchunk *c, *n;
+{ segchunk *c;
 
-  c = s->first;
+  if ( (c = s->first) )
+  { segchunk *n;
 
-  for(; c; c = n)
-  { n = c->next;
-    if ( c->allocated )
-      PL_free(c);
+    if ( !c->allocated )		/* statically allocated first chunk */
+    { n = c->next;
+
+      c->next = NULL;
+      s->last = c;
+      s->base = s->top = c->top;
+      s->max  = addPointer(c, c->size);
+      s->count = 0;
+
+      for(c=n; c; c = n)
+      { n = c->next;
+	PL_free(c);
+      }
+    } else				/* all dynamic chunks */
+    { for(; c; c = n)
+      { n = c->next;
+	PL_free(c);
+      }
+      memset(s, 0, sizeof(*s));
+    }
   }
-
-  memset(s, 0, sizeof(*s));
 }
