@@ -83,7 +83,9 @@ typedef struct
 
 void
 resetTerm()
-{ LOCK();
+{ GET_LD
+
+  LOCK();
   if ( capabilities == NULL )
   { capabilities = newHTable(16);
   } else
@@ -101,7 +103,9 @@ resetTerm()
 
 static int
 initTerm(void)
-{ if ( term_initialised == STAT_START )
+{ GET_LD
+
+  if ( term_initialised == STAT_START )
   { char term[100];
 
     term_initialised = STAT_ERROR;
@@ -152,7 +156,8 @@ out:
 
 static Entry
 lookupEntry(atom_t name, atom_t type)
-{ Symbol s;
+{ GET_LD
+  Symbol s;
   Entry e;
 
   LOCK();
@@ -183,7 +188,9 @@ lookupEntry(atom_t name, atom_t type)
       if ( (s = tgetstr(stringAtom(name), &string_area)) != NULL )
         e->value  = PL_new_atom(s);	/* locked: ok */
     } else
-    { warning("tgetent/3: Illegal type");
+    { term_t t = PL_new_term_ref();
+      PL_put_atom(t, type);
+      PL_domain_error("termcap_type", t);
       freeHeap(e, sizeof(entry));
       e = NULL;
       goto out;
@@ -201,7 +208,8 @@ out:
 
 word
 pl_tty_get_capability(term_t name, term_t type, term_t value)
-{ Entry e;
+{ GET_LD
+  Entry e;
   atom_t n, t;
 
   if ( !PL_get_atom_ex(name, &n) || !PL_get_atom_ex(type, &t) )
@@ -218,13 +226,15 @@ pl_tty_get_capability(term_t name, term_t type, term_t value)
 
 static int
 tputc(int chr)
-{ return Sputc(chr, Suser_output);
+{ GET_LD
+  return Sputc(chr, Suser_output);
 }
 
 
 word
 pl_tty_goto(term_t x, term_t y)
-{ Entry e;
+{ GET_LD
+  Entry e;
   char *s;
   int ix, iy;
 
@@ -255,7 +265,7 @@ pl_tty_put(term_t a, term_t affcnt)
 { char *s;
   int n;
 
-  if ( PL_get_chars_ex(a, &s, CVT_ALL) &&
+  if ( PL_get_chars(a, &s, CVT_ALL|CVT_EXCEPTION) &&
        PL_get_integer_ex(affcnt, &n) )
   { tputs(s, n, tputc);
     succeed;
@@ -267,7 +277,8 @@ pl_tty_put(term_t a, term_t affcnt)
 
 word
 pl_tty_size(term_t r, term_t c)
-{ int rows, cols;
+{ GET_LD
+  int rows, cols;
 
 #ifdef __unix__
   int iorval;
@@ -327,7 +338,8 @@ getModuleFunction(const char *module, const char *name)
 
 word
 pl_tty_size(term_t r, term_t c)
-{ int (*ScreenCols)(void *h) = getModuleFunction("plterm", "ScreenCols");
+{ GET_LD
+  int (*ScreenCols)(void *h) = getModuleFunction("plterm", "ScreenCols");
   int (*ScreenRows)(void *h) = getModuleFunction("plterm", "ScreenRows");
 
   if ( ScreenCols && ScreenRows )
