@@ -3,9 +3,10 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2011, University of Amsterdam
+			      VU University Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -24,7 +25,8 @@
 
 #include <math.h>
 #include "pl-incl.h"
-#include "pl-ctype.h"
+#include "os/pl-dtoa.h"
+#include "os/pl-ctype.h"
 #include <stdio.h>			/* sprintf() */
 #ifdef HAVE_LOCALE_H
 #include <locale.h>
@@ -40,53 +42,32 @@
 #define HAVE_FPCLASSIFY 1
 #endif
 
-typedef struct visited
-{ Word address;				/* we have done this address */
-  struct visited *next;			/* next already visited */
-} visited;
-
 typedef struct
 { int   flags;				/* PL_WRT_* flags */
   int   max_depth;			/* depth limit */
   int   depth;				/* current depth */
+  atom_t spacing;			/* Where to insert spaces */
   Module module;			/* Module for operators */
   IOSTREAM *out;			/* stream to write to */
-  visited *visited;			/* visited (attributed-) variables */
+  term_t portray_goal;			/* call/2 activated portray hook */
+  term_t write_options;			/* original write options */
+  term_t prec_opt;			/* term in write options with prec */
 } write_options;
 
-static bool	writeTerm2(term_t term, int prec, write_options *options);
-static bool	writeTerm(term_t t, int prec, write_options *options);
-
-static Word
-address_of(term_t t)
-{ Word adr = valTermRef(t);
-
-  deRef(adr);
-  switch(tag(*adr))
-  { case TAG_ATTVAR:
-      return adr;
-    case TAG_COMPOUND:
-      return valPtr(*adr);
-    default:
-      return NULL;			/* non-recursive structure */
-  }
-}
-
-
-static int
-has_visited(visited *v, Word addr)
-{ for( ; v; v=v->next )
-  { if ( v->address == addr )
-      succeed;
-  }
-
-  fail;
-}
-
+static bool	writeTerm2(term_t term, int prec,
+			   write_options *options, bool arg) WUNUSED;
+static bool	writeTerm(term_t t, int prec,
+			  write_options *options) WUNUSED;
+static bool	writeArgTerm(term_t t, int prec,
+			     write_options *options, bool arg) WUNUSED;
+static int	PutToken(const char *s, IOSTREAM *stream);
+static int	writeAtom(atom_t a, write_options *options);
+static int	callPortray(term_t arg, int prec, write_options *options);
 
 char *
 varName(term_t t, char *name)
-{ Word adr = valTermRef(t);
+{ GET_LD
+  Word adr = valTermRef(t);
 
   deRef(adr);
 
@@ -99,6 +80,96 @@ varName(term_t t, char *name)
 }
 
 
+static int
+atomIsVarName(atom_t a)
+{ Atom atom = atomValue(a);
+
+  if ( false(atom->type, PL_BLOB_TEXT) || atom->length == 0 )
+    fail;
+  if ( isUCSAtom(atom) )
+  { pl_wchar_t *w = (pl_wchar_t*)atom->name;
+    size_t len = atom->length / sizeof(pl_wchar_t);
+
+    return atom_varnameW(w, len);
+  } else
+  { const char *s = atom->name;
+    size_t len = atom->length;
+
+    if ( isUpper(*s) || *s == '_' )
+    { for(s++; --len > 0; s++)
+      { if ( !isAlpha(*s) )
+	  return FALSE;
+      }
+
+      return TRUE;
+    }
+
+    return FALSE;
+  }
+}
+
+
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Return:	TRUE:  processes
+	FALSE: not processed
+	-1:    error
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
+static int
+writeNumberVar(term_t t, write_options *options ARG_LD)
+{ Word p = valTermRef(t);
+  Functor f;
+
+  deRef(p);
+  if ( !isTerm(*p) )
+    return FALSE;
+  f = valueTerm(*p);
+
+  if ( f->definition != FUNCTOR_isovar1 )
+    return FALSE;
+
+  if ( LD->var_names.numbervars_frame )
+  { FliFrame fr = (FliFrame)valTermRef(LD->var_names.numbervars_frame);
+
+    assert(fr->magic == FLI_MAGIC);
+    if ( fr->mark.globaltop > (Word)f )
+      return FALSE;			/* older $VAR term */
+  }
+
+  p = &f->arguments[0];
+  deRef(p);
+  if ( isInteger(*p) )
+  { int64_t n = valInteger(*p);
+    char buf[32];			/* Max is H354745078340568300 */
+
+    if ( n < 0 )
+    { sprintf(buf, "S_" INT64_FORMAT, -n);
+    } else
+    { int i = (int)(n % 26);
+      int64_t j = n / 26;
+
+      if ( j == 0 )
+      { buf[0] = i+'A';
+	buf[1] = EOS;
+      } else
+      { sprintf(buf, "%c" INT64_FORMAT, i+'A', j);
+      }
+    }
+
+    return PutToken(buf, options->out) ? TRUE : -1;
+  }
+
+  if ( isAtom(*p) && atomIsVarName(*p) )
+  { write_options o2 = *options;
+    clear(&o2, PL_WRT_QUOTED);
+
+    return writeAtom(*p, &o2) ? TRUE : -1;
+  }
+
+  return FALSE;
+}
+
+
 #define AT_LOWER	0
 #define AT_QUOTE	1
 #define AT_FULLSTOP	2
@@ -176,6 +247,15 @@ PutString(const char *str, IOSTREAM *s)
 }
 
 
+static bool
+PutComma(write_options *options)
+{ if ( options->spacing == ATOM_next_argument )
+    return PutString(", ", options->out);
+  else
+    return PutString(",", options->out);
+}
+
+
 static bool
 PutStringN(const char *str, size_t length, IOSTREAM *s)
 { size_t i;
@@ -195,19 +275,33 @@ PutOpenToken() inserts a space in the output stream if the last-written
 and given character require a space to ensure a token-break.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
+#define LAST_C_RESERVED		0x110000 /* Above Unicode range */
+#define PREFIX_SIGN		(LAST_C_RESERVED+1)
+
+#define isquote(c) ((c) == '\'' || (c) == '"')
+
 static bool
 needSpace(int c, IOSTREAM *s)
 { if ( c == EOF )
   { s->lastc = EOF;
     return FALSE;
-  } else if ( s->lastc != EOF &&
-	      ((isAlphaW(s->lastc) && isAlphaW(c)) ||
-	       (isSymbolW(s->lastc) && isSymbolW(c)) ||
-	       (s->lastc != '(' && !isBlank(s->lastc) && c == '(') ||
-	       (c == '\'' && isDigit(s->lastc))) )
-  { return TRUE;
   }
 
+  if ( s->lastc == PREFIX_SIGN )	/* avoid passing to is*W() functions */
+  { if ( isDigit(c) || isSymbolW(c) )
+      return TRUE;
+    return FALSE;
+  }
+
+  if ( s->lastc != EOF &&
+       ((isAlphaW(s->lastc) && isAlphaW(c)) ||
+	(isSymbolW(s->lastc) && isSymbolW(c)) ||
+	(s->lastc != '(' && !isBlank(s->lastc) && c == '(') ||
+	(c == '\'' && (isDigit(s->lastc))) ||
+	(isquote(c) && s->lastc == c)
+       ) )
+    return TRUE;
+
   return FALSE;
 }
 
@@ -281,7 +375,8 @@ PutCloseBrace(IOSTREAM *s)
 static bool
 putQuoted(int c, int quote, int flags, IOSTREAM *stream)
 { if ( (flags & PL_WRT_CHARESCAPES) )
-  { if ( !(c < 0xff && isControl(c)) && c != quote && c != '\\' )
+  { if ( c == ' ' ||
+	 (!(c < 0xff && !isGraph(c)) && c != quote && c != '\\') )
     { TRY(Putc(c, stream));
     } else
     { char esc[8];
@@ -328,10 +423,11 @@ putQuoted(int c, int quote, int flags, IOSTREAM *stream)
 	fail;
     }
   } else
-  { if ( c == quote )
-    { TRY(Putc(c, stream) && Putc(c, stream));
-    } else
-    { return Putc(c, stream);
+  { if ( !Putc(c, stream) )
+      fail;
+    if ( c == quote || c == '\\' )	/* write '' or \\ */
+    { if ( !Putc(c, stream) )
+	fail;
     }
   }
 
@@ -358,50 +454,50 @@ writeQuoted(IOSTREAM *stream, const char *text, size_t len, int quote,
 #if O_ATTVAR
 static bool
 writeAttVar(term_t av, write_options *options)
-{ char buf[32];
+{ GET_LD
+  char buf[32];
 
   TRY(PutToken(varName(av, buf), options->out));
 
   if ( (options->flags & PL_WRT_ATTVAR_DOTS) )
   { return PutString("{...}", options->out);
   } else if ( (options->flags & PL_WRT_ATTVAR_WRITE) )
-  { fid_t fid = PL_open_foreign_frame();
-    term_t a = PL_new_term_ref();
-    visited v;
+  { fid_t fid;
+    term_t a;
 
-    v.address = address_of(av);
-    if ( has_visited(options->visited, v.address) )
-      succeed;
-    v.next = options->visited;
-    options->visited = &v;
-    Sputc('{', options->out);
-    PL_get_attr(av, a);
+    if ( !(fid = PL_open_foreign_frame()) )
+      return FALSE;
+
+    Sputcode('{', options->out);
+    a = PL_new_term_ref();
+    PL_get_attr__LD(av, a PASS_LD);
     if ( !writeTerm(a, 1200, options) )
-      goto error;
-    Sputc('}', options->out);
-    PL_discard_foreign_frame(fid);
+      return FALSE;
+    Sputcode('}', options->out);
+    PL_close_foreign_frame(fid);
 
-    options->visited = v.next;
-    succeed;
-
-  error:
-    options->visited = v.next;
-    fail;
+    return TRUE;
   } else if ( (options->flags & PL_WRT_ATTVAR_PORTRAY) &&
 	      GD->cleaning <= CLN_PROLOG )
-  { fid_t fid   = PL_open_foreign_frame();
-    predicate_t pred;
+  { predicate_t pred;
     IOSTREAM *old;
+    wakeup_state wstate;
+    int rc;
 
     pred = _PL_predicate("portray_attvar", 1, "$attvar",
 			 &GD->procedures.portray_attvar1);
 
+    if ( !saveWakeup(&wstate, TRUE PASS_LD) )
+      return FALSE;
     old = Scurout;
     Scurout = options->out;
-    PL_call_predicate(NULL, PL_Q_NODEBUG, pred, av);
+    rc = PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_PASS_EXCEPTION, pred, av);
+    if ( rc != TRUE && !PL_exception(0) )
+      rc = TRUE;
     Scurout = old;
+    restoreWakeup(&wstate PASS_LD);
 
-    PL_discard_foreign_frame(fid);
+    return rc;
   }
 
   succeed;
@@ -431,6 +527,24 @@ static int				/* FALSE, TRUE or TRUE_WITH_SPACE */
 writeAtom(atom_t a, write_options *options)
 { Atom atom = atomValue(a);
 
+  if ( (options->flags & PL_WRT_BLOB_PORTRAY) &&
+       false(atom->type, PL_BLOB_TEXT) &&
+       GD->cleaning <= CLN_PROLOG )
+  { GET_LD
+    int rc;
+    fid_t fid;
+    term_t av;
+
+    if ( !(fid = PL_open_foreign_frame()) )
+      return FALSE;
+    av = PL_new_term_ref();
+    PL_put_atom(av, a);
+    rc = callPortray(av, 1200, options);
+    PL_close_foreign_frame(fid);
+    if ( rc == TRUE )
+      return TRUE;
+  }
+
   if ( atom->type->write )
     return (*atom->type->write)(options->out, a, options->flags);
   if ( false(atom->type, PL_BLOB_TEXT) )
@@ -480,33 +594,19 @@ writeUCSAtom(IOSTREAM *fd, atom_t atom, int flags)
   size_t len = a->length/sizeof(pl_wchar_t);
   pl_wchar_t *e = &s[len];
 
-  if ( flags & PL_WRT_QUOTED )
+  if ( (flags&PL_WRT_QUOTED) && !unquoted_atomW(s, len, fd) )
   { pl_wchar_t quote = L'\'';
-    int rc;
 
-    if ( isLowerW(*s) )
-    { pl_wchar_t *q;
-
-      for(q=s; q<e; q++)
-      { if ( !isAlphaW(*q) || Scanrepresent(*q, fd) < 0 )
-	  break;
-      }
-      if ( q == e )
-        goto unquoted;
-    }
-
-    TRY(Putc(quote, fd));
+    TRY(PutOpenToken(quote, fd) &&
+	Putc(quote, fd));
 
     while(s < e)
     { TRY(putQuoted(*s++, quote, flags, fd));
     }
 
-    rc = Putc(quote, fd);
-
-    return rc;
+    return Putc(quote, fd);
   }
 
-unquoted:
   if ( s < e && !PutOpenToken(s[0], fd) )
     return FALSE;
   for( ; s<e; s++)
@@ -536,7 +636,8 @@ get_chr_from_text(const PL_chars_t *t, int index)
 
 static int
 writeString(term_t t, write_options *options)
-{ PL_chars_t txt;
+{ GET_LD
+  PL_chars_t txt;
 
   PL_get_text(t, &txt, CVT_STRING);
 
@@ -575,55 +676,94 @@ writeString(term_t t, write_options *options)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-Formatting a float. This is very  complicated   as  we must write floats
-such that it can be read as a float. This means using the conventions of
-the C locale and if the float happens to be integer as <int>.0.
+Formatting a float. This used  to  use   sprintf(),  but  there  are two
+problems with this. First of all, this uses the current locale, which is
+complicated to avoid. Second, it does not provide a mode that guarantees
+reliable read-back. Using %g gets closest,   but %.15g doesn't guarantee
+read-back and %.17g does, but prints 0.1 as 0.100..001, etc.
 
-Switching the locale is no option as  locale handling is not thread-safe
-and may have unwanted  consequences  for   embedding.  There  is  a intptr_t
-discussion on the very same topic on  the Python mailinglist. Many hacks
-are proposed, none is very satisfactory.   Richard  O'Keefe suggested to
-use ecvt(), fcvt() and gcvt(). These  are   not  thread-safe.  The GNU C
-library provides *_r() variations that  can   do  the  trick. An earlier
-patch used localeconv() to find the  decimal   point,  but  this is both
-complicated and not thread-safe.
+This uses dtoa.c. See pl-dtoa.c for how this is packed into SWI-Prolog.
 
-Finally, with help of Richard we decided  to replace the first character
-that is not a digit nor [eE], as this must be the decimal point.
+TBD: The number of cases are large. We should see whether it is possible
+to clean this up a bit. The 5 cases   as  such are real: there is no way
+around these.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 char *
-format_float(double f, char *buf, const char *format)
-{ char *q;
+format_float(double f, char *buf)
+{ char *end, *o=buf;
+  int decpt, sign;
+  char *s = dtoa(f, 0, 30, &decpt, &sign, &end);
 
-  sprintf(buf, format, f);
+  DEBUG(2, Sdprintf("decpt=%d, sign=%d, len = %d, '%s'\n",
+		    decpt, sign, end-s, s));
 
-  q = buf;
-  if ( *q == '-' )			/* skip -?[0-9]* */
-    q++;
-  while(*q && (isDigit(*q) || *q <= ' '))
-    q++;
+  if ( sign )
+    *o++ = '-';
 
-  switch( *q )
-  { case '\0':
-      *q++ = '.';
-      *q++ = '0';
-      *q = EOS;
-      break;
-    case 'e':
-    case 'E':
-      break;
-    default:
-      *q = '.';
+  if ( decpt <= 0 )			/* decimal dot before */
+  { if ( decpt <= -4 )
+    { *o++ = s[0];
+      *o++ = '.';
+      if ( end-s > 1 )
+      { memcpy(o, s+1, end-s-1);
+	o += end-s-1;
+      } else
+	*o++ = '0';
+      sprintf(o, "e%d", decpt-1);
+    } else
+    { int i;
+
+      *o++ = '0';
+      *o++ = '.';
+      for(i=0; i < -decpt; i++)
+	*o++ = '0';
+      memcpy(o, s, end-s);
+      o[end-s] = 0;
+    }
+  } else if ( end-s > decpt )		/* decimal dot inside */
+  { memcpy(o, s, decpt);
+    o += decpt;
+    *o++ = '.';
+    memcpy(o, s+decpt, end-s-decpt);
+    o[end-s-decpt] = 0;
+  } else				/* decimal dot after */
+  { int i;
+    int trailing = decpt-(int)(end-s);
+
+    if ( decpt > 15 )			/* over precision: use eE */
+    { *o++ = s[0];
+      *o++ = '.';
+      if ( end-s > 1 )
+      { trailing += (int)(end-s)-1;
+	memcpy(o, s+1, end-s-1);
+	o += end-s-1;
+      } else
+	*o++ = '0';
+      sprintf(o, "e+%d", trailing);
+    } else				/* within precision trail with .0 */
+    { memcpy(o, s, end-s);
+      o += end-s;
+
+      for(i=(int)(end-s); i<decpt; i++)
+	*o++ = '0';
+      *o++ = '.';
+      *o++ = '0';
+      *o = 0;
+    }
   }
 
+  freedtoa(s);
+
   return buf;
 }
 
 
 static bool
 WriteNumber(Number n, write_options *options)
-{ switch(n->type)
+{ GET_LD
+
+  switch(n->type)
   { case V_INTEGER:
     { char buf[32];
 
@@ -642,7 +782,14 @@ WriteNumber(Number n, write_options *options)
       else
 	buf = PL_malloc(sz);
 
-      mpz_get_str(buf, 10, n->value.mpz);
+      /* mpz_get_str() can perform large intermediate allocations :-( */
+      EXCEPTION_GUARDED({ LD->gmp.persistent++;
+			  mpz_get_str(buf, 10, n->value.mpz);
+			  LD->gmp.persistent--;
+			},
+			{ LD->gmp.persistent--;
+			  rc = PL_rethrow();
+			})
       rc = PutToken(buf, options->out);
       if ( buf != tmp )
 	PL_free(buf);
@@ -662,7 +809,8 @@ WriteNumber(Number n, write_options *options)
 
 static bool
 writePrimitive(term_t t, write_options *options)
-{ double f;
+{ GET_LD
+  double f;
   atom_t a;
   char buf[32];
   IOSTREAM *out = options->out;
@@ -749,7 +897,7 @@ writePrimitive(term_t t, write_options *options)
     } else
     { char buf[100];
 
-      format_float(f, buf, LD->float_format);
+      format_float(f, buf);
 
       return PutToken(buf, out);
     }
@@ -765,48 +913,113 @@ writePrimitive(term_t t, write_options *options)
 }
 
 
-word
-pl_nl1(term_t stream)
+static int
+pl_nl__LD(term_t stream ARG_LD)
 { IOSTREAM *s;
 
-  if ( getOutputStream(stream, &s) )
+  if ( getTextOutputStream(stream, &s) )
   { Sputcode('\n', s);
     return streamStatus(s);
   }
 
-  fail;
+  return FALSE;
 }
 
-word
-pl_nl()
-{ return pl_nl1(0);
+
+static
+PRED_IMPL("nl", 1, nl, 0)
+{ PRED_LD
+
+  return pl_nl__LD(A1 PASS_LD);
+}
+
+static
+PRED_IMPL("nl", 0, nl, 0)
+{ PRED_LD
+
+  return pl_nl__LD(0 PASS_LD);
 }
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Call user:portray/1 if defined.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
-static bool
-callPortray(term_t arg, write_options *options)
-{ predicate_t portray;
+static int
+put_write_options(term_t opts_in, write_options *options)
+{ GET_LD
+  term_t newlist = PL_new_term_ref();
+  term_t precopt = PL_new_term_ref();
+  fid_t fid = PL_open_foreign_frame();
+  term_t head = PL_new_term_ref();
+  term_t tail = PL_copy_term_ref(opts_in);
+  term_t newhead = PL_new_term_ref();
+  term_t newtail = PL_copy_term_ref(newlist);
+  int rc = TRUE;
+
+  while(rc && PL_get_list(tail, head, tail))
+  { if ( !PL_is_functor(head, FUNCTOR_priority1) )
+      rc = ( PL_unify_list(newtail, newhead, newtail) &&
+	     PL_unify(newhead, head) );
+  }
+
+  if ( rc )
+  { rc = ( PL_unify_list(newtail, head, newtail) &&
+	   PL_unify_functor(head, FUNCTOR_priority1) &&
+	   PL_get_arg(1, head, precopt) &&
+	   PL_unify_nil(newtail) );
+  }
+  if ( rc )
+  { options->write_options = newlist;
+    options->prec_opt = precopt;
+  }
+
+  PL_close_foreign_frame(fid);
+  return rc;
+}
+
+
+
+
+static int
+callPortray(term_t arg, int prec, write_options *options)
+{ predicate_t pred;
 
   if ( GD->cleaning > CLN_PROLOG )
     fail;				/* avoid dangerous callbacks */
 
-  portray = _PL_predicate("portray", 1, "user", &GD->procedures.portray);
+  if ( options->portray_goal )
+  { pred = _PL_predicate("call", 3, "user", &GD->procedures.call3);
+  } else
+  { pred = _PL_predicate("portray", 1, "user", &GD->procedures.portray);
+    if ( !pred->definition->definition.clauses )
+      return FALSE;
+  }
 
-  if ( portray->definition->definition.clauses )
-  { fid_t wake    = saveWakeup(PASS_LD1);
-    fid_t fid     = PL_open_foreign_frame();
+  { GET_LD
+    wakeup_state wstate;
     IOSTREAM *old = Scurout;
     int rval;
+    term_t av;
 
+    if ( !saveWakeup(&wstate, TRUE PASS_LD) )
+      return FALSE;
     Scurout = options->out;
-    rval = PL_call_predicate(NULL, PL_Q_NODEBUG, portray, arg);
-    Scurout = old;
+    if ( options->portray_goal )
+    { av = PL_new_term_refs(3);
 
-    PL_discard_foreign_frame(fid);
-    restoreWakeup(wake PASS_LD);
+      PL_put_term(av+0, options->portray_goal);
+      PL_put_term(av+1, arg);
+      PL_unify_integer(options->prec_opt, prec);
+      PL_put_term(av+2, options->write_options);
+    } else
+    { av = arg;
+    }
+    rval = PL_call_predicate(NULL, PL_Q_NODEBUG|PL_Q_PASS_EXCEPTION,
+			     pred, av);
+    if ( !rval && PL_exception(0) )
+      rval = -1;
+    Scurout = old;
+    restoreWakeup(&wstate PASS_LD);
 
     return rval;
   }
@@ -816,45 +1029,50 @@ callPortray(term_t arg, write_options *options)
 
 
 static bool
-writeTerm(term_t t, int prec, write_options *options)
-{ int rval;
+writeArgTerm(term_t t, int prec, write_options *options, bool arg)
+{ GET_LD
+  int rval;
   int levelSave = options->depth;
-  fid_t fid = PL_open_foreign_frame();
+  fid_t fid;
 
-  if ( ++options->depth > options->max_depth && options->max_depth )
-    rval = PutString("...", options->out);
-  else if ( PL_is_compound(t) )
-  { visited v;
+  if ( !(fid = PL_open_foreign_frame()) )
+    return FALSE;
 
-    v.address = address_of(t);
-    if ( has_visited(options->visited, v.address) )
-    { rval = PutString("**", options->out);
-    } else
-    { v.next = options->visited;
-      options->visited = &v;
-      rval = writeTerm2(t, prec, options);
-      options->visited = v.next;
-    }
-  } else
-  { rval = writeTerm2(t, prec, options);
+  if ( PL_handle_signals() < 0 )
+  { rval = FALSE;
+    goto out;
   }
 
+  if ( ++options->depth > options->max_depth && options->max_depth )
+  { PutOpenToken('.', options->out);
+    rval = PutString("...", options->out);
+  } else
+  { rval = writeTerm2(t, prec, options, arg);
+  }
+
+out:
   options->depth = levelSave;
   PL_close_foreign_frame(fid);
 
   return rval;
 }
 
+static bool
+writeTerm(term_t t, int prec, write_options *options)
+{
+  return writeArgTerm(t, prec, options, FALSE);
+}
 
 static bool
-writeList2(term_t list, write_options *options, int cyclic)
-{ term_t head = PL_new_term_ref();
+writeList(term_t list, write_options *options)
+{ GET_LD
+  term_t head = PL_new_term_ref();
   term_t l    = PL_copy_term_ref(list);
 
   TRY(Putc('[', options->out));
   for(;;)
   { PL_get_list(l, head, l);
-    TRY(writeTerm(head, 999, options));
+    TRY(writeArgTerm(head, 999, options, TRUE));
 
     if ( PL_get_nil(l) )
       break;
@@ -862,24 +1080,11 @@ writeList2(term_t list, write_options *options, int cyclic)
       return PutString("|...]", options->out);
     if ( !PL_is_functor(l, FUNCTOR_dot2) )
     { TRY(Putc('|', options->out));
-      TRY(writeTerm(l, 999, options));
+      TRY(writeArgTerm(l, 999, options, TRUE));
       break;
     }
 
-					/* cycle detection */
-    { Word addr =  address_of(l);
-
-      if ( has_visited(options->visited, addr) )
-      { return PutString("|**]", options->out);
-      } else if ( cyclic )
-      { visited *v = alloca(sizeof(*v));
-	v->address = addr;
-	v->next = options->visited;
-	options->visited = v;
-      }
-    }
-
-    TRY(PutString(", ", options->out));
+    TRY(PutComma(options));
   }
 
   return Putc(']', options->out);
@@ -887,34 +1092,28 @@ writeList2(term_t list, write_options *options, int cyclic)
 
 
 static bool
-writeList(term_t list, write_options *options)
-{ visited *v = options->visited;
-  Word tail;
-  int rc;
-
-  skip_list(valTermRef(list), &tail PASS_LD);
-  rc = writeList2(list, options, isList(*tail));
-  options->visited = v;
-
-  return rc;
-}
-
-
-static bool
-writeTerm2(term_t t, int prec, write_options *options)
-{ atom_t functor;
+writeTerm2(term_t t, int prec, write_options *options, bool arg)
+{ GET_LD
+  atom_t functor;
   int arity, n;
   int op_type, op_pri;
   atom_t a;
   IOSTREAM *out = options->out;
 
   if ( !PL_is_variable(t) &&
-       true(options, PL_WRT_PORTRAY) &&
-       callPortray(t, options) )
-    succeed;
+       true(options, PL_WRT_PORTRAY) )
+  { switch( callPortray(t, prec, options) )
+    { case TRUE:
+	return TRUE;
+      case FALSE:
+	break;
+      default:
+	return FALSE;
+    }
+  }
 
   if ( PL_get_atom(t, &a) )
-  { if ( priorityOperator(NULL, a) > prec )
+  { if ( !arg && prec < 1200 && priorityOperator(NULL, a) > 0 )
     { if ( PutOpenBrace(out) &&
 	   writeAtom(a, options) &&
 	   PutCloseBrace(out) )
@@ -926,54 +1125,23 @@ writeTerm2(term_t t, int prec, write_options *options)
   if ( !PL_get_name_arity(t, &functor, &arity) )
   { return writePrimitive(t, options);
   } else
-  { if ( functor == ATOM_isovar &&			/* $VAR/1 */
-	 true(options, PL_WRT_NUMBERVARS) )
-    { int n;
-      atom_t a;
-      term_t arg = PL_new_term_ref();
-
-      PL_get_arg(1, t, arg);
-      if ( PL_get_integer(arg, &n) && n >= 0 )
-      { int i = n % 26;
-	int j = n / 26;
-	char buf[16];
-
-	if ( j == 0 )
-	{ buf[0] = i+'A';
-	  buf[1] = EOS;
-	} else
-	{ sprintf(buf, "%c%d", i+'A', j);
-	}
-
-	return PutToken(buf, out);
-      }
-      if ( PL_get_atom(arg, &a) )
-      { write_options o2 = *options;
-	clear(&o2, PL_WRT_QUOTED);
-
-	return writeAtom(a, &o2);
+  { if ( true(options, PL_WRT_NUMBERVARS) )
+    { switch( writeNumberVar(t, options PASS_LD) )
+      { case -1:
+	  return FALSE;
+	case TRUE:
+	  return TRUE;
       }
     }
 
-
     if ( false(options, PL_WRT_IGNOREOPS) )
     { term_t arg = PL_new_term_ref();
 
       if ( arity == 1 )
       { if ( functor == ATOM_curl )	/* {a,b,c} */
-	{ term_t a = PL_new_term_ref();
-
-	  PL_get_arg(1, t, arg);
+	{ _PL_get_arg(1, t, arg);
 	  TRY(Putc('{', out));
-	  for(;;)
-	  { if ( !PL_is_functor(arg, FUNCTOR_comma2) )
-	      break;
-	    PL_get_arg(1, arg, a);
-	    TRY(writeTerm(a, 999, options) &&
-		PutString(", ", out));
-	    PL_get_arg(2, arg, arg);
-	  }
-	  TRY(writeTerm(arg, 999, options) &&
+	  TRY(writeTerm(arg, 1200, options) &&
 	      Putc('}', out));
 
 	  succeed;
@@ -987,26 +1155,21 @@ writeTerm2(term_t t, int prec, write_options *options)
 
 	  embrace = ( op_pri > prec );
 
-	  PL_get_arg(1, t, arg);
+	  _PL_get_arg(1, t, arg);
 	  if ( embrace )
-	  { TRY(PutOpenBrace(out));
-	  }
+	    TRY(PutOpenBrace(out));
 	  TRY(writeAtom(functor, options));
 
 				/* +/-(Number) : avoid parsing as number */
-	  if ( (functor == ATOM_minus || functor == ATOM_plus) &&
-	       PL_is_number(arg) )
-	  { TRY(Putc('(', out));
-	    TRY(writeTerm(arg, 999, options));
-	    TRY(Putc(')', out));
-	  } else
-	  { TRY(writeTerm(arg,
-			  op_type == OP_FX ? op_pri-1 : op_pri,
-			  options));
-	  }
+	  if ( functor == ATOM_minus || functor == ATOM_plus )
+	    options->out->lastc = PREFIX_SIGN;
+
+	  TRY(writeTerm(arg,
+			op_type == OP_FX ? op_pri-1 : op_pri,
+			options));
+
 	  if ( embrace )
-	  { TRY(PutCloseBrace(out));
-	  }
+	   TRY(PutCloseBrace(out));
 
 	  succeed;
 	}
@@ -1016,7 +1179,7 @@ writeTerm2(term_t t, int prec, write_options *options)
 			     &op_type, &op_pri) )
 	{ term_t arg = PL_new_term_ref();
 
-	  PL_get_arg(1, t, arg);
+	  _PL_get_arg(1, t, arg);
 	  if ( op_pri > prec )
 	    TRY(PutOpenBrace(out));
 	  TRY(writeTerm(arg,
@@ -1038,8 +1201,8 @@ writeTerm2(term_t t, int prec, write_options *options)
 	{ term_t l = PL_new_term_ref();
 	  term_t r = PL_new_term_ref();
 
-	  PL_get_arg(1, t, l);
-	  PL_get_arg(2, t, r);
+	  _PL_get_arg(1, t, l);
+	  _PL_get_arg(2, t, r);
 
 	  if ( op_pri > prec )
 	    TRY(PutOpenBrace(out));
@@ -1048,7 +1211,7 @@ writeTerm2(term_t t, int prec, write_options *options)
 				? op_pri-1 : op_pri,
 			options));
 	  if ( functor == ATOM_comma )
-	  { TRY(PutString(", ", out));
+	  { TRY(PutComma(options));
 	  } else
 	  { switch(writeAtom(functor, options))
 	    { case FALSE:
@@ -1067,6 +1230,10 @@ writeTerm2(term_t t, int prec, write_options *options)
 	}
       }
     }
+
+    if ( (options->flags&PL_WRT_LIST) && arity == 2 && functor == ATOM_dot )
+      return writeList(t, options);
+
 					/* functor(<args> ...) */
     { term_t a = PL_new_term_ref();
 
@@ -1074,9 +1241,9 @@ writeTerm2(term_t t, int prec, write_options *options)
 	  Putc('(', out));
       for(n=0; n<arity; n++)
       { if (n > 0)
-	  TRY(PutString(", ", out));
-	PL_get_arg(n+1, t, a);
-	TRY(writeTerm(a, 999, options));
+	  TRY(PutComma(options));
+	_PL_get_arg(n+1, t, a);
+	TRY(writeArgTerm(a, 999, options, TRUE));
       }
       return Putc(')', out);
     }
@@ -1084,6 +1251,88 @@ writeTerm2(term_t t, int prec, write_options *options)
 }
 
 
+		 /*******************************
+		 *	  CYCLE HANDLING	*
+		 *******************************/
+
+static int
+reunify_acyclic_substitutions(term_t substitutions, term_t cycles,
+			      write_options *options)
+{ GET_LD
+  term_t s_tail, c_tail, s_head, c_head, var, value;
+  intptr_t count = 0;
+
+  if ( !(s_tail = PL_copy_term_ref(substitutions)) ||
+       !(c_tail = PL_copy_term_ref(cycles)) ||
+       !(s_head = PL_new_term_ref()) ||
+       !(c_head = PL_new_term_ref()) ||
+       !(var    = PL_new_term_ref()) ||
+       !(value  = PL_new_term_ref()) )
+    return FALSE;
+
+  while(PL_get_list(s_tail, s_head, s_tail))
+  { _PL_get_arg(1, s_head, var);
+    _PL_get_arg(2, s_head, value);
+    if ( PL_var_occurs_in(var, value) )
+    { if ( (options->flags&PL_WRT_NUMBERVARS) )
+      { if ( !PL_unify_term(var,
+			    PL_FUNCTOR, FUNCTOR_isovar1,
+			      PL_INTPTR, --count) )
+	  return FALSE;
+      }
+
+      if ( !PL_unify_list(c_tail, c_head, c_tail) ||
+	   !PL_unify(c_head, s_head) )
+	return FALSE;
+    } else
+    { if ( !PL_unify(var, value) )
+	return FALSE;
+    }
+  }
+
+  return PL_unify_nil(c_tail);
+}
+
+
+static int
+writeTopTerm(term_t term, int prec, write_options *options)
+{ GET_LD
+  int rc;
+
+  if ( (!(options->flags&PL_WRT_NO_CYCLES) && options->max_depth) ||
+       PL_is_acyclic(term) )
+  { rc = writeTerm(term, prec, options);
+  } else
+  { fid_t fid;
+    term_t template, substitutions, cycles, at_term;
+
+    if ( options->flags & PL_WRT_NO_CYCLES )
+      return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_cyclic_term, term);
+
+    if ( !(fid = PL_open_foreign_frame()) ||
+	 !(template = PL_new_term_ref()) ||
+	 !(substitutions = PL_new_term_ref()) ||
+	 !(cycles = PL_new_term_ref()) ||
+	 !(at_term = PL_new_term_ref()) ||
+	 !PL_factorize_term(term, template, substitutions) ||
+	 !reunify_acyclic_substitutions(substitutions, cycles, options) ||
+	 !PL_unify_term(at_term,
+			PL_FUNCTOR, FUNCTOR_xpceref2,
+			  PL_TERM, template,
+			  PL_TERM, cycles) )
+      return FALSE;
+    rc = writeTerm(at_term, prec, options);
+    PL_discard_foreign_frame(fid);
+  }
+
+  return rc;
+}
+
+
+		 /*******************************
+		 *	      TOPLEVEL		*
+		 *******************************/
+
 int
 writeAttributeMask(atom_t a)
 { if ( a == ATOM_ignore )
@@ -1099,11 +1348,23 @@ writeAttributeMask(atom_t a)
 }
 
 
+static int
+writeBlobMask(atom_t a)
+{ if ( a == ATOM_default )
+  { return 0;
+  } else if ( a == ATOM_portray )
+  { return PL_WRT_BLOB_PORTRAY;
+  } else
+    return -1;
+}
+
+
 static const opt_spec write_term_options[] =
 { { ATOM_quoted,	    OPT_BOOL },
   { ATOM_ignore_ops,	    OPT_BOOL },
   { ATOM_numbervars,        OPT_BOOL },
   { ATOM_portray,           OPT_BOOL },
+  { ATOM_portray_goal,      OPT_TERM },
   { ATOM_character_escapes, OPT_BOOL },
   { ATOM_max_depth,	    OPT_INT  },
   { ATOM_module,	    OPT_ATOM },
@@ -1111,30 +1372,40 @@ static const opt_spec write_term_options[] =
   { ATOM_attributes,	    OPT_ATOM },
   { ATOM_priority,	    OPT_INT },
   { ATOM_partial,	    OPT_BOOL },
-  { NULL_ATOM,	     	    0 }
+  { ATOM_spacing,	    OPT_ATOM },
+  { ATOM_blobs,		    OPT_ATOM },
+  { ATOM_cycles,	    OPT_BOOL },
+  { NULL_ATOM,		    0 }
 };
 
 word
 pl_write_term3(term_t stream, term_t term, term_t opts)
-{ bool quoted     = FALSE;
+{ GET_LD
+  bool quoted     = FALSE;
   bool ignore_ops = FALSE;
   bool numbervars = -1;			/* not set */
   bool portray    = FALSE;
+  term_t gportray = 0;
   bool bqstring   = truePrologFlag(PLFLAG_BACKQUOTED_STRING);
   bool charescape = -1;			/* not set */
   atom_t mname    = ATOM_user;
   atom_t attr     = ATOM_nil;
+  atom_t blobs    = ATOM_nil;
   int  priority   = 1200;
   bool partial    = FALSE;
+  bool cycles     = TRUE;
   IOSTREAM *s;
   write_options options;
+  int rc;
 
   memset(&options, 0, sizeof(options));
+  options.spacing = ATOM_standard;
 
   if ( !scan_options(opts, 0, ATOM_write_option, write_term_options,
-		     &quoted, &ignore_ops, &numbervars, &portray,
+		     &quoted, &ignore_ops, &numbervars, &portray, &gportray,
 		     &charescape, &options.max_depth, &mname,
-		     &bqstring, &attr, &priority, &partial) )
+		     &bqstring, &attr, &priority, &partial, &options.spacing,
+		     &blobs, &cycles) )
     fail;
 
   if ( attr == ATOM_nil )
@@ -1147,19 +1418,43 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
 
     options.flags |= mask;
   }
+  if ( blobs != ATOM_nil )
+  { int mask = writeBlobMask(blobs);
+
+    if ( mask < 0 )
+      return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_write_option, opts);
+
+    options.flags |= mask;
+  }
   if ( priority < 0 || priority > OP_MAXPRIORITY )
   { term_t t = PL_new_term_ref();
     PL_put_integer(t, priority);
 
     return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_operator_priority, t);
   }
-  if ( !getOutputStream(stream, &s) )
-    fail;
+  switch( options.spacing )
+  { case ATOM_standard:
+    case ATOM_next_argument:
+      break;
+    default:
+    { term_t t = PL_new_term_ref();
+      PL_put_atom(t, options.spacing);
+
+      return PL_error(NULL, 0, NULL, ERR_DOMAIN, ATOM_spacing, t);
+    }
+  }
 
   options.module = lookupModule(mname);
   if ( charescape == TRUE ||
        (charescape == -1 && true(options.module, CHARESCAPE)) )
     options.flags |= PL_WRT_CHARESCAPES;
+  if ( gportray )
+  { options.portray_goal = gportray;
+    if ( !put_write_options(opts, &options) ||
+	 !PL_qualify(options.portray_goal, options.portray_goal) )
+      return FALSE;
+    portray = TRUE;
+  }
   if ( numbervars == -1 )
     numbervars = (portray ? TRUE : FALSE);
 
@@ -1168,19 +1463,22 @@ pl_write_term3(term_t stream, term_t term, term_t opts)
   if ( numbervars ) options.flags |= PL_WRT_NUMBERVARS;
   if ( portray )    options.flags |= PL_WRT_PORTRAY;
   if ( bqstring )   options.flags |= PL_WRT_BACKQUOTED_STRING;
+  if ( !cycles )    options.flags |= PL_WRT_NO_CYCLES;
 
+  if ( !getTextOutputStream(stream, &s) )
+    fail;
   options.out = s;
   if ( !partial )
     PutOpenToken(EOF, s);		/* reset this */
   if ( (options.flags & PL_WRT_QUOTED) && !(s->flags&SIO_REPPL) )
   { s->flags |= SIO_REPPL;
-    writeTerm(term, priority, &options);
+    rc = writeTopTerm(term, priority, &options);
     s->flags &= ~SIO_REPPL;
   } else
-  { writeTerm(term, priority, &options);
+  { rc = writeTopTerm(term, priority, &options);
   }
 
-  return streamStatus(s);
+  return streamStatus(s) && rc;
 }
 
 
@@ -1200,16 +1498,18 @@ PL_write_term(IOSTREAM *s, term_t term, int precedence, int flags)
   options.module    = MODULE_user;
 
   PutOpenToken(EOF, s);			/* reset this */
-  return writeTerm(term, precedence, &options);
+  return writeTopTerm(term, precedence, &options);
 }
 
 
 static word
 do_write2(term_t stream, term_t term, int flags)
-{ IOSTREAM *s;
+{ GET_LD
+  IOSTREAM *s;
 
-  if ( getOutputStream(stream, &s) )
+  if ( getTextOutputStream(stream, &s) )
   { write_options options;
+    int rc;
 
     memset(&options, 0, sizeof(options));
     options.flags     = flags;
@@ -1221,18 +1521,18 @@ do_write2(term_t stream, term_t term, int flags)
       options.flags |= PL_WRT_BACKQUOTED_STRING;
 
     PutOpenToken(EOF, s);		/* reset this */
-    writeTerm(term, 1200, &options);
+    rc = writeTopTerm(term, 1200, &options);
 
-    return streamStatus(s);
+    return streamStatus(s) && rc;
   }
 
-  fail;
+  return FALSE;
 }
 
 
 word
 pl_write2(term_t stream, term_t term)
-{ return do_write2(stream, term, 0);
+{ return do_write2(stream, term, PL_WRT_NUMBERVARS);
 }
 
 word
@@ -1248,17 +1548,23 @@ pl_print2(term_t stream, term_t term)
 
 word
 pl_write_canonical2(term_t stream, term_t term)
-{ fid_t fid = PL_open_foreign_frame();
+{ GET_LD
   nv_options options;
   word rc;
 
+  BEGIN_NUMBERVARS();
+
   options.functor = FUNCTOR_isovar1;
   options.on_attvar = AV_SKIP;
-  options.singletons = TRUE;
-  numberVars(term, &options, 0 PASS_LD);
-  rc = do_write2(stream, term,
-		 PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS);
-  PL_discard_foreign_frame(fid);
+  options.singletons = PL_is_acyclic(term);
+  options.numbered_check = FALSE;
+
+  rc = ( numberVars(term, &options, 0 PASS_LD) >= 0 &&
+	 do_write2(stream, term,
+		   PL_WRT_QUOTED|PL_WRT_IGNOREOPS|PL_WRT_NUMBERVARS)
+       );
+
+  END_NUMBERVARS();
 
   return rc;
 }
@@ -1285,7 +1591,9 @@ pl_write_canonical(term_t term)
 
 word					/* for debugging purposes! */
 pl_writeln(term_t term)
-{ if ( pl_write2(0, term) && pl_nl() )
+{ if ( PL_write_term(Serror, term, 1200,
+		     PL_WRT_QUOTED|PL_WRT_NUMBERVARS) &&
+       Sdprintf("\n") >= 0 )
     succeed;
 
   fail;
@@ -1310,12 +1618,46 @@ PRED_IMPL("$put_token", 2, put_token, 0)
   fail;
 }
 
+/** '$put_quoted_codes'(+Stream, +Quote, +Codes, +Options)
+
+Emit Codes using the escaped character  syntax,   but  does not emit the
+start and end-code itself. Options is  currently ignored. It is intended
+to provide additional preferences, so as using \uXXXX, \UXXXXXXXX, etc.
+*/
+
+static
+PRED_IMPL("$put_quoted", 4, put_quoted_codes, 0)
+{ IOSTREAM *out;
+  pl_wchar_t *w;
+  size_t i, len;
+  int quote;
+  int flags = PL_WRT_CHARESCAPES;
+  int rc = TRUE;
+
+  if ( !PL_get_stream_handle(A1, &out) ||
+       !PL_get_char_ex(A2, &quote, FALSE) ||
+       !PL_get_wchars(A3, &len, &w, CVT_LIST|CVT_STRING|CVT_EXCEPTION) )
+    return FALSE;
+
+  for(i=0; rc && i<len; i++)
+    rc = putQuoted(w[i], quote, flags, out);
+
+  if ( rc )
+    rc = PL_release_stream(out);
+
+  return rc;
+}
+
+
 
 		 /*******************************
 		 *      PUBLISH PREDICATES	*
 		 *******************************/
 
 BeginPredDefs(write)
+  PRED_DEF("nl", 0, nl, 0)
+  PRED_DEF("nl", 1, nl, 0)
   PRED_DEF("$put_token", 2, put_token, 0)
+  PRED_DEF("$put_quoted", 4, put_quoted_codes, 0)
 EndPredDefs
 
