@@ -1,4 +1,4 @@
-/*  $Id: rc.h,v 1.15 2007/02/20 11:17:33 jan Exp $
+/*  $Id$
 
     Part of SWI-Prolog
 
@@ -25,6 +25,10 @@
 #ifndef RC_H_INCLUDED
 #define RC_H_INCLUDED
 
+#if defined(__MINGW32__) && !defined(__WINDOWS__)
+#define __WINDOWS__ 1
+#endif
+
 #ifdef __WINDOWS__
 #  ifdef WIN64
 #    include "../config/win64.h"
@@ -40,13 +44,17 @@
 #  else
      typedef void *WIN_HANDLE;
 #  endif
-#  if (_MSC_VER < 1300)
+#  if defined(_MSC_VER)
+#    if (_MSC_VER < 1300)
 typedef long intptr_t;
 typedef unsigned long uintptr_t;
-#  else
-#    include <stddef.h>
-#  endif
+#    else
+#      include <stddef.h>
+#    endif
 typedef intptr_t ssize_t;		/* signed version of size_t */
+#  elif defined(__MINGW32__)
+#    include <stdint.h>
+#  endif
 #else
 #include <config.h>
 #include <unistd.h>
@@ -57,7 +65,7 @@ typedef intptr_t ssize_t;		/* signed version of size_t */
 #include <dmalloc.h>			/* Use www.dmalloc.com debugger */
 #endif
 #ifdef HAVE_UXNT_H
-#include "../win32/uxnt/uxnt.h"
+#include "../os/windows/uxnt.h"
 #endif
 
 #include <stdio.h>
