@@ -3,9 +3,9 @@
     Part of SWI-Prolog
 
     Author:        Jan Wielemaker and Anjo Anjewierden
-    E-mail:        wielemak@science.uva.nl
+    E-mail:        J.Wielemaker@cs.vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (C): 1985-2007, University of Amsterdam
+    Copyright (C): 1985-2009, University of Amsterdam
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Lesser General Public
@@ -26,7 +26,7 @@
 %:- set_prolog_flag(trace_gc, true).
 :- asserta(user:file_search_path(library, '../packages/plunit')).
 :- [library(plunit)].
-:- set_test_options([load(always), silent(true), sto(true)]).
+:- set_test_options([load(always), silent(true), sto(true), cleanup(true)]).
 
 :- use_module(library(lists)).
 
@@ -60,36 +60,6 @@ available test sets. The public goals are:
 
 :- setenv('TZ', 'CET').
 
-		 /*******************************
-		 *    CONDITIONAL COMPILATION	*
-		 *******************************/
-
-:- dynamic
-	user:include_code/1.
-
-term_expansion((:- if(G)), []) :-
-	prolog_load_context(module, M),
-	(   catch(M:G, E, (print_message(error, E), fail))
-	->  asserta(user:include_code(true))
-	;   asserta(user:include_code(false))
-	).
-term_expansion((:- else), []) :-
-	(   retract(user:include_code(X))
-	->  (   X == true
-	    ->  X2 = false
-	    ;   X2 = true
-	    ),
-	    asserta(user:include_code(X2))
-	;   throw(error(context_error(no_if), _))
-	).
-term_expansion((:- endif), []) :-
-	retract(user:include_code(_)), !.
-
-term_expansion(_, []) :-
-	user:include_code(X), !,
-	X == false.
-
-
 		 /*******************************
 		 *	      SYNTAX		*
 		 *******************************/
@@ -127,6 +97,14 @@ syntax(char-2) :-
 	52 == 0'\x34.
 syntax(char-3) :-
 	"\\" =:= 0'\\.
+syntax(char-4) :-
+	1-48 == 1-0'0.
+syntax(cannot_start_term-1) :-
+	catch(term_to_atom(_T, 'p(]'), E, true),
+	E = error(syntax_error(cannot_start_term), _).
+
+
+:- op(100, yf, af).
 
 syntax(string-1) :-
 	'\c ' == ''.
@@ -140,26 +118,51 @@ syntax(quote-3) :-
 	'\x61\' == a.
 syntax(quote-4) :-
 	char_code('\'', 39).
+syntax(quote-5) :-
+	0'\' == 0''.
+syntax(quote-6) :-
+	0'\' == 0'''.
+syntax(quote-7) :-
+	atom_to_term('\'\\\n\'', T, _),
+	T == ''.
+syntax(base-1) :-
+	1'+'1 == 1+1.
+syntax(base-2) :-
+	16'af == 175.
+syntax(base-3) :-
+	10'af' == af(10).
+syntax(base-4) :-
+	_0 = 1, B is _0'+'1, B == 2.
+syntax(base-5) :-
+	A is 1.0e+0'+'1, A == 2.0.
 syntax(number-2) :-
 	catch(atom_to_term('2\'', _, _), E, true),
-	E = error(syntax_error(illegal_number), _).
+	E = error(syntax_error(end_of_file), _).
 syntax(zero-1) :-
-	term_to_atom(T, 'hello(\000\"\000\x")'),
+	term_to_atom(T, 'hello("\000\x")'),
 	T == hello([0, 120]).
 syntax(latin-1) :-
 	atom_codes(A, [247]),
 	atom_to_term(A, T, []),
 	atom_codes(T, [247]).
 
+:- op(0, yf, af).
+
 
 		 /*******************************
 		 *	       WRITE		*
 		 *******************************/
 
 write_test(q-1) :-
-	term_to_atom(-(0), X), X == '-(0)'.
+	T = -(0),
+	term_to_atom(T, X),
+	term_to_atom(T2, X),
+	T == T2.
 write_test(q-2) :-
-	term_to_atom(+(0), X), X == '+(0)'.
+	T = +(0),
+	term_to_atom(T, X),
+	term_to_atom(T2, X),
+	T == T2.
 write_test(q-3) :-
 	term_to_atom(+(a), X), X == '+a'.
 write_test(q-4) :-
@@ -169,12 +172,15 @@ write_test(q-5) :-
 write_test(q-6) :-
 	term_to_atom('*/*', X), X == '*/*'.
 write_test(q-7) :-
-	term_to_atom(p(0|a), X), X == 'p((0 \'|\' a))'.
+	term_to_atom(p((0|a)), X), X == 'p((0 \'|\' a))'.
 write_test(q-8) :-
-	term_to_atom(p(a|b), X), X == 'p((a\'|\'b))'.
+	term_to_atom(p((a|b)), X), X == 'p((a\'|\'b))'.
 write_test(c-1) :-
 	T = [a,b,c|T],
-	term_to_atom(T, X), X == '[a, b, c|**]'.
+	term_to_atom(T, X),
+	term_to_atom(@(T2,S2), X),
+	maplist(call, S2),
+	T2 =@= T.
 
 
 		 /*******************************
@@ -288,6 +294,22 @@ unifiable(unifiable-1) :-
 unifiable(unifiable-2) :-
 	unifiable(a(X,X), a(Y,Z), S),
 	S == [Z=X, Y=X].
+unifiable(gc-1) :-
+	trim_stacks,
+	(   between(2, 10, N),
+	    Len is 1<<N,
+	    numlist(1, Len, L1),
+	    attvar_list(Len, L2),
+	    unifiable(L1, L2, _Unifier),
+	    fail
+	;   true
+	).
+
+attvar_list(0, []) :- !.
+attvar_list(N, [H|T]) :-
+	freeze(H, integer(H)),
+	N2 is N - 1,
+	attvar_list(N2, T).
 
 
 		 /*******************************
@@ -340,9 +362,9 @@ arithmetic(arith-7) :-
 arithmetic(arith-8) :-
 	-6 is min(-6, -5.5).
 arithmetic(arith-9) :-
-	4000 =:= integer(10000 * float_fractional_part(1e10 + 0.4)).
+	4000 =:= integer(10000 * float_fractional_part(1.0e10 + 0.4)).
 arithmetic(arith-10) :-
-	-4000 =:= integer(10000 * float_fractional_part(-1e10 - 0.4)).
+	-4000 =:= integer(10000 * float_fractional_part(-1.0e10 - 0.4)).
 arithmetic(float_fractional_part-1) :-
 	0 is float_fractional_part(4).
 arithmetic(arith-11) :-
@@ -393,7 +415,7 @@ arithmetic(integer-2) :-
 arithmetic(int-1) :-
 	A is 1<<31, integer(A).
 arithmetic(cmp-1) :-
-	A is 100e6, 67 < A.
+	A is 100.0e6, 67 < A.
 
 
 		 /*******************************
@@ -415,10 +437,11 @@ floattest(float-1) :-
 floattest(float-2) :-
 	ftest.
 floattest(float-3) :-
-	erase_all(f),
-	recorda(f, 6.7),
+	erase_all(f),			% play safe
+	recorda(f, 6.7, Ref),
 	recorded(f, X),
-	X == 6.7.
+	X == 6.7,
+	erase(Ref).
 floattest(float-4) :-
 	X is 10.67,
 	X == 10.67.
@@ -436,49 +459,6 @@ floattest(float-8) :-
 	E = error(evaluation_error(float_overflow), _).
 
 
-		 /*******************************
-		 *	 PROLOG FUNCTIONS	*
-		 *******************************/
-
-:- arithmetic_function(ten/0).
-:- arithmetic_function(twice/1).
-:- arithmetic_function(mean/2).
-:- arithmetic_function(euler/0).
-:- arithmetic_function(fail/0).
-:- arithmetic_function(except/0).
-
-ten(10).
-twice(X, R) :-
-	R is X * 2.
-mean(X1, X2, R) :-
-	R is (X1 + X2)/2.
-
-euler(2.71828).
-
-fail(_) :- fail.
-
-except(_) :-
-	throw(error(foobar)).
-
-arithmetic_functions(func-1) :-
-	A is ten, A =:= 10.
-arithmetic_functions(func-2) :-
-	A is twice(5), A =:= 10.
-arithmetic_functions(func-3) :-
-	A is mean(0, 20), A =:= 10.
-arithmetic_functions(func-4) :-
-        Exp = 6*euler*7*1,		% test functions corrupting stack
-        EE is Exp,
-	EE =:= 6*euler*7*1.
-arithmetic_functions(fail-1) :-
-	catch(_ is fail, E, true),
-	compound(E), E = error(E2, _),
-	compound(E2), E2 = failure_error(_).
-arithmetic_functions(except-1) :-
-	catch(_ is except, E, true),
-	E == error(foobar).
-
-
 		 /*******************************
 		 *     UNBOUNDED ARITHMETIC	*
 		 *******************************/
@@ -512,13 +492,6 @@ ratp(Count, In, Out) :-
         T is In + (In rdiv 2),
 	ratp(Count0, T, Out).
 
-:- arithmetic_function(idiv/2).
-
-idiv(Dd,Dr,Iq):-
-        Q is Dd/Dr,
-        rational(Q,Qt,Qn),
-        Iq is Qt//Qn.
-
 dec(X, Y) :-
 	Y is X - 1.
 
@@ -544,17 +517,17 @@ gmp(sign-1) :-
 	-1 =:= sign(-X),
 	1 =:= sign(X).
 gmp(floor-1) :-
-	A is floor(1e20),
+	A is floor(1.0e20),
 	integer(A),
-	1e20 =:= float(A).
+	1.0e20 =:= float(A).
 gmp(floor-2) :-
 	0 is floor(9 rdiv 10),
 	-1 is floor(-1 rdiv 10),
 	-1 is floor(-9 rdiv 10).
 gmp(ceil-1) :-
-	A is ceil(1e20),
+	A is ceil(1.0e20),
 	integer(A),
-	1e20 =:= float(A).
+	1.0e20 =:= float(A).
 gmp(ceil-2) :-
 	1 is ceil(9 rdiv 10),
 	0 is ceil(-1 rdiv 10),
@@ -722,9 +695,6 @@ gmp(fmtf-1) :-
 	ratp(999, 1, X),
 	format(atom(S), '~5f', [X]),
 	sub_atom(S, _, _, 0, '935376.65824').
-gmp(idiv-1) :-
-	Qi is idiv(3 rdiv 2,2 rdiv 5),
-	Qi == 3.
 gmp(random) :-
 	A is random((1<<200)-((1<<200)-20)),
 	A < 20.
@@ -786,6 +756,8 @@ wchars(cmp-2) :-
 
 foo:hello(world).
 
+ten(10).
+
 meta(call-1) :-
 	call(ten(X)),
 	X == 10.
@@ -820,7 +792,8 @@ meta(call-11) :-
 	forall(A, true),
 	flag(a, 3, Old).
 meta(call-12) :-
-	catch(call(1), E, true),
+	G = 1,				% avoid in-line expansion
+	catch(call(G), E, true),
 	error(E, type_error(callable, _)).
 meta(apply-1) :-
 	apply(=, [a,a]).
@@ -873,7 +846,7 @@ cleanup(clean-4) :-
 	E == a.
 cleanup(clean-5) :-
 	catch(call_cleanup(throw(a), throw(b)), E, true),
-	E == b.
+	E == a.				% ISO Proposal
 cleanup(clean-6) :-
 	catch(call_cleanup(true, throw(b)), E, true),
 	E == b.
@@ -919,24 +892,24 @@ dl_fail(N) :-
 	NN is N - 1,
 	dl_fail(NN).
 
-:- arithmetic_function(fac/1).
-
 depth_limit(depth-1) :-
-	call_with_depth_limit(dl_det(1), 10, 1),
+	G = dl_det(1),
+	call_with_depth_limit(G, 10, 1),
 	deterministic(true).
 depth_limit(depth-2) :-
-	call_with_depth_limit(dl_det(10), 10, 10).
+	G = dl_det(10),
+	call_with_depth_limit(G, 10, 10).
 depth_limit(depth-3) :-
-	call_with_depth_limit(dl_det(10), 9, depth_limit_exceeded).
+	G = dl_det(10),
+	call_with_depth_limit(G, 9, depth_limit_exceeded).
 depth_limit(ndet-1) :-
+	G = dl_ndet(5),
 	findall(X,
-		call_with_depth_limit(dl_ndet(5), 10, X),
+		call_with_depth_limit(G, 10, X),
 		L),
 	L = [5, depth_limit_exceeded].
 depth_limit(fail-1) :-
 	\+ call_with_depth_limit(dl_fail(2), 10, _).
-depth_limit(arith-1) :-
-	call_with_depth_limit(_A is fac(10), 8, depth_limit_exceeded).
 
 
 		 /*******************************
@@ -1073,23 +1046,36 @@ sets(setof-2) :-
 		   compound(_A1, _B1)-[1, 2]]
 	).
 sets(vars-1) :-
-	'$e_free_variables'(A^satisfy(B^C^(setof(D:E,
-						 (country(E), area(E, D)),
-						 C),
-					   aggregate(max, C, B),
-					   in(B, A),
-					   {place(A)})),
-			    Free),
-	Free == v(D, E).
+	'$free_variable_set'(X^(m:Y^hello(X,Y)), G, V),
+	G == m:hello(X,Y),
+	V == v.
 sets(bagof-1) :-
 	List = [_,_,_],
 	bagof(X, member(X, List), Xs),
 	Xs == List.
-sets(setof-3) :-
+sets(bagof-2) :-
+	Goal = member(_, [1,2]),
+	bagof(Goal, Goal, Xs),
+	Xs == [member(1, [1,2]), member(2, [1,2])].
+sets(setof-3001) :-
+	List = [_,_,_],
+	setof(X, member(X, List), Xs),
+	Xs = List.
+sets(setof-3002) :-
 	List = [_,_,_],
 	setof(X, member(X, List), Xs),
 	sort(List, LS),
 	Xs == LS.
+sets(setof-4) :-
+	findall(f(X,Y,Z), setof(t,(X=Y;Y=Z;Z=X),_), Fs),
+	Fs = [_,_,_],
+	Ms = [f(X,X,_),f(_,Y,Y),f(X,_,X)],
+	\+ ( member(M, Ms), \+ ( member(F, Fs), F =@= M ) ).
+sets(setof-5) :- % Bart Demoen's example
+	findall(f(X,Z,A,B,Out), ( Given = [f(X,1),f(2,Z)], bagof(A,member(f(A,B),Given),Out) ), Fs),
+	Fs = [_,_],
+	Ms = [f(X, Z, A, Z, [2]), f(X, Z, A, 1, [X])],
+	\+ ( member(M, Ms), \+ ( member(F, Fs), F =@= M ) ).
 
 
 		 /*******************************
@@ -1101,9 +1087,9 @@ atom_handling(name-1) :-
 atom_handling(name-2) :-
 	name(V, "5"), V == 5.
 atom_handling(name-3) :-
-	name(V, "5e4"), V =:= 50000.
+	name(V, "5.0e4"), V =:= 50000.
 atom_handling(name-4) :-
-	name(V, "5e4a"), V == '5e4a'.
+	name(V, "5.0e4a"), V == '5.0e4a'.
 atom_handling(name-5) :-
 	name(V, ""), V == ''.
 
@@ -1123,6 +1109,10 @@ atom_handling(concat-5) :-
 atom_handling(concat-6) :-
 	atom_concat(X, a, a),
 	X == ''.
+atom_handling(concat-7) :-
+	atom_concat(X, Y, ''),
+	X == '',
+	Y == ''.
 
 atom_handling(number-1) :-
 	atom_number('42', X), X == 42.
@@ -1172,6 +1162,13 @@ string_handling(cmp-1) :-
 
 :- set_prolog_flag(backquoted_string, false).
 
+string_handling(atom-1) :-
+	string_to_atom(an_atom, X),
+	X == an_atom.
+string_handling(list-1) :-
+	string_to_atom("a_list", X),
+	X == a_list.
+
 
 		 /*******************************
 		 *	       DYNAMIC		*
@@ -1200,8 +1197,8 @@ proc(retract-2) :-
 	Body == (A is B + 3).
 proc(retract-3) :-
 	assert(myunit(1)),
-	assert(myunit(2) :- x),
-	retract(myunit(2) :- X),
+	assert((myunit(2) :- x)),
+	retract((myunit(2) :- X)),
 	X == x,
 	retractall(myunit(_)).		% cleanup
 proc(current_predicate-1) :-
@@ -1236,11 +1233,15 @@ tcl(a).
 tcl(b) :- true.
 tcl(c) :- write(hello).
 tcl(a(X)) :- b(X).
+tcl(x(G)) :- G.
+tcl(a(X,X)) :- a(X).
 
 mtcl:tcl(a) :- a.
 mtcl:tcl(b) :- a, b.
 mtcl:(tcl(c) :- a, b).
 
+cl(call-1) :-
+	clause(tcl(x(G)), X), X == call(G).
 cl(clause-1) :-
 	clause(tcl(a), X), X == true.
 cl(clause-2) :-
@@ -1252,6 +1253,10 @@ cl(clause-4) :-
 cl(clause-5) :-
 	clause(tcl(H), b(a)), H == a(a).
 cl(clause-6) :-
+	clause(tcl(a(a,X)), B),
+	X == a,
+	B == a(a).
+cl(clause-7) :-
 	clause(mtcl:tcl(H), user:a), H == a.
 
 
@@ -1284,40 +1289,46 @@ erase_all(_).
 record(recorda-1) :-
 	erase_all(r1),
 	mkterm(T0),
-	recorda(r1, T0),
+	recorda(r1, T0, Ref),
 	recorded(r1, T1),
-	T0 =@= T1.
+	T0 =@= T1,
+	erase(Ref).
 record(recorda-2) :-
 	erase_all(r2),
 	mkterm(T0),
 	recorda(r2, T0, Ref),
 	recorded(K, T1, Ref),
 	K == r2,
-	T0 =@= T1.
+	T0 =@= T1,
+	erase(Ref).
 record(recorda-3) :-
 	erase_all(r3),
 	\+ current_key(r3),
-	recorda(r3, test),
-	current_key(r3).
+	recorda(r3, test, Ref),
+	current_key(r3),
+	erase(Ref).
 record(recorda-4) :-
 	erase_all(r4),
-	recorda(r4, aap),
-	recorda(r4, noot),
-	recordz(r4, mies),
+	recorda(r4, aap, R1),
+	recorda(r4, noot, R2),
+	recordz(r4, mies, R3),
 	findall(X, recorded(r4, X), Xs),
-	Xs = [noot, aap, mies].
+	Xs = [noot, aap, mies],
+	erase(R1), erase(R2), erase(R3).
 record(recorda-5) :-
-	recorda(bla,sign(a,(b,c),d)),
+	recorda(bla,sign(a,(b,c),d), Ref),
 	\+ recorded(bla, sign(_,(B,B),_)),
 	\+ (recorded(bla,S),
-	    S=sign(_,(B,B),_)).
+	    S=sign(_,(B,B),_)),
+	erase(Ref).
 record(erase-1) :-
 	erase_all(r5),
 	recorda(r5, aap, R),
-	recorda(r5, noot),
+	recorda(r5, noot, R2),
 	erase(R),
 	findall(X, recorded(r5, X), Xs),
-	Xs = [noot].
+	Xs = [noot],
+	erase(R2).
 record(erase-2) :-
 	retractall(a(_)),
 	assert(a(1), Ref),
@@ -1368,8 +1379,9 @@ compiler(assert-6) :-
 	numlist(0, 100, L),
 	append(L, [_], A),
 	T =.. [x|A],
-	assert(T),
-	T.
+	assert(T, Ref),
+	once(T),
+	erase(Ref).
 compiler(assert-7) :-
 	Body = (a(A), A=[B|C], a(B), a(C)),
 	Clause = (at:-Body),
@@ -1389,11 +1401,6 @@ flag(arith-1) :-
 	flag(f, V, V+1),
 	flag(f, NV, Old),
 	NV == 1.
-flag(arith-2) :-
-	flag(f, Old, 100),
-	flag(f, V, mean(V, 0)),
-	flag(f, NV, Old),
-	NV == 50.
 
 
 		 /*******************************
@@ -1410,7 +1417,8 @@ update(assert-1) :-
 	     a(X),
 	     assert(a(3)),
 	     X = 3
-	   ).
+	   ),
+	retractall(a(_)).
 update(retract-1) :-
 	retractall(a(_)),
 	(   assert(a(1)),
@@ -1420,14 +1428,16 @@ update(retract-1) :-
 	    fail
 	;   findall(X, a(X), Xs),
 	    Xs = [3,3]
-	).
+	),
+	retractall(a(_)).
 update(retract-2) :-
 	retractall(a(_)),
 	assert(a(1)),
 	assert(a(2)),
 	a(X),
 	ignore(retract(a(2))),
-	X = 2.
+	X = 2,
+	retractall(a(_)).
 
 
 		 /*******************************
@@ -1445,24 +1455,6 @@ softcut2(A) :-
 	;   A = 1
 	).
 
-do_block :-
-	exit(notmyblock, ok).
-
-bb(a) :-
-	!(myblock).
-bb(b).
-
-b1 :- b2.
-b1.
-
-b2 :- exit(test, b).
-
-b3 :- b4.
-b3.
-
-b4 :-
-	!(test).
-
 /* c*: tests for handling !
 */
 
@@ -1504,23 +1496,6 @@ control(softcut-1) :-
 	findall(A, softcut1(A), [1,2]).
 control(softcut-2) :-
 	findall(A, softcut2(A), [1]).
-control(block-1) :-
-	catch(block(myblock, do_block, _), E, true),
-	error(E, existence_error(block, notmyblock)).
-control(block-2) :-
-	block(notmyblock, do_block, X),
-	X == ok.
-control(block-3) :-
-	\+ (   block(myblock, bb(X), _),
-	       X == b
-	   ).
-control(block-4) :-
-	block(test, b1, B),
-	B == b,
-	'$get_predicate_attribute'(b1, references, 0).
-control(block-5) :-
-	block(test, b3, _),
-	'$get_predicate_attribute'(b3, references, 0).
 control(cut-1) :-
 	c1.
 control(cut-2) :-
@@ -1595,51 +1570,7 @@ exception(context-2) :-
 	error_context(E, undef/0).
 exception(catch-gc) :-
 	catch(tcatch, E, true),
-	subsumes_chk(ok(error(_,_)), E).
-
-
-		 /*******************************
-		 *	  RESOURCE ERRORS	*
-		 *******************************/
-
-choice.
-choice.
-
-local_overflow :-
-	choice,
-	local_overflow.
-
-global_overflow(X) :-			% Causes gracefully signalled overflow
-	global_overflow(s(X)).
-
-string_overflow(StringList) :-
-	string_overflow2(StringList),
-	is_list(StringList).		% avoid GC of list
-
-string_overflow2([H|T]) :-		% Causes PL_throw() overflow
-	format(string(H), '~txx~1000000|', []),
-	string_overflow2(T).
-
-
-resource(stack-1) :-
-	catch(local_overflow, E, true),
-	E = error(resource_error(stack), local).
-resource(stack-2) :-			% VERY slow with -DO_SECURE
-	catch(global_overflow(0), E, true),
-	E = error(resource_error(stack), global).
-resource(stack-3) :-
-	catch(string_overflow(_), E1, true),
-	E1 = error(resource_error(stack), global),
-	catch(string_overflow(_), E2, true),
-	E2 = error(resource_error(stack), global).
-resource(stack-4) :-
-	catch(string_overflow(_), E1, true),
-	E1 = error(resource_error(stack), global),
-	catch(global_overflow(_), E2, true),
-	E2 = error(resource_error(stack), global).
-resource(stack-5) :-
-	catch(length(_L,10000000), E, true),
-	E = error(resource_error(stack), global).
+	subsumes_term(ok(error(_,_)), E).
 
 
 		 /*******************************
@@ -1651,12 +1582,6 @@ make_data(N, s(X)) :-
 	NN is N - 1,
 	make_data(NN, X).
 
-gc(shift-1) :-
-	(   feature(dynamic_stacks, true)
-	->  true
-	;   MinFree is 400 * 1024,
-	    stack_parameter(global, min_free, _, MinFree)
-	).
 gc(gc-1) :-
 	garbage_collect.
 gc(gc-2) :-			% Beautiful crash.  See compilation of \+
@@ -1671,20 +1596,25 @@ gc(gc-3) :-
 gc(gc-4) :-
 	catch(_,_,garbage_collect).
 gc(gc-5) :-
-	catch(25,_,garbage_collect).
+	G = 25,
+	catch(G,_,garbage_collect).
 gc(gc-6) :-
-	catch(1.25,_,garbage_collect).
+	G = 1.25,
+	catch(G,_,garbage_collect).
 gc(agc-1) :-
 	garbage_collect_atoms.
-gc(agc-2) :-
+gc(agc-2) :-		% not if concurrent: this is too simple.  There
+			% are enough tests for AGC in the rest of the suite.
 	(   current_prolog_flag(agc_margin, Margin),
-	    Margin > 0
-	->  UpTo is Margin*2,
+	    Margin > 0,
+	    \+ current_prolog_flag(test_concurrent, true)
+	->  garbage_collect_atoms,
+	    UpTo is Margin*2+10,
 	    statistics(agc_gained, Gained0),
 	    forall(between(0, UpTo, X), atom_concat(foobar, X, _)),
 	    statistics(agc_gained, Gained1),
 	    Gained is Gained1 - Gained0,
-	    Gained > UpTo - 10		% might be some junk
+	    Gained > UpTo - 20		% might be some junk
 	;   true			% no atom-gc
 	).
 
@@ -1961,8 +1891,10 @@ copy_term(av-3) :-
 	A = f(Z),
 	Y == Z.
 copy_term(av-4) :-
-	freeze(X, true),
-	freeze(X, Done = true),
+	G1 = true,
+	G2 = (Done = true),
+	freeze(X, G1),
+	freeze(X, G2),
 	copy_term(X, Y),
 	X = ok,
 	Done == true,
@@ -1997,7 +1929,11 @@ copy_term(nat-2) :-			% cyclic term
 		 *******************************/
 
 term_hash(simple-1) :-
-	term_hash(aap, 9270206).
+	term_hash(aap, X),
+	memberchk(X, [ 9270206,		% little endian
+		       16674642		% big endian
+		     ]).
+
 term_hash(simple-2) :-			% small int
 	term_hash(42, X),
 	memberchk(X, [ 12280004,	% little endian
@@ -2018,12 +1954,21 @@ term_hash(simple-5) :-
 	string_to_list(S, "hello world"),
 	term_hash(S, 13985775).
 term_hash(compound-1) :-
-	term_hash(hello(world), 12599352).
+	term_hash(hello(world), X),
+	memberchk(X, [ 2512014,		% little endian
+		       4285241		% big endian
+		     ]).
 term_hash(compound-2) :-
-	X = x(a),
-	term_hash(hello(X, X), 5826661).
+	A = x(a),
+	term_hash(hello(A, A), X),
+	memberchk(X, [ 6171734,		% little endian
+		       13034251		% big endian
+		     ]).
 term_hash(compound-3) :-
-	term_hash(hello(x(a), x(a)), 5826661).
+	term_hash(hello(x(a), x(a)), X),
+	memberchk(X, [ 6171734,		% little endian
+		       13034251		% big endian
+		     ]).
 
 
 		 /*******************************
@@ -2135,13 +2080,10 @@ popen(cat-2) :-
 	->  Cmd = 'cmd /c rem true'
 	;   Cmd = true
 	),
-	absolute_file_name(swi('library/MANUAL'), Manual),
-	open(Manual, read, Fd),
 	open(pipe(Cmd), write, Pipe),
-	catch(copy_stream_data(Fd, Pipe),
+	catch(forall(between(1, 10000, _), format(Pipe, '0123456789~n', [])),
 	      E,
 	      true),
-	close(Fd),
 	catch(close(Pipe), _, true),	% ???
 	(   var(E)
 	->  format(user_error, 'No exception?~n', []),
@@ -2187,18 +2129,37 @@ collect_data(C, Fd, [C|T]) :-
 timeout(pipe-1) :-
 	(   current_prolog_flag(pipe, true),
 	    \+ current_prolog_flag(windows, true) % cannot wait on pipes
-	->  open(pipe('echo xx && sleep 2 && echo xx.'), read, In,
+	->  open(pipe('echo + && sleep 1 && echo xx.'), read, In,
 		 [ bom(false)
 		 ]),
-	    set_stream(In, timeout(1)),
+	    set_stream(In, timeout(0.2)),
 	    wait_for_input([In], [In], infinite),
 	    catch(read(In, _), E1, true),
-	    E1 = error(timeout_error(read, _), _),
-	    wait_for_input([In], [In], infinite),
-	    catch(read(In, Term), E2, true),
-	    var(E2),
-	    Term == xx,
-	    close(In)
+	    (	E1 = error(timeout_error(read, _), _)
+	    ->	wait_for_input([In], [In], infinite),
+		catch(read(In, Term), E2, true),
+		(   var(E2)
+		->  (   Term == xx
+		    ->	close(In)
+		    ;	format(user_error, 'Term == ~q~n', [Term]),
+			true
+		    )
+		;   format(user_error, 'E2 == ~q~n', [E2]),
+		    fail
+		)
+	    ;   var(E1)
+	    ->	(   Term == (+ xx)
+		->  close(In),
+		    format(user_error,
+			   'timeout(pipe-1): ~q (machine heavy loaded?)~n',
+			   [Term])
+		;   format(user_error,
+			   'var(E1) && Term == ~q~n', [Term]),
+		    fail
+		)
+	    ;	format(user_error, 'E1 == ~q~n', [E1]),
+		fail
+	    )
 	;   true
 	).
 
@@ -2209,9 +2170,8 @@ timeout(pipe-1) :-
 
 :- dynamic
 	testfile/1.
-:- initialization
-	prolog_load_context(file, File),
-	assert(testfile(File)).
+:- prolog_load_context(file, File),
+   assert(testfile(File)).
 
 root(Root) :-
 	(   current_prolog_flag(windows, true)
@@ -2266,7 +2226,7 @@ file(absfile-2) :-			% canoniseDir() caching issues
 	delete_directory(Y),
 	atom_concat(_, YF, Abs2).
 file(ext-2) :-
-	\+ file_name_extension(foo, _, 'bar.pl'). 	% Bug#69
+	\+ file_name_extension(foo, _, 'bar.pl').	% Bug#69
 file(open-1) :-
 	catch(open(foobar, read, _, [lock(qqq)]), E, true),
 	E =@= error(domain_error(lock, qqq), context(system:open/4, _)).
@@ -2444,7 +2404,7 @@ thread(at_exit-1) :-
 	retract(at_exit_called).
 thread(status-1) :-
 	thread_create(true, Id, []),
-	between(0, 10, _),
+	between(0, 100, _),
 	sleep(0.01),
 	thread_property(Id2, status(Status)),
 	Id2 == Id,
@@ -2470,14 +2430,16 @@ mutex(trylock-1) :-
 	thread_get_message(locked),
 	\+ mutex_trylock(Mutex),
 	thread_send_message(Id, done),
-	thread_join(Id, true).
+	thread_join(Id, true),
+	mutex_destroy(Mutex).
 mutex(unlock-1) :-
 	gensym(mutex, Mutex),
 	mutex_lock(Mutex),
 	mutex_unlock(Mutex),
 	catch(mutex_unlock(Mutex), E, true),
-	E == error(permission_error(mutex, unlock, Mutex),
-		   context(mutex_unlock/1, 'not locked')).
+	E == error(permission_error(unlock, mutex, Mutex),
+		   context(mutex_unlock/1, 'not locked')),
+	mutex_destroy(Mutex).
 mutex(destroy-1) :-
 	gensym(mutex, Mutex),
 	mutex_create(Mutex),
@@ -2513,8 +2475,12 @@ follow_links(File, File).
 run_test_script(Script) :-
 	file_base_name(Script, Base),
 	file_name_extension(Pred, _, Base),
-	load_files(Script, [silent(true)]),
-	Pred.
+	load_files(Script, [silent(true), if(changed)]),
+	(   current_prolog_flag(verbose, normal)
+	->  format('(~w)', [Base]), flush_output
+	;   true
+	),
+	call_test(Pred, script).
 
 run_test_scripts(Directory) :-
 	(   script_dir(ScriptDir),
@@ -2534,10 +2500,10 @@ run_scripts([]).
 run_scripts([H|T]) :-
 	(   catch(run_test_script(H), Except, true)
 	->  (   var(Except)
-	    ->  put(.), flush
+	    ->  put_ok
 	    ;   Except = blocked(Reason)
 	    ->  assert(blocked(H, Reason)),
-		put(!), flush
+		put_blocked
 	    ;   script_failed(H, Except)
 	    )
 	;   script_failed(H, fail)
@@ -2615,7 +2581,6 @@ testset(thread) :-
 	current_prolog_flag(threads, true).
 testset(mutex) :-
 	current_prolog_flag(threads, true).
-testset(resource).
 
 %	unicode_file_locale/0
 %
@@ -2650,6 +2615,7 @@ testdir('Tests/attvar').
 testdir('Tests/library').
 testdir('Tests/charset').
 testdir('Tests/clp').
+testdir('Tests/GC').
 testdir('Tests/thread') :-
 	current_prolog_flag(threads, true).
 
@@ -2662,6 +2628,9 @@ test :-
 	retractall(blocked(_,_)),
 	forall(testset(Set), runtest(Set)),
 	scripts,
+	garbage_collect,
+	garbage_collect_atoms,
+	trim_stacks,
 	statistics,
 	report_blocked,
 	report_failed.
@@ -2690,6 +2659,37 @@ report_failed :-
         ;   format('~nAll tests passed~n', [])
 	).
 
+%%	call_test(:Goal, +Line)
+%
+%	Call the actual test. If dmalloc/3 is provided, call through
+%	this leak-detection hook.
+%
+%	@see test/dmalloc.pl
+
+:- meta_predicate
+	call_test(0, +).
+
+:- if(current_predicate(dmalloc/3)).
+call_test(Goal, script) :- !,
+	dmalloc((Goal->true), '*** Script ~w ***', [Goal]).
+call_test(Goal, Line) :-
+	dmalloc((Goal->true), '*** Line ~d: ~p ***', [Line, Goal]).
+:- elif(current_prolog_flag(test_concurrent, true)).
+call_test(Goal, _Line) :-
+	test_name(Goal, Name),
+	with_mutex(Name, Goal).
+
+test_name(M:G, Name) :- !,
+	functor(G, Pred, Arity),
+	format(atom(Name), 'test ~w:~w/~d', [M,Pred,Arity]).
+test_name(G, Name) :- !,
+	functor(G, Pred, Arity),
+	format(atom(Name), 'test ~w/~d', [Pred,Arity]).
+:- else.
+call_test(Goal, _Line) :-
+	Goal, !.
+:- endif.
+
 runtest(Name) :-
 	format('Running test set "~w" ', [Name]),
 	flush,
@@ -2697,12 +2697,17 @@ runtest(Name) :-
 	findall(Head-R, nth_clause_head(Head, R), Heads),
 	unique_heads(Heads),
 	member(Head-R, Heads),
-	(   catch(Head, Except, true)
+	clause_property(R, line_count(Line)),
+	(   current_prolog_flag(verbose, normal)
+	->  format('(~w)', [Line]), flush_output
+	;   true
+	),
+	(   catch(call_test(Head, Line), Except, true)
 	->  (   var(Except)
-	    ->  put(.), flush_output
+	    ->  put_ok
 	    ;   Except = blocked(Reason)
 	    ->  assert(blocked(Head, Reason)),
-		put(!), flush_output
+		put_blocked
 	    ;   test_failed(R, Except)
 	    )
 	;   test_failed(R, fail)
@@ -2732,6 +2737,18 @@ check_uniqye([_|T]) :-
 	check_uniqye(T).
 
 
+:- if(current_prolog_flag(test_concurrent, true)).
+put_ok :-
+	thread_self(Me),
+	atom_concat(tester_, Id, Me), !,
+	write(user_error, Id).
+:- endif.
+put_ok :-
+	write(user_error, .).
+
+put_blocked :-
+	write(user_error, !).
+
 test_failed(R, Except) :-
 	clause(Head, _, R),
 	functor(Head, Name, 1),
@@ -2757,7 +2774,7 @@ blocked(Reason) :-
 %	2nd context argument.
 
 error(error(Ex, _Ctx), Expected) :-
-	subsumes_chk(Expected, Ex), !.
+	subsumes_term(Expected, Ex), !.
 error(error(Ex, _Ctx), Expected) :-
 	format('~NWrong exception: ~p (expected ~p)~n', [Ex, Expected]),
 	fail.
